@@ -40,6 +40,8 @@ module Reference = struct
     | `Type (r, s) -> render_resolved (r :> t) ^ "." ^ TypeName.to_string s
     | `Constructor (r, s) ->
         render_resolved (r :> t) ^ "." ^ ConstructorName.to_string s
+    | `PolyConstructor (r, s) ->
+        render_resolved (r :> t) ^ ".`" ^ ConstructorName.to_string s
     | `Field (r, s) -> render_resolved (r :> t) ^ "." ^ FieldName.to_string s
     | `Extension (r, s) ->
         render_resolved (r :> t) ^ "." ^ ExtensionName.to_string s
@@ -48,9 +50,8 @@ module Reference = struct
     | `Exception (r, s) ->
         render_resolved (r :> t) ^ "." ^ ExceptionName.to_string s
     | `Value (r, s) -> render_resolved (r :> t) ^ "." ^ ValueName.to_string s
-    | `Class (r, s) -> render_resolved (r :> t) ^ "." ^ ClassName.to_string s
-    | `ClassType (r, s) ->
-        render_resolved (r :> t) ^ "." ^ ClassTypeName.to_string s
+    | `Class (r, s) -> render_resolved (r :> t) ^ "." ^ TypeName.to_string s
+    | `ClassType (r, s) -> render_resolved (r :> t) ^ "." ^ TypeName.to_string s
     | `Method (r, s) ->
         (* CR trefis: do we really want to print anything more than [s] here?  *)
         render_resolved (r :> t) ^ "." ^ MethodName.to_string s
@@ -59,12 +60,25 @@ module Reference = struct
         render_resolved (r :> t) ^ "." ^ InstanceVariableName.to_string s
     | `Label (_, s) -> LabelName.to_string s
 
+  let render_path (tag, cs) =
+    let tag =
+      match tag with
+      | `TRelativePath -> "./"
+      | `TAbsolutePath -> "/"
+      | `TCurrentPackage -> "//"
+    in
+    tag ^ String.concat "/" cs
+
   let rec render_unresolved : Reference.t -> string =
     let open Reference in
     function
     | `Resolved r -> render_resolved r
     | `Root (n, _) -> n
     | `Dot (p, f) -> render_unresolved (p :> t) ^ "." ^ f
+    | `Page_path p -> render_path p
+    | `Asset_path p -> render_path p
+    | `Module_path p -> render_path p
+    | `Any_path p -> render_path p
     | `Module (p, f) ->
         render_unresolved (p :> t) ^ "." ^ ModuleName.to_string f
     | `ModuleType (p, f) ->
@@ -80,9 +94,9 @@ module Reference = struct
     | `Exception (p, f) ->
         render_unresolved (p :> t) ^ "." ^ ExceptionName.to_string f
     | `Value (p, f) -> render_unresolved (p :> t) ^ "." ^ ValueName.to_string f
-    | `Class (p, f) -> render_unresolved (p :> t) ^ "." ^ ClassName.to_string f
+    | `Class (p, f) -> render_unresolved (p :> t) ^ "." ^ TypeName.to_string f
     | `ClassType (p, f) ->
-        render_unresolved (p :> t) ^ "." ^ ClassTypeName.to_string f
+        render_unresolved (p :> t) ^ "." ^ TypeName.to_string f
     | `Method (p, f) ->
         render_unresolved (p :> t) ^ "." ^ MethodName.to_string f
     | `InstanceVariable (p, f) ->
@@ -93,7 +107,7 @@ module Reference = struct
   let to_ir : ?text:Inline.t -> Reference.t -> Inline.t =
    fun ?text ref ->
     match ref with
-    | `Resolved r -> (
+    | `Resolved r ->
         (* IDENTIFIER MUST BE RENAMED TO DEFINITION. *)
         let id = Reference.Resolved.identifier r in
         let rendered = render_resolved r in
@@ -103,18 +117,14 @@ module Reference = struct
           | Some s -> s
         and tooltip =
           (* Add a tooltip if the content is not the rendered reference. *)
-          match text with None -> None | Some _ -> Some rendered
+          match text with
+          | None -> None
+          | Some _ -> Some rendered
         in
-        match Url.from_identifier ~stop_before:false id with
-        | Ok url ->
-            let target = InternalLink.Resolved url in
-            let link = { InternalLink.target; content; tooltip } in
-            [ inline @@ Inline.InternalLink link ]
-        | Error (Not_linkable _) -> content
-        | Error exn ->
-            (* FIXME: better error message *)
-            Printf.eprintf "Id.href failed: %S\n%!" (Url.Error.to_string exn);
-            content)
+        let url = Url.from_identifier ~stop_before:false id in
+        let target = Target.Internal (Resolved url) in
+        let link = { Link.target; content; tooltip } in
+        [ inline @@ Inline.Link link ]
     | _ -> (
         let s = render_unresolved ref in
         match text with
@@ -123,9 +133,9 @@ module Reference = struct
             [ inline @@ Inline.Source s ]
         | Some content ->
             let link =
-              { InternalLink.target = Unresolved; content; tooltip = Some s }
+              { Link.target = Internal Unresolved; content; tooltip = Some s }
             in
-            [ inline @@ Inline.InternalLink link ])
+            [ inline @@ Inline.Link link ])
 end
 
 let leaf_inline_element : Comment.leaf_inline_element -> Inline.one = function
@@ -170,7 +180,7 @@ let rec inline_element : Comment.inline_element -> Inline.t = function
         | [] -> [ inline @@ Text target ]
         | _ -> non_link_inline_element_list content
       in
-      [ inline @@ Link (target, content) ]
+      [ inline @@ Link { target = External target; content; tooltip = None } ]
 
 and inline_element_list elements =
   List.concat
@@ -263,6 +273,32 @@ let rec nestable_block_element :
            and raise warnings *)
       in
       [ block @@ Table { data; align } ]
+  | `Media (href, media, content) ->
+      let content =
+        match (content, href) with
+        | "", `Reference path ->
+            Reference.render_unresolved (path :> Comment.Reference.t)
+        | "", `Link href -> href
+        | _ -> content
+      in
+      let url =
+        match href with
+        | `Reference (`Resolved r) -> (
+            let id =
+              Odoc_model.Paths.Reference.Resolved.Asset.(identifier (r :> t))
+            in
+            match Url.from_asset_identifier id with
+            | url -> Target.Internal (Resolved url))
+        | `Reference _ -> Internal Unresolved
+        | `Link href -> External href
+      in
+      let i =
+        match media with
+        | `Audio -> Block.Audio (url, content)
+        | `Video -> Video (url, content)
+        | `Image -> Image (url, content)
+      in
+      [ block i ]
 
 and paragraph : Comment.paragraph -> Block.one = function
   | [ { value = `Raw_markup (target, s); _ } ] ->
@@ -308,7 +344,14 @@ let tag : Comment.tag -> Description.one =
   | `See (kind, target, content) ->
       let value =
         match kind with
-        | `Url -> mk_value (Inline.Link (target, [ inline @@ Text target ]))
+        | `Url ->
+            mk_value
+              (Inline.Link
+                 {
+                   target = External target;
+                   content = [ inline @@ Text target ];
+                   tooltip = None;
+                 })
         | `File -> mk_value (Inline.Source (source_of_code target))
         | `Document -> mk_value (Inline.Text target)
       in
@@ -372,7 +415,7 @@ let standalone docs =
   Utils.flatmap ~f:item_element
   @@ List.map (fun x -> x.Odoc_model.Location_.value) docs
 
-let to_ir (docs : Comment.docs) =
+let to_ir (docs : Comment.elements) =
   Utils.flatmap ~f:block_element
   @@ List.map (fun x -> x.Odoc_model.Location_.value) docs
 
