@@ -6,6 +6,7 @@ let string_of_syntax = function OCaml -> "ml" | Reason -> "re"
 
 type page = {
   filename : Fpath.t;
+  path : Url.Path.t;
   content : Format.formatter -> unit;
   children : page list;
 }
@@ -23,22 +24,19 @@ type input =
 
 type 'a t = {
   name : string;
-  render : 'a -> Types.Document.t -> page list;
-  extra_documents : 'a -> input -> syntax:syntax -> Types.Document.t list;
+  render : 'a -> Sidebar.t option -> Types.Document.t -> page list;
+  filepath : 'a -> Url.Path.t -> Fpath.t;
 }
 
 let document_of_page ~syntax v =
   match syntax with Reason -> Reason.page v | OCaml -> ML.page v
 
-let documents_of_source_tree ~syntax v =
-  match syntax with Reason -> Reason.source_tree v | OCaml -> ML.source_tree v
+let documents_of_implementation ~syntax v =
+  match syntax with
+  | Reason -> Reason.implementation v
+  | OCaml -> ML.implementation v
 
 let document_of_compilation_unit ~syntax v =
   match syntax with
   | Reason -> Reason.compilation_unit v
   | OCaml -> ML.compilation_unit v
-
-let document_of_source ~syntax =
-  match syntax with
-  | Reason -> Reason.source_page (* Currently, both functions are equivalent *)
-  | OCaml -> ML.source_page
