@@ -14,6 +14,8 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *)
 
+open Odoc_utils
+
 module Url = Odoc_document.Url
 module Html = Tyxml.Html
 
@@ -31,56 +33,77 @@ let html_of_toc toc =
 
 let html_of_search () =
   let search_bar =
-    Html.(input ~a:[ a_class [ "search-bar" ]; a_placeholder "🔎 Search..." ] ())
+    Html.(
+      input
+        ~a:[ a_class [ "search-bar" ]; a_placeholder "🔎 Type '/' to search..." ]
+        ())
   in
   let snake = Html.(div ~a:[ a_class [ "search-snake" ] ] []) in
   let search_result = Html.div ~a:[ Html.a_class [ "search-result" ] ] [] in
   Html.(
     div ~a:[ a_class [ "search-inner" ] ] [ search_bar; snake; search_result ])
 
-let sidebar toc =
-  match toc with
-  | [] -> []
-  | _ -> [ Html.nav ~a:[ Html.a_class [ "odoc-toc" ] ] (html_of_toc toc) ]
-
-let html_of_breadcrumbs (breadcrumbs : Types.breadcrumb list) =
-  let make_navigation ~up_url rest =
-    [
-      Html.nav
-        ~a:[ Html.a_class [ "odoc-nav" ] ]
-        ([ Html.a ~a:[ Html.a_href up_url ] [ Html.txt "Up" ]; Html.txt " – " ]
-        @ rest);
-    ]
+let sidebars ~global_toc ~local_toc =
+  let local_toc =
+    match local_toc with
+    | [] -> []
+    | _ :: _ ->
+        [
+          Html.nav
+            ~a:[ Html.a_class [ "odoc-toc"; "odoc-local-toc" ] ]
+            (html_of_toc local_toc);
+        ]
   in
-  match List.rev breadcrumbs with
-  | [] -> [] (* Can't happen - there's always the current page's breadcrumb. *)
-  | [ _ ] -> [] (* No parents *)
-  | [ { name = "index"; _ }; x ] ->
-      (* Special case leaf pages called 'index' with one parent. This is for files called
-          index.mld that would otherwise clash with their parent. In particular,
-          dune and odig both cause this situation right now. *)
-      let up_url = "../index.html" in
-      let parent_name = x.name in
-      make_navigation ~up_url [ Html.txt parent_name ]
-  | current :: up :: bs ->
-      let space = Html.txt " " in
-      let sep = [ space; Html.entity "#x00BB"; space ] in
-      let html =
-        (* Create breadcrumbs *)
-        Utils.list_concat_map ?sep:(Some sep)
-          ~f:(fun (breadcrumb : Types.breadcrumb) ->
+  let global_toc =
+    match global_toc with
+    | None -> []
+    | Some c ->
+        [ Html.nav ~a:[ Html.a_class [ "odoc-toc"; "odoc-global-toc" ] ] c ]
+  in
+  match local_toc @ global_toc with
+  | [] -> []
+  | tocs -> [ Html.div ~a:[ Html.a_class [ "odoc-tocs" ] ] tocs ]
+
+let html_of_breadcrumbs (breadcrumbs : Types.breadcrumbs) =
+  let make_navigation ~up_url rest =
+    let up =
+      match up_url with
+      | None -> []
+      | Some up_url ->
+          [ Html.a ~a:[ Html.a_href up_url ] [ Html.txt "Up" ]; Html.txt " – " ]
+    in
+    [ Html.nav ~a:[ Html.a_class [ "odoc-nav" ] ] (up @ rest) ]
+  in
+  let space = Html.txt " " in
+  let sep = [ space; Html.entity "#x00BB"; space ] in
+  let html =
+    (* Create breadcrumbs *)
+    List.concat_map_sep ~sep
+      ~f:(fun (breadcrumb : Types.breadcrumb) ->
+        match breadcrumb.href with
+        | Some href ->
             [
               [
                 Html.a
-                  ~a:[ Html.a_href breadcrumb.href ]
-                  [ Html.txt breadcrumb.name ];
+                  ~a:[ Html.a_href href ]
+                  (breadcrumb.name
+                    :> Html_types.flow5_without_interactive Html.elt list);
               ];
-            ])
-          (up :: bs)
-        |> List.flatten
-      in
-      make_navigation ~up_url:up.href
-        (List.rev html @ sep @ [ Html.txt current.name ])
+            ]
+        | None ->
+            [ (breadcrumb.name :> Html_types.nav_content_fun Html.elt list) ])
+      breadcrumbs.parents
+    |> List.flatten
+  in
+  let current_name :> Html_types.nav_content_fun Html.elt list =
+    breadcrumbs.current.name
+  in
+  let rest =
+    if List.is_empty breadcrumbs.parents then current_name
+    else html @ sep @ current_name
+  in
+  make_navigation ~up_url:breadcrumbs.up_url
+    (rest :> [< Html_types.nav_content_fun > `A `PCDATA `Wbr ] Html.elt list)
 
 let file_uri ~config ~url (base : Types.uri) file =
   match base with
@@ -96,7 +119,7 @@ let default_meta_elements ~config ~url =
     Html.meta ~a:[ Html.a_charset "utf-8" ] ();
     Html.link ~rel:[ `Stylesheet ] ~href:odoc_css_uri ();
     Html.meta
-      ~a:[ Html.a_name "generator"; Html.a_content "odoc 2.4.4" ]
+      ~a:[ Html.a_name "generator"; Html.a_content "odoc 3.0.0" ]
       ();
     Html.meta
       ~a:
@@ -107,7 +130,8 @@ let default_meta_elements ~config ~url =
       ();
   ]
 
-let page_creator ~config ~url ~uses_katex header breadcrumbs toc content =
+let page_creator ~config ~url ~uses_katex ~global_toc header breadcrumbs
+    local_toc content =
   let theme_uri = Config.theme_uri config in
   let support_uri = Config.support_uri config in
   let search_uris = Config.search_uris config in
@@ -115,7 +139,7 @@ let page_creator ~config ~url ~uses_katex header breadcrumbs toc content =
 
   let head : Html_types.head Html.elt =
     let title_string =
-      Printf.sprintf "%s (%s)" url.name (String.concat "." path)
+      Printf.sprintf "%s (%s)" url.name (String.concat ~sep:"." path)
     in
 
     let file_uri = file_uri ~config ~url in
@@ -133,7 +157,7 @@ let page_creator ~config ~url ~uses_katex header breadcrumbs toc content =
           let search_urls =
             let search_url name = Printf.sprintf "'%s'" name in
             let search_urls = List.map search_url search_urls in
-            "[" ^ String.concat "," search_urls ^ "]"
+            "[" ^ String.concat ~sep:"," search_urls ^ "]"
           in
           (* The names of the search scripts are put into a js variable. Then
              the code in [odoc_search.js] load them into a webworker. *)
@@ -209,7 +233,7 @@ let search_urls = %s;
     html_of_breadcrumbs breadcrumbs
     @ search_bar
     @ [ Html.header ~a:[ Html.a_class [ "odoc-preamble" ] ] header ]
-    @ sidebar toc
+    @ sidebars ~global_toc ~local_toc
     @ [ Html.div ~a:[ Html.a_class [ "odoc-content" ] ] content ]
   in
 
@@ -222,21 +246,23 @@ let search_urls = %s;
   in
   content
 
-let make ~config ~url ~header ~breadcrumbs ~toc ~uses_katex content children =
-  let filename = Link.Path.as_filename ~is_flat:(Config.flat config) url in
+let make ~config ~url ~header ~breadcrumbs ~sidebar ~toc ~uses_katex content
+    children =
+  let filename = Link.Path.as_filename ~config url in
   let content =
-    page_creator ~config ~url ~uses_katex header breadcrumbs toc content
+    page_creator ~config ~url ~uses_katex ~global_toc:sidebar header breadcrumbs
+      toc content
   in
-  { Odoc_document.Renderer.filename; content; children }
+  { Odoc_document.Renderer.filename; content; children; path = url }
 
 let path_of_module_of_source ppf url =
   match url.Url.Path.parent with
   | Some parent ->
       let path = Link.Path.for_printing parent in
-      Format.fprintf ppf " (%s)" (String.concat "." path)
+      Format.fprintf ppf " (%s)" (String.concat ~sep:"." path)
   | None -> ()
 
-let src_page_creator ~breadcrumbs ~config ~url ~header name content =
+let src_page_creator ~breadcrumbs ~config ~url ~header ~sidebar name content =
   let head : Html_types.head Html.elt =
     let title_string =
       Format.asprintf "Source: %s%a" name path_of_module_of_source url
@@ -247,6 +273,7 @@ let src_page_creator ~breadcrumbs ~config ~url ~header name content =
   let body =
     html_of_breadcrumbs breadcrumbs
     @ [ Html.header ~a:[ Html.a_class [ "odoc-preamble" ] ] header ]
+    @ sidebars ~global_toc:sidebar ~local_toc:[]
     @ content
   in
   (* We never indent as there is a bug in tyxml and it would break lines inside
@@ -262,9 +289,9 @@ let src_page_creator ~breadcrumbs ~config ~url ~header name content =
   in
   content
 
-let make_src ~config ~url ~breadcrumbs ~header title content =
-  let filename = Link.Path.as_filename ~is_flat:(Config.flat config) url in
+let make_src ~config ~url ~breadcrumbs ~header ~sidebar title content =
+  let filename = Link.Path.as_filename ~config url in
   let content =
-    src_page_creator ~breadcrumbs ~config ~url ~header title content
+    src_page_creator ~breadcrumbs ~config ~url ~header ~sidebar title content
   in
-  { Odoc_document.Renderer.filename; content; children = [] }
+  { Odoc_document.Renderer.filename; content; children = []; path = url }
