@@ -1,22 +1,16 @@
 open Odoc_document.Types
 open Types
 module Doctree = Odoc_document.Doctree
-
-let rec list_concat_map ?sep ~f = function
-  | [] -> []
-  | [ x ] -> f x
-  | x :: xs -> (
-      let hd = f x in
-      let tl = list_concat_map ?sep ~f xs in
-      match sep with None -> hd @ tl | Some sep -> hd @ (sep :: tl))
+module Url = Odoc_document.Url
 
 module Link = struct
   let rec flatten_path ppf (x : Odoc_document.Url.Path.t) =
-    match x.parent with
-    | Some p ->
-        Fmt.pf ppf "%a-%a-%s" flatten_path p Odoc_document.Url.Path.pp_kind
-          x.kind x.name
-    | None -> Fmt.pf ppf "%a-%s" Odoc_document.Url.Path.pp_kind x.kind x.name
+    let pp_parent ppf = function
+      | Some p -> Format.fprintf ppf "%a-" flatten_path p
+      | None -> ()
+    in
+    Format.fprintf ppf "%a%a%s" pp_parent x.parent
+      Url.Path.pp_disambiguating_prefix x.kind x.name
 
   let page p = Format.asprintf "%a" flatten_path p
 
@@ -48,10 +42,10 @@ module Link = struct
     ( List.map segment_to_string dir,
       String.concat "." (List.map segment_to_string file) )
 
-  let filename url =
+  let filename ?(add_ext = true) url =
     let dir, file = get_dir_and_file url in
     let file = Fpath.(v (String.concat dir_sep (dir @ [ file ]))) in
-    Fpath.(add_ext "tex" file)
+    if add_ext then Fpath.add_ext "tex" file else file
 end
 
 let style = function
@@ -97,19 +91,13 @@ let list kind pp ppf x =
 
 let escape_entity = function "#45" -> "-" | "gt" -> ">" | s -> s
 
-let filter_map f x =
-  List.rev
-  @@ List.fold_left
-       (fun acc x -> match f x with Some x -> x :: acc | None -> acc)
-       [] x
-
 let elt_size (x : elt) =
   match x with
   | Txt _ | Internal_ref _ | External_ref _ | Label _ | Style _ | Inlined_code _
   | Code_fragment _ | Tag _ | Break _ | Ligaturable _ ->
       Small
   | List _ | Section _ | Verbatim _ | Raw _ | Code_block _ | Indented _
-  | Description _ ->
+  | Description _ | Image _ ->
       Large
   | Table _ | Layout_table _ -> Huge
 
@@ -126,7 +114,9 @@ let layout_table = function
         | Empty, _ -> None
         | (Small | Large | Huge), x -> Some x
       in
-      let filter_row row = filter_map filter_empty @@ List.combine mask row in
+      let filter_row row =
+        Odoc_utils.List.filter_map filter_empty @@ List.combine mask row
+      in
       let row_size = List.fold_left max Empty mask in
       [ Layout_table { row_size; tbl = List.map filter_row m } ]
 
@@ -140,13 +130,11 @@ let entity ~in_source ~verbatim x =
   if in_source && not verbatim then Ligaturable (escape_entity x)
   else Txt [ escape_entity x ]
 
-(** Tables with too many rows are hard to typeset correctly on
-    the same page.
+(** Tables with too many rows are hard to typeset correctly on the same page.
     Splitting tables on multiple pages is unreliable with longtable + hyperref.
-    Thus we limit the height of the tables that we render as latex tables.
-    This variable is kept separated because we may want to make it tunable
-    by the user.
-*)
+    Thus we limit the height of the tables that we render as latex tables. This
+    variable is kept separated because we may want to make it tunable by the
+    user. *)
 let small_table_height_limit = 10
 
 let rec pp_elt ppf = function
@@ -179,6 +167,7 @@ let rec pp_elt ppf = function
   | Indented x -> Raw.indent pp ppf x
   | Ligaturable s -> Fmt.string ppf s
   | Tag (s, t) -> tag s ppf t
+  | Image target -> Raw.includegraphics Fpath.pp ppf target
 
 and pp ppf = function
   | [] -> ()
@@ -228,16 +217,16 @@ let source k (t : Source.t) =
     | Elt i -> k i
     | Tag (None, l) -> tokens l
     | Tag (Some s, l) -> [ Tag (s, tokens l) ]
-  and tokens t = list_concat_map t ~f:token in
+  and tokens t = Odoc_utils.List.concat_map token t in
   tokens t
 
-let rec internalref ~verbatim ~in_source (t : InternalLink.t) =
+let rec internalref ~verbatim ~in_source (t : Target.internal) (c : Inline.t) =
   let target =
-    match t.target with
-    | InternalLink.Resolved uri -> Link.label uri
+    match t with
+    | Target.Resolved uri -> Link.label uri
     | Unresolved -> "xref-unresolved"
   in
-  let text = Some (inline ~verbatim ~in_source t.content) in
+  let text = Some (inline ~verbatim ~in_source c) in
   let short = in_source in
   Internal_ref { short; target; text }
 
@@ -247,10 +236,11 @@ and inline ~in_source ~verbatim (l : Inline.t) =
     | Text _s -> assert false
     | Linebreak -> [ Break Line ]
     | Styled (style, c) -> [ Style (style, inline ~verbatim ~in_source c) ]
-    | Link (ext, c) ->
+    | Link { target = External ext; content = c; _ } ->
         let content = inline ~verbatim:false ~in_source:false c in
         [ External_ref (ext, Some content) ]
-    | InternalLink c -> [ internalref ~in_source ~verbatim c ]
+    | Link { target = Internal ref_; content = c; _ } ->
+        [ internalref ~in_source ~verbatim ref_ c ]
     | Source c ->
         [ Inlined_code (source (inline ~verbatim:false ~in_source:true) c) ]
     | Math s -> [ Raw (Format.asprintf "%a" Raw.math s) ]
@@ -288,10 +278,30 @@ let non_empty_code_fragment c =
   let s = source (inline ~verbatim:false ~in_source:true) c in
   match s with [] -> [] | _ :: _ as l -> [ Code_fragment l ]
 
+let alt_text ~in_source (target : Target.t) alt =
+  let text = txt ~verbatim:false ~in_source:false [ alt ] in
+  let break = if in_source then [] else [ Break Paragraph ] in
+  match target with
+  | Internal _ -> text @ break
+  | External l -> [ External_ref (l, Some text) ] @ break
+
+let image ~in_source (internal_url : Url.t) alt =
+  let dir, file = Link.get_dir_and_file internal_url.page in
+  match Fpath.(get_ext @@ v file) with
+  (* list imported from pdftex.def *)
+  | "" | ".pdf" | ".png" | ".jpg" | ".mps" | ".jpeg" | ".jbig2" | ".jb2"
+  | ".PDF" | ".PNG" | ".JPG" | ".JPEG" | ".JBIG2" | ".JB2" ->
+      let fpath = Fpath.v (String.concat Fpath.dir_sep (dir @ [ file ])) in
+      [ Image fpath ]
+  | _ -> alt_text ~in_source (Internal (Resolved internal_url)) alt
+
 let rec block ~in_source (l : Block.t) =
   let one (t : Block.one) =
     match t.desc with
     | Inline i -> inline ~verbatim:false ~in_source:false i
+    | Image (Internal (Resolved x), alt) -> image ~in_source x alt
+    | Image (t, alt) | Audio (t, alt) | Video (t, alt) ->
+        alt_text ~in_source t alt
     | Paragraph i ->
         inline ~in_source:false ~verbatim:false i
         @ if in_source then [] else [ Break Paragraph ]
@@ -316,7 +326,7 @@ let rec block ~in_source (l : Block.t) =
           Break Paragraph;
         ]
   in
-  list_concat_map l ~f:one
+  Odoc_utils.List.concat_map one l
 
 and table_block { Table.data; align } =
   let data =
@@ -463,7 +473,7 @@ module Doc = struct
       if with_children then link_children ppf children else ()
     in
     let content ppf = Fmt.pf ppf "@[<v>%a@,%t@]@." pp content children_input in
-    { Odoc_document.Renderer.filename; content; children }
+    { Odoc_document.Renderer.filename; content; children; path = url }
 end
 
 module Page = struct
@@ -477,11 +487,12 @@ module Page = struct
     List.flatten @@ List.map (subpage ~with_children) subpages
 
   and page ~with_children p =
-    let { Page.preamble; items = i; url; _ } =
+    let { Page.items = i; url; _ } =
       Doctree.Labels.disambiguate_page ~enter_subpages:true p
     and subpages = subpages ~with_children @@ Doctree.Subpages.compute p in
     let i = Doctree.Shift.compute ~on_sub i in
-    let header = items (Doctree.PageTitle.render_title p @ preamble) in
+    let header, preamble = Doctree.PageTitle.render_title p in
+    let header = items (header @ preamble) in
     let content = items i in
     let page = Doc.make ~with_children url (header @ content) subpages in
     page
@@ -489,4 +500,6 @@ end
 
 let render ~with_children = function
   | Document.Page page -> [ Page.page ~with_children page ]
-  | Source_page _ | Asset _ -> []
+  | Source_page _ -> []
+
+let filepath url = Link.filename ~add_ext:false url
