@@ -1,8 +1,7 @@
 (** Raw latex primitives:
     - macro definitions
     - env defitions
-    - text escaping
-*)
+    - text escaping *)
 
 type pr = Format.formatter -> unit
 
@@ -127,6 +126,8 @@ let code_block pp ppf x =
   Fmt.cut ppf ();
   mend ppf name
 
+let includegraphics pp = create "includegraphics" pp
+
 let section pp = create "section" pp
 
 let subsection pp = create "subsection" pp
