@@ -25,11 +25,17 @@ open Odoc_model.Lang
 
 module Env = Ident_env
 
+type env = Cmi.env = {
+  ident_env : Ident_env.t;
+  warnings_tag : string option;
+}
+
+
 let read_core_type env ctyp =
   Cmi.read_type_expr env ctyp.ctyp_type
 
 let rec read_pattern env parent doc pat =
-  let locs _id = None in
+  let source_loc = None in
   let open Signature in
     match pat.pat_desc with
     | Tpat_any -> []
@@ -39,22 +45,22 @@ let rec read_pattern env parent doc pat =
     | Tpat_var(id,_,_uid) ->
 #endif
         let open Value in
-        let id = Env.find_value_identifier env id in
+        let id = Env.find_value_identifier env.ident_env id in
           Cmi.mark_type_expr pat.pat_type;
           let type_ = Cmi.read_type_expr env pat.pat_type in
           let value = Abstract in
-          [Value {id; locs = locs id; doc; type_; value}]
+          [Value {id; source_loc; doc; type_; value}]
 #if OCAML_VERSION < (5,2, 0)
     | Tpat_alias(pat, id, _) ->
 #else
     | Tpat_alias(pat, id, _,_) ->
 #endif
         let open Value in
-        let id = Env.find_value_identifier env id in
+        let id = Env.find_value_identifier env.ident_env id in
           Cmi.mark_type_expr pat.pat_type;
           let type_ = Cmi.read_type_expr env pat.pat_type in
           let value = Abstract in
-          Value {id; locs = locs id; doc; type_; value} :: read_pattern env parent doc pat
+          Value {id; source_loc; doc; type_; value} :: read_pattern env parent doc pat
     | Tpat_constant _ -> []
     | Tpat_tuple pats ->
         List.concat (List.map (read_pattern env parent doc) pats)
@@ -85,7 +91,7 @@ let rec read_pattern env parent doc pat =
 
 let read_value_binding env parent vb =
   let container = (parent : Identifier.Signature.t :> Identifier.LabelParent.t) in
-  let doc = Doc_attr.attached_no_tag container vb.vb_attributes in
+  let doc = Doc_attr.attached_no_tag ~warnings_tag:env.warnings_tag container vb.vb_attributes in
     read_pattern env parent doc vb.vb_pat
 
 let read_value_bindings env parent vbs =
@@ -95,7 +101,7 @@ let read_value_bindings env parent vbs =
       (fun acc vb ->
          let open Signature in
         let comments =
-          Doc_attr.standalone_multiple container vb.vb_attributes in
+          Doc_attr.standalone_multiple container ~warnings_tag:env.warnings_tag vb.vb_attributes in
          let comments = List.map (fun com -> Comment com) comments in
          let vb = read_value_binding env parent vb in
           List.rev_append vb (List.rev_append comments acc))
@@ -105,9 +111,9 @@ let read_value_bindings env parent vbs =
 
 let read_type_extension env parent tyext =
   let open Extension in
-  let type_path = Env.Path.read_type env tyext.tyext_path in
+  let type_path = Env.Path.read_type env.ident_env tyext.tyext_path in
   let container = (parent : Identifier.Signature.t :> Identifier.LabelParent.t) in
-  let doc = Doc_attr.attached_no_tag container tyext.tyext_attributes in
+  let doc = Doc_attr.attached_no_tag ~warnings_tag:env.warnings_tag container tyext.tyext_attributes in
   let type_params =
     List.map (fun (ctyp, _) -> ctyp.ctyp_type) tyext.tyext_params
   in
@@ -136,14 +142,14 @@ let read_type_extension env parent tyext =
     rendered. For example, [constraint] items are read separately and not
     associated with their comment. *)
 let mk_class_comment = function
-  | [] -> None
+  | { Odoc_model.Comment.elements = []; _} -> None
   | doc -> Some (ClassSignature.Comment (`Docs doc))
 
 let rec read_class_type_field env parent ctf =
   let open ClassSignature in
   let open Odoc_model.Names in
   let container = (parent : Identifier.ClassSignature.t :> Identifier.LabelParent.t) in
-  let doc = Doc_attr.attached_no_tag container ctf.ctf_attributes in
+  let doc = Doc_attr.attached_no_tag ~warnings_tag:env.warnings_tag container ctf.ctf_attributes in
   match ctf.ctf_desc with
   | Tctf_val(name, mutable_, virtual_, typ) ->
       let open InstanceVariable in
@@ -164,7 +170,7 @@ let rec read_class_type_field env parent ctf =
       let expr = read_class_signature env parent [] cltyp in
       Some (Inherit {Inherit.expr; doc})
   | Tctf_attribute attr ->
-      match Doc_attr.standalone container attr with
+      match Doc_attr.standalone container ~warnings_tag:env.warnings_tag attr with
       | None -> None
       | Some doc -> Some (Comment doc)
 
@@ -172,7 +178,7 @@ and read_class_signature env parent params cltyp =
   let open ClassType in
     match cltyp.cltyp_desc with
     | Tcty_constr(p, _, params) ->
-        let p = Env.Path.read_class_type env p in
+        let p = Env.Path.read_class_type env.ident_env p in
         let params = List.map (read_core_type env) params in
           Constr(p, params)
     | Tcty_signature csig ->
@@ -193,7 +199,7 @@ and read_class_signature env parent params cltyp =
         let items, (doc, doc_post) = Doc_attr.extract_top_comment_class items in
         let items =
           match doc_post with
-          | [] -> items
+          | { elements = []; _ } -> items
           | _ -> Comment (`Docs doc_post) :: items
         in
         Signature {self; items; doc}
@@ -224,7 +230,7 @@ let rec read_class_field env parent cf =
   let open ClassSignature in
   let open Odoc_model.Names in
   let container = (parent : Identifier.ClassSignature.t :> Identifier.LabelParent.t) in
-  let doc = Doc_attr.attached_no_tag container (cf.cf_attributes) in
+  let doc = Doc_attr.attached_no_tag ~warnings_tag:env.warnings_tag container (cf.cf_attributes) in
   match cf.cf_desc with
   | Tcf_val({txt = name; _}, mutable_, _, kind, _) ->
       let open InstanceVariable in
@@ -264,7 +270,7 @@ let rec read_class_field env parent cf =
       Some (Inherit {Inherit.expr; doc})
   | Tcf_initializer _ -> mk_class_comment doc
   | Tcf_attribute attr ->
-      match Doc_attr.standalone container attr with
+      match Doc_attr.standalone container ~warnings_tag:env.warnings_tag attr with
       | None -> None
       | Some doc -> Some (Comment doc)
 
@@ -289,7 +295,7 @@ and read_class_structure env parent params cl =
         let items, (doc, doc_post) = Doc_attr.extract_top_comment_class items in
         let items =
           match doc_post with
-          | [] -> items
+          | { elements = []; _ } -> items
           | _ -> Comment (`Docs doc_post) :: items
         in
         Signature {self; items; doc}
@@ -331,10 +337,10 @@ let rec read_class_expr env parent params cl =
 
 let read_class_declaration env parent cld =
   let open Class in
-  let id = Env.find_class_identifier env cld.ci_id_class in
-  let locs = None in
+  let id = Env.find_class_identifier env.ident_env cld.ci_id_class in
+  let source_loc = None in
   let container = (parent : Identifier.Signature.t :> Identifier.LabelParent.t) in
-  let doc = Doc_attr.attached_no_tag container cld.ci_attributes in
+  let doc = Doc_attr.attached_no_tag ~warnings_tag:env.warnings_tag container cld.ci_attributes in
     Cmi.mark_class_declaration cld.ci_decl;
     let virtual_ = (cld.ci_virt = Virtual) in
     let clparams =
@@ -346,13 +352,13 @@ let read_class_declaration env parent cld =
         clparams
     in
     let type_ = read_class_expr env (id :> Identifier.ClassSignature.t) clparams cld.ci_expr in
-    { id; locs; doc; virtual_; params; type_; expansion = None }
+    { id; source_loc; doc; virtual_; params; type_; expansion = None }
 
 let read_class_declarations env parent clds =
   let container = (parent : Identifier.Signature.t :> Identifier.LabelParent.t) in
   let open Signature in
   List.fold_left begin fun (acc, recursive) cld ->
-    let comments = Doc_attr.standalone_multiple container cld.ci_attributes in
+    let comments = Doc_attr.standalone_multiple container ~warnings_tag:env.warnings_tag cld.ci_attributes in
     let comments = List.map (fun com -> Comment com) comments in
     let cld = read_class_declaration env parent cld in
     ((Class (recursive, cld))::(List.rev_append comments acc), And)
@@ -378,8 +384,8 @@ let rec read_module_expr env parent label_parent mexpr =
               let id, env =
                 match id_opt with
                 | None -> Identifier.Mk.parameter (parent, Odoc_model.Names.ModuleName.make_std "_"), env
-                | Some id -> let env = Env.add_parameter parent id (ModuleName.of_ident id) env in
-                  Env.find_parameter_identifier env id, env
+                | Some id -> let e' = Env.add_parameter parent id (ModuleName.of_ident id) env.ident_env in
+                  Env.find_parameter_identifier e' id, {env with ident_env=e'}
               in
               let arg = Cmti.read_module_type env (id :> Identifier.Signature.t) label_parent arg in
 
@@ -389,12 +395,13 @@ let rec read_module_expr env parent label_parent mexpr =
         Functor (f_parameter, res)
 #else
     | Tmod_functor(id, _, arg, res) ->
-        let new_env = Env.add_parameter parent id (ModuleName.of_ident id) env in
+        let new_env = Env.add_parameter parent id (ModuleName.of_ident id) env.ident_env in
+        let new_env = {env with ident_env = new_env} in
         let f_parameter =
           match arg with
           | None -> FunctorParameter.Unit
           | Some arg ->
-              let id = Env.find_parameter_identifier new_env id in
+              let id = Env.find_parameter_identifier new_env.ident_env id in
               let arg = Cmti.read_module_type env (id :> Identifier.Signature.t) label_parent arg in
               Named { FunctorParameter. id; expr = arg; }
         in
@@ -435,17 +442,17 @@ and read_module_binding env parent mb =
       match mb.mb_id with
       | None -> None
       | Some id ->
-        let id = Env.find_module_identifier env id in
+        let mid = Env.find_module_identifier env.ident_env id in
 #else
-  let id = Env.find_module_identifier env mb.mb_id in
+  let mid = Env.find_module_identifier env.ident_env mb.mb_id in
 #endif
-  let id = (id :> Identifier.Module.t) in
-  let locs = None in
+  let id = (mid :> Identifier.Module.t) in
+  let source_loc = None in
   let container = (parent : Identifier.Signature.t :> Identifier.LabelParent.t) in
-  let doc, canonical = Doc_attr.attached Odoc_model.Semantics.Expect_canonical container mb.mb_attributes in
+  let doc, canonical = Doc_attr.attached ~warnings_tag:env.warnings_tag Odoc_model.Semantics.Expect_canonical container mb.mb_attributes in
   let type_, canonical =
     match unwrap_module_expr_desc mb.mb_expr.mod_desc with
-    | Tmod_ident (p, _) -> (Alias (Env.Path.read_module env p, None), canonical)
+    | Tmod_ident (p, _) -> (Alias (Env.Path.read_module env.ident_env p, None), canonical)
     | _ ->
         let id = (id :> Identifier.Signature.t) in
         let expr, canonical =
@@ -453,19 +460,19 @@ and read_module_binding env parent mb =
         in
         (ModuleType expr, canonical)
   in
-  let canonical = (canonical :> Path.Module.t option) in
+  let canonical = match canonical with | None -> None | Some s -> Some (Doc_attr.conv_canonical_module s) in
   let hidden =
 #if OCAML_VERSION >= (4,10,0)
-    match canonical, mb.mb_id with
-    | None, Some id -> Odoc_model.Root.contains_double_underscore (Ident.name id)
-    | _, _ -> false
+    match canonical, mid.iv with
+    | None, (`Module (_, n) | `Parameter (_, n) | `Root (_, n)) -> Odoc_model.Names.ModuleName.is_hidden n
+    | Some _, _ -> false
 #else
-    match canonical with
-    | None -> Odoc_model.Root.contains_double_underscore (Ident.name mb.mb_id)
-    | _ -> false
+    match canonical, mid.iv with
+    | None, (`Module (_, n) | `Parameter (_, n) | `Root (_, n)) -> Odoc_model.Names.ModuleName.is_hidden n
+    | Some _, _ -> false
 #endif
   in
-  Some {id; locs; doc; type_; canonical; hidden; }
+  Some {id; source_loc; doc; type_; canonical; hidden; }
 
 and read_module_bindings env parent mbs =
   let container = (parent : Identifier.Signature.t :> Identifier.LabelParent.t)
@@ -473,7 +480,7 @@ and read_module_bindings env parent mbs =
   let open Signature in
   List.fold_left
     (fun (acc, recursive) mb ->
-      let comments = Doc_attr.standalone_multiple container mb.mb_attributes in
+      let comments = Doc_attr.standalone_multiple container ~warnings_tag:env.warnings_tag mb.mb_attributes in
       let comments = List.map (fun com -> Comment com) comments in
       match read_module_binding env parent mb with
       | Some mb ->
@@ -543,7 +550,7 @@ and read_structure_item env parent item =
           Cmti.read_class_type_declarations env parent cltyps
     | Tstr_attribute attr ->
       let container = (parent : Identifier.Signature.t :> Identifier.LabelParent.t) in
-          match Doc_attr.standalone container attr with
+          match Doc_attr.standalone container ~warnings_tag:env.warnings_tag attr with
           | None -> []
           | Some doc -> [Comment doc]
 
@@ -551,11 +558,12 @@ and read_include env parent incl =
   let open Include in
   let loc = Doc_attr.read_location incl.incl_loc in
   let container = (parent : Identifier.Signature.t :> Identifier.LabelParent.t) in
-  let doc, status = Doc_attr.attached Odoc_model.Semantics.Expect_status container incl.incl_attributes in
+  let doc, status = Doc_attr.attached ~warnings_tag:env.warnings_tag Odoc_model.Semantics.Expect_status container incl.incl_attributes in
   let decl_modty =
     match unwrap_module_expr_desc incl.incl_mod.mod_desc with
     | Tmod_ident(p, _) ->
-      Some (ModuleType.U.TypeOf {t_desc = ModuleType.StructInclude (Env.Path.read_module env p); t_expansion=None })
+      let p = Env.Path.read_module env.ident_env p in
+      Some (ModuleType.U.TypeOf (ModuleType.StructInclude p, p))
     | _ ->
       let mty = read_module_expr env parent container incl.incl_mod in
       umty_of_mty mty
@@ -571,7 +579,7 @@ and read_include env parent incl =
 
 and read_open env parent o =
   let container = (parent : Identifier.Signature.t :> Identifier.LabelParent.t) in
-  let doc = Doc_attr.attached_no_tag container o.open_attributes in
+  let doc = Doc_attr.attached_no_tag ~warnings_tag:env.warnings_tag container o.open_attributes in
   #if OCAML_VERSION >= (4,8,0)
   let signature = o.open_bound_items in
   #else
@@ -584,7 +592,8 @@ and read_structure :
       'tags. 'tags Odoc_model.Semantics.handle_internal_tags -> _ -> _ -> _ ->
       _ * 'tags =
  fun internal_tags env parent str ->
-  let env = Env.add_structure_tree_items parent str env in
+  let e' = Env.add_structure_tree_items parent str env.ident_env in
+  let env = { env with ident_env=e' } in
   let items, (doc, doc_post), tags =
     let classify item =
       match item.str_desc with
@@ -592,7 +601,7 @@ and read_structure :
       | Tstr_attribute attr -> Some (`Attribute attr)
       | _ -> None
     in
-    Doc_attr.extract_top_comment internal_tags ~classify parent str.str_items
+    Doc_attr.extract_top_comment internal_tags ~warnings_tag:env.warnings_tag ~classify parent str.str_items
   in
   let items =
     List.fold_left
@@ -602,16 +611,25 @@ and read_structure :
     |> List.rev
   in
   match doc_post with
-  | [] ->
-    ({ Signature.items; compiled = false; doc }, tags)
+  | { elements = [] ; _} ->
+    ({ Signature.items; compiled = false; removed = []; doc }, tags)
   | _ ->
-    ({ Signature.items = Comment (`Docs doc_post) :: items; compiled=false; doc }, tags)
+    ({ Signature.items = Comment (`Docs doc_post) :: items; compiled=false; removed = []; doc }, tags)
 
-let read_implementation root name impl =
-  let id = Identifier.Mk.root (root, Odoc_model.Names.ModuleName.make_std name) in
-  let sg, canonical =
-    read_structure Odoc_model.Semantics.Expect_canonical (Env.empty ()) id impl
+let read_implementation root name ~warnings_tag impl =
+  let id =
+    Identifier.Mk.root (root, Odoc_model.Names.ModuleName.make_std name)
   in
-  (id, sg, (canonical :> Odoc_model.Paths.Path.Module.t option))
+  let sg, canonical =
+    read_structure Odoc_model.Semantics.Expect_canonical
+      { ident_env = Env.empty (); warnings_tag }
+      id impl
+  in
+  let canonical =
+    match canonical with
+    | None -> None
+    | Some s -> Some (Doc_attr.conv_canonical_module s)
+  in
+  (id, sg, canonical)
 
 let _ = Cmti.read_module_expr := read_module_expr
