@@ -20,8 +20,6 @@ open Names
 module Id = Paths.Identifier
 module P = Paths.Path
 
-type type_ident = Paths.Identifier.Path.Type.t
-
 module LocHashtbl = Hashtbl.Make(struct
     type t = Location.t
     let equal l1 l2 = l1 = l2
@@ -41,7 +39,7 @@ type t =
     classes : Id.Class.t Ident.tbl;
     class_types : Id.ClassType.t Ident.tbl;
     loc_to_ident : Id.t LocHashtbl.t;
-    hidden : Ident.t list; (* we use term hidden to mean shadowed and idents_in_doc_off_mode items*)
+    shadowed : Ident.t list;
   }
 
 let empty () =
@@ -57,7 +55,7 @@ let empty () =
     classes = Ident.empty;
     class_types = Ident.empty;
     loc_to_ident = LocHashtbl.create 100;
-    hidden = [];
+    shadowed = [];
   }
 
 (* The boolean is an override for whether it should be hidden - true only for
@@ -83,10 +81,34 @@ type items =
   | `Include of item list
 ]
 
+let extract_visibility =
+  let open Compat in
+  function
+  | Sig_type (_, _, _, vis)
+  | Sig_module (_, _, _, _, vis)
+  | Sig_modtype (_, _, vis)
+  | Sig_value (_, _, vis)
+  | Sig_class (_, _, _, vis)
+  | Sig_class_type (_, _, _, vis)
+  | Sig_typext (_, _, _, vis) ->
+      vis
+
 let rec extract_signature_type_items vis items =
   let open Compat in
-    match items with
-    | Sig_type(id, td, _, vis') :: rest when vis=vis' ->
+  match items with
+  | item :: rest ->
+    let vis' = extract_visibility item in
+    if vis = vis' then
+      let hidden = vis' = Hidden in
+      extract_signature_type_items_extract vis ~hidden item rest
+    else
+      extract_signature_type_items_skip vis item rest
+  | [] -> []
+
+and extract_signature_type_items_extract vis ~hidden item rest =
+  let open Compat in
+  match item, rest with
+    | Sig_type(id, td, _, _), _ ->
       if Btype.is_row_name (Ident.name id)
       then extract_signature_type_items vis rest
       else
@@ -104,56 +126,64 @@ let rec extract_signature_type_items vis items =
 #endif
             List.map (fun c -> `Constructor (c.Types.cd_id, id, Some c.cd_loc)) cstrs
           | Type_open -> [] in
-        `Type (id, vis'=Hidden, None) :: constrs @ extract_signature_type_items vis rest 
+        `Type (id, hidden, None) :: constrs @ extract_signature_type_items vis rest
 
-    | Sig_module(id, _, _, _, vis') :: rest when vis=vis' ->
-      `Module (id, vis'=Hidden, None) :: extract_signature_type_items vis rest
+    | Sig_module(id, _, _, _, _), _ ->
+      `Module (id, hidden, None) :: extract_signature_type_items vis rest
 
-    | Sig_modtype(id, _, vis') :: rest when vis=vis' ->
-      `ModuleType (id, vis'=Hidden, None) :: extract_signature_type_items vis rest
-    
-    | Sig_value(id, _, vis') :: rest when vis=vis' ->
-      `Value (id, vis'=Hidden, None) :: extract_signature_type_items vis rest
+    | Sig_modtype(id, _, _), _ ->
+      `ModuleType (id, hidden, None) :: extract_signature_type_items vis rest
+
+    | Sig_value(id, _, _), _ ->
+      `Value (id, hidden, None) :: extract_signature_type_items vis rest
 #if OCAML_VERSION < (5,1,0)
-    | Sig_class(id, _, _, vis') :: Sig_class_type(ty_id, _, _, _)
-        :: Sig_type(obj_id, _, _, _) :: Sig_type(cl_id, _, _, _) :: rest when vis=vis' ->
-      `Class (id, ty_id, obj_id, Some cl_id, vis'=Hidden, None) :: extract_signature_type_items vis rest
+    | Sig_class(id, _, _, _),
+      Sig_class_type(ty_id, _, _, _)
+      :: Sig_type(obj_id, _, _, _)
+      :: Sig_type(cl_id, _, _, _) :: _ ->
+      `Class (id, ty_id, obj_id, Some cl_id, hidden, None)
+      :: extract_signature_type_items vis rest
 
-    | Sig_class_type(id, _, _, vis') :: Sig_type(obj_id, _, _, _)
-      :: Sig_type(cl_id, _, _, _) :: rest when vis=vis' ->
-      `ClassType (id, obj_id, Some cl_id, vis'=Hidden, None) :: extract_signature_type_items vis rest
+    | Sig_class_type(id, _, _, _),
+      Sig_type(obj_id, _, _, _) :: Sig_type(cl_id, _, _, _) :: _ ->
+      `ClassType (id, obj_id, Some cl_id, hidden, None)
+      :: extract_signature_type_items vis rest
 #else
-    | Sig_class(id, _, _, vis') :: Sig_class_type(ty_id, _, _, _)
-        :: Sig_type(obj_id, _, _, _) :: rest when vis=vis' ->
-      `Class (id, ty_id, obj_id, None, vis'=Hidden, None) :: extract_signature_type_items vis rest
+    | Sig_class(id, _, _, _),
+      Sig_class_type(ty_id, _, _, _) :: Sig_type(obj_id, _, _, _) :: _ ->
+      `Class (id, ty_id, obj_id, None, hidden, None)
+      :: extract_signature_type_items vis rest
 
-    | Sig_class_type(id, _, _, vis') :: Sig_type(obj_id, _, _, _)  :: rest when vis=vis' ->
-      `ClassType (id, obj_id, None, vis'=Hidden, None) :: extract_signature_type_items vis rest
+    | Sig_class_type(id, _, _, _), Sig_type(obj_id, _, _, _) :: _ ->
+      `ClassType (id, obj_id, None, hidden, None)
+      :: extract_signature_type_items vis rest
 #endif
 
-    | Sig_typext (id, constr, Text_exception, vis') :: rest when vis=vis' ->
+    | Sig_typext (id, constr, Text_exception, _), _ ->
       `Exception (id, Some constr.ext_loc)
       :: extract_signature_type_items vis rest
 
-    | Sig_typext (id, constr, _, vis') :: rest when vis=vis'->
+    | Sig_typext (id, constr, _, _), _ ->
       `Extension (id, Some constr.ext_loc)
       :: extract_signature_type_items vis rest
 
-    | Sig_class_type(_, _, _, _) :: Sig_type(_, _, _, _)
-      :: Sig_type(_, _, _, _) :: rest
-    | Sig_class(_, _, _, _) :: Sig_class_type(_, _, _, _)
-        :: Sig_type(_, _, _, _) :: Sig_type(_, _, _, _) :: rest
-    | Sig_typext (_,_,_,_) :: rest
-    | Sig_modtype(_, _, _) :: rest
-    | Sig_module(_, _, _, _, _) :: rest
-    | Sig_type(_, _, _, _) :: rest
-    | Sig_value (_, _, _) :: rest ->
-        extract_signature_type_items vis rest
+  | Sig_class _, _
+  | Sig_class_type _, _ -> assert false
 
-    | Sig_class _ :: _
-    | Sig_class_type _ :: _ -> assert false
+and extract_signature_type_items_skip vis item rest =
+  let open Compat in
+  match item, rest with
+  | Sig_class_type _, Sig_type _ :: Sig_type _ :: rest
+  | Sig_class _, Sig_class_type _ :: Sig_type _ :: Sig_type _ :: rest
+  | Sig_typext  _, rest
+  | Sig_modtype _, rest
+  | Sig_module _, rest
+  | Sig_type _, rest
+  | Sig_value  _, rest ->
+    extract_signature_type_items vis rest
 
-    | [] -> []
+  | Sig_class _, _
+  | Sig_class_type _, _ -> assert false
 
 #if OCAML_VERSION >= (4,8,0)
 
@@ -163,14 +193,6 @@ let extract_extended_open o =
 #endif
 
 
-let concat_map f l =
-  let rec aux f acc = function
-    | [] -> List.rev acc
-    | x :: l ->
-       let xs = f x in
-       aux f (List.rev_append xs acc) l
-  in aux f [] l
-
 let rec extract_signature_tree_items : bool -> Typedtree.signature_item list -> items list = fun hide_item items ->
   let open Typedtree in
   match items with
@@ -179,7 +201,7 @@ let rec extract_signature_tree_items : bool -> Typedtree.signature_item list ->
 #else
   | { sig_desc = Tsig_type (_, decls); _} :: rest ->
 #endif
-    concat_map (fun decl ->
+    Odoc_utils.List.concat_map (fun decl ->
       if Btype.is_row_name (Ident.name decl.typ_id)
       then []
       else
@@ -315,7 +337,7 @@ let rec extract_structure_tree_items : bool -> Typedtree.structure_item list ->
 #else
     | { str_desc = Tstr_type (_, decls); _ } :: rest -> (* TODO: handle rec_flag *)
 #endif
-  concat_map (fun decl ->
+  Odoc_utils.List.concat_map (fun decl ->
       `Type (decl.typ_id, hide_item, Some decl.typ_loc) ::
         (match decl.typ_kind with
           Ttype_abstract -> []
@@ -449,15 +471,15 @@ let add_items : Id.Signature.t -> item list -> t -> t = fun parent items env ->
     match items with
     | `Type (t, is_hidden_item, loc) :: rest ->
       let name = Ident.name t in
-      let is_hidden = is_hidden_item || type_name_exists name rest in
-        let identifier, hidden =
-        if is_hidden
-        then Mk.type_(parent, TypeName.internal_of_string name), t :: env.hidden
-        else Mk.type_(parent, TypeName.make_std name), env.hidden
+      let is_shadowed = type_name_exists name rest in
+      let identifier, shadowed =
+        if is_shadowed
+        then Mk.type_(parent, TypeName.shadowed_of_string name), t :: env.shadowed
+        else Mk.type_(parent, (if is_hidden_item then TypeName.hidden_of_string else TypeName.make_std) name), env.shadowed
       in
       let types = Ident.add t identifier env.types in
       (match loc with | Some l -> LocHashtbl.add env.loc_to_ident l (identifier :> Id.any) | _ -> ());
-      inner rest { env with types; hidden }
+      inner rest { env with types; shadowed }
 
     | `Constructor (t, t_parent, loc) :: rest ->
       let name = Ident.name t in
@@ -485,54 +507,53 @@ let add_items : Id.Signature.t -> item list -> t -> t = fun parent items env ->
 
     | `Value (t, is_hidden_item, loc) :: rest ->
       let name = Ident.name t in
-      let is_hidden = is_hidden_item || value_name_exists name rest in
-      let identifier, hidden =
-        if is_hidden
-        then Mk.value(parent, ValueName.internal_of_string name), t :: env.hidden
-        else Mk.value(parent, ValueName.make_std name), env.hidden
+      let is_shadowed = value_name_exists name rest in
+      let identifier, shadowed =
+        if is_shadowed
+        then Mk.value(parent, ValueName.shadowed_of_string name), t :: env.shadowed
+        else Mk.value(parent, (if is_hidden_item then ValueName.hidden_of_string else ValueName.make_std) name), env.shadowed
       in
       let values = Ident.add t identifier env.values in
       (match loc with | Some l -> LocHashtbl.add env.loc_to_ident l (identifier :> Id.any) | _ -> ());
-      inner rest { env with values; hidden }
+      inner rest { env with values; shadowed }
 
     | `ModuleType (t, is_hidden_item, loc) :: rest ->
       let name = Ident.name t in
-      let is_hidden = is_hidden_item || module_type_name_exists name rest in
-      let identifier, hidden =
-        if is_hidden
-        then Mk.module_type(parent, ModuleTypeName.internal_of_string name), t :: env.hidden
-        else Mk.module_type(parent, ModuleTypeName.make_std name), env.hidden
+      let is_shadowed = module_type_name_exists name rest in
+      let identifier, shadowed =
+        if is_shadowed
+        then Mk.module_type(parent, ModuleTypeName.shadowed_of_string name), t :: env.shadowed
+        else Mk.module_type(parent,(if is_hidden_item then ModuleTypeName.hidden_of_string else ModuleTypeName.make_std) name), env.shadowed
       in
       let module_types = Ident.add t identifier env.module_types in
       (match loc with | Some l -> LocHashtbl.add env.loc_to_ident l (identifier :> Id.any) | _ -> ());
-      inner rest { env with module_types; hidden }
+      inner rest { env with module_types; shadowed }
 
     | `Module (t, is_hidden_item, loc) :: rest ->
       let name = Ident.name t in
-      let double_underscore = Odoc_model.Root.contains_double_underscore name in
-      let is_hidden = is_hidden_item || module_name_exists name rest || double_underscore in
-      let identifier, hidden =
-        if is_hidden 
-        then Mk.module_(parent, ModuleName.internal_of_string name), t :: env.hidden
-        else Mk.module_(parent, ModuleName.make_std name), env.hidden
+      let is_shadowed = module_name_exists name rest in
+      let identifier, shadowed =
+        if is_shadowed 
+        then Mk.module_(parent, ModuleName.shadowed_of_string name), t :: env.shadowed
+        else Mk.module_(parent, (if is_hidden_item then ModuleName.hidden_of_string else ModuleName.make_std) name), env.shadowed
       in
-      let path = `Identifier(identifier, is_hidden) in
+      let path = `Identifier(identifier, is_hidden_item || is_shadowed) in
       let modules = Ident.add t identifier env.modules in
       let module_paths = Ident.add t path env.module_paths in
       (match loc with | Some l -> LocHashtbl.add env.loc_to_ident l (identifier :> Id.any) | _ -> ());
-      inner rest { env with modules; module_paths; hidden }
+      inner rest { env with modules; module_paths; shadowed }
 
     | `Class (t,t2,t3,t4, is_hidden_item, loc) :: rest ->
       let name = Ident.name t in
-      let is_hidden = is_hidden_item || class_name_exists name rest in
+      let is_shadowed = class_name_exists name rest in
       let class_types = match t4 with
         | None -> [t;t2;t3]
         | Some t4 -> [t;t2;t3;t4]
       in
-      let identifier, hidden =
-        if is_hidden 
-        then Mk.class_(parent, ClassName.internal_of_string name), class_types @ env.hidden
-        else Mk.class_(parent, ClassName.make_std name), env.hidden
+      let identifier, shadowed =
+        if is_shadowed 
+        then Mk.class_(parent, TypeName.shadowed_of_string name), class_types @ env.shadowed
+        else Mk.class_(parent, (if is_hidden_item then TypeName.hidden_of_string else TypeName.make_std) name), env.shadowed
       in
 
       let classes =
@@ -541,25 +562,25 @@ let add_items : Id.Signature.t -> item list -> t -> t = fun parent items env ->
       
       (match loc with | Some l -> LocHashtbl.add env.loc_to_ident l (identifier :> Id.any) | _ -> ());
 
-      inner rest { env with classes; hidden }
+      inner rest { env with classes; shadowed }
 
     | `ClassType (t,t2,t3, is_hidden_item, loc) :: rest ->
       let name = Ident.name t in
-      let is_hidden = is_hidden_item || class_type_name_exists name rest in
+      let is_shadowed = class_type_name_exists name rest in
       let class_types = match t3 with
         | None -> [t;t2]
         | Some t3 -> [t;t2;t3]
       in
-      let identifier, hidden =
-        if is_hidden 
-        then Mk.class_type(parent, ClassTypeName.internal_of_string name), class_types @ env.hidden
-        else Mk.class_type(parent, ClassTypeName.make_std name), env.hidden
+      let identifier, shadowed =
+        if is_shadowed 
+        then Mk.class_type(parent, TypeName.shadowed_of_string name), class_types @ env.shadowed
+        else Mk.class_type(parent, (if is_hidden_item then TypeName.hidden_of_string else TypeName.make_std) name), env.shadowed
       in
       let class_types =
         List.fold_right (fun id class_types -> Ident.add id identifier class_types)
           class_types env.class_types in
       (match loc with | Some l -> LocHashtbl.add env.loc_to_ident l (identifier :> Id.any) | _ -> ());
-      inner rest { env with class_types; hidden }
+      inner rest { env with class_types; shadowed }
 
     | [] -> env
     in inner items env
@@ -621,16 +642,15 @@ let find_extension_identifier env id =
 let find_value_identifier env id =
   Ident.find_same id env.values
 
-(** Lookup a type in the environment. If it isn't found, it's assumed to be a
-    core type. *)
+(** Lookup a type in the environment. If it isn't found, it means it's a core
+    type. *)
 let find_type env id =
-  try (Ident.find_same id env.types :> Id.Path.Type.t)
+  try Some (Ident.find_same id env.types :> Id.Path.Type.t)
   with Not_found -> (
-    try (Ident.find_same id env.classes :> Id.Path.Type.t)
+    try Some (Ident.find_same id env.classes :> Id.Path.Type.t)
     with Not_found -> (
-      try (Ident.find_same id env.class_types :> Id.Path.Type.t)
-      with Not_found ->
-        (Paths.Identifier.Mk.core_type (Ident.name id) :> type_ident)))
+      try Some (Ident.find_same id env.class_types :> Id.Path.Type.t)
+      with Not_found -> None))
 
 let find_class_type env id =
   try
@@ -646,11 +666,11 @@ let find_class_type_identifier env id =
 
 let is_shadowed
  env id =
-    List.mem id env.hidden
+    List.mem id env.shadowed
 module Path = struct
 
   let read_module_ident env id =
-    if Ident.persistent id then `Root (Ident.name id)
+    if Ident.persistent id then `Root (ModuleName.of_ident id)
     else
       try find_module env id
       with Not_found -> assert false
@@ -661,13 +681,18 @@ module Path = struct
     with Not_found -> assert false
 
   let read_type_ident env id =
-    `Identifier (find_type env id, false)
+    match find_type env id with
+    | Some id -> `Identifier (id , false)
+    | None -> `Resolved (`CoreType (TypeName.of_ident id))
+
+  let read_value_ident env id : Paths.Path.Value.t =
+    `Identifier (find_value_identifier env id, false)
 
   let read_class_type_ident env id : Paths.Path.ClassType.t =
     try
       `Identifier (find_class_type env id, false)
     with Not_found ->
-      `Dot(`Root "*", (Ident.name id))
+      `DotT (`Root (ModuleName.make_std "*"), (TypeName.of_ident id))
       (* TODO remove this hack once the fix for PR#6650
          is in the OCaml release *)
 
@@ -692,9 +717,9 @@ module Path = struct
   let rec read_module : t -> Path.t -> Paths.Path.Module.t = fun env -> function
     | Path.Pident id -> read_module_ident env id
 #if OCAML_VERSION >= (4,8,0)
-    | Path.Pdot(p, s) -> `Dot(read_module env p, s)
+    | Path.Pdot(p, s) -> `Dot(read_module env p, ModuleName.make_std s)
 #else
-    | Path.Pdot(p, s, _) -> `Dot(read_module env p, s)
+    | Path.Pdot(p, s, _) -> `Dot(read_module env p, ModuleName.make_std s)
 #endif
     | Path.Papply(p, arg) -> `Apply(read_module env p, read_module env arg)
 #if OCAML_VERSION >= (5,1,0)
@@ -704,9 +729,9 @@ module Path = struct
   let read_module_type env = function
     | Path.Pident id -> read_module_type_ident env id
 #if OCAML_VERSION >= (4,8,0)
-    | Path.Pdot(p, s) -> `Dot(read_module env p, s)
+    | Path.Pdot(p, s) -> `DotMT(read_module env p, ModuleTypeName.make_std s)
 #else
-    | Path.Pdot(p, s, _) -> `Dot(read_module env p, s)
+    | Path.Pdot(p, s, _) -> `DotMT(read_module env p, ModuleTypeName.make_std s)
 #endif
     | Path.Papply(_, _)-> assert false
 #if OCAML_VERSION >= (5,1,0)
@@ -716,9 +741,9 @@ module Path = struct
   let read_class_type env = function
     | Path.Pident id -> read_class_type_ident env id
 #if OCAML_VERSION >= (4,8,0)
-    | Path.Pdot(p, s) -> `Dot(read_module env p, strip_hash s)
+    | Path.Pdot(p, s) -> `DotT(read_module env p, TypeName.make_std (strip_hash s))
 #else
-    | Path.Pdot(p, s, _) -> `Dot(read_module env p, strip_hash s)
+    | Path.Pdot(p, s, _) -> `DotT(read_module env p, TypeName.make_std (strip_hash s))
 #endif
     | Path.Papply(_, _)-> assert false
 #if OCAML_VERSION >= (5,1,0)
@@ -732,15 +757,27 @@ module Path = struct
 #endif
     | Path.Pident id -> read_type_ident env id
 #if OCAML_VERSION >= (4,8,0)
-    | Path.Pdot(p, s) -> `Dot(read_module env p, strip_hash s)
+    | Path.Pdot(p, s) -> `DotT(read_module env p, TypeName.make_std (strip_hash s))
 #else
-    | Path.Pdot(p, s, _) -> `Dot(read_module env p, strip_hash s)
+    | Path.Pdot(p, s, _) -> `DotT(read_module env p, TypeName.make_std (strip_hash s))
 #endif
     | Path.Papply(_, _)-> assert false
 #if OCAML_VERSION >= (5,1,0)
     | Path.Pextra_ty (p,_) -> read_type env p
 #endif
 
+  let read_value env = function
+    | Path.Pident id -> read_value_ident env id
+#if OCAML_VERSION >= (4,8,0)
+    | Path.Pdot(p, s) -> `DotV(read_module env p, ValueName.make_std s)
+#else
+    | Path.Pdot(p, s, _) -> `DotV(read_module env p, ValueName.make_std s)
+#endif
+    | Path.Papply(_, _) -> assert false
+#if OCAML_VERSION >= (5,1,0)
+    | Path.Pextra_ty _ -> assert false
+#endif
+
 end
 
 module Fragment = struct
