@@ -40,6 +40,8 @@ module Path : sig
   val read_type : t -> Path.t -> Paths.Path.Type.t
 
   val read_class_type : t -> Path.t -> Paths.Path.ClassType.t
+
+  val read_value : t -> Path.t -> Paths.Path.Value.t
 end
 
 val find_module : t -> Ident.t -> Paths.Path.Module.t
@@ -53,7 +55,7 @@ val find_module_type : t -> Ident.t -> Paths.Identifier.ModuleType.t
 
 val find_value_identifier : t -> Ident.t -> Paths.Identifier.Value.t
 
-val find_type : t -> Ident.t -> Paths.Identifier.Path.Type.t
+val find_type : t -> Ident.t -> Paths.Identifier.Path.Type.t option
 
 val find_constructor_identifier : t -> Ident.t -> Paths.Identifier.Constructor.t
 
