@@ -1,6 +1,8 @@
+module ManLink = Link
 open Odoc_document
 open Types
 open Doctree
+module Link = ManLink
 
 (*
 Manpages relies on the (g|t|n)roff document language.
@@ -247,7 +249,7 @@ let strip l =
               { h with desc = Styled (sty, List.rev @@ loop [] content) }
             in
             loop (h :: acc) t
-        | Link (_, content) | InternalLink { content; _ } ->
+        | Link { content; _ } ->
             let acc = loop acc content in
             loop acc t
         | Source code ->
@@ -303,9 +305,9 @@ and inline (l : Inline.t) =
           x ++ inline rest
       | Linebreak -> break ++ inline rest
       | Styled (sty, content) -> style sty (inline content) ++ inline rest
-      | Link (href, content) ->
+      | Link { target = External href; content; _ } ->
           env "UR" "UE" href (inline @@ strip content) ++ inline rest
-      | InternalLink { content; _ } ->
+      | Link { content; _ } ->
           font "CI" (inline @@ strip content) ++ inline rest
       | Source content -> source_code content ++ inline rest
       | Math s -> math s ++ inline rest
@@ -365,6 +367,8 @@ let rec block (l : Block.t) =
       let continue r = if r = [] then noop else vspace ++ block r in
       match b.desc with
       | Inline i -> inline i ++ continue rest
+      | Video (_, content) | Audio (_, content) | Image (_, content) ->
+          str "%s" content ++ continue rest
       | Paragraph i -> inline i ++ continue rest
       | List (list_typ, l) ->
           let f n b =
@@ -538,9 +542,8 @@ let on_sub subp =
 
 let page p =
   reset_heading ();
-  let header =
-    Doctree.PageTitle.render_title p @ Shift.compute ~on_sub p.preamble
-  in
+  let header, preamble = Doctree.PageTitle.render_title p in
+  let header = header @ Shift.compute ~on_sub preamble in
   let i = Shift.compute ~on_sub p.items in
   macro "TH" {|%s 3 "" "Odoc" "OCaml Library"|} p.url.name
   ++ macro "SH" "Name"
@@ -558,8 +561,10 @@ and render_page (p : Page.t) =
   and children = Utils.flatmap ~f:subpage @@ Subpages.compute p in
   let content ppf = Format.fprintf ppf "%a@." Roff.pp (page p) in
   let filename = Link.as_filename p.url in
-  { Renderer.filename; content; children }
+  { Renderer.filename; content; children; path = p.url }
 
 let render = function
   | Document.Page page -> [ render_page page ]
-  | Source_page _ | Asset _ -> []
+  | Source_page _ -> []
+
+let filepath url = Link.as_filename ~add_ext:false url
