@@ -3,11 +3,9 @@ open Odoc_document
 let for_printing url = List.map snd @@ Url.Path.to_list url
 
 let segment_to_string (kind, name) =
-  match kind with
-  | `Module | `Page | `LeafPage | `Class -> name
-  | _ -> Format.asprintf "%a-%s" Odoc_document.Url.Path.pp_kind kind name
+  Format.asprintf "%a%s" Url.Path.pp_disambiguating_prefix kind name
 
-let as_filename (url : Url.Path.t) =
+let as_filename ?(add_ext = true) (url : Url.Path.t) =
   let components = Url.Path.to_list url in
   let dir, path =
     Url.Path.split
@@ -17,7 +15,7 @@ let as_filename (url : Url.Path.t) =
   let dir = List.map segment_to_string dir in
   let path = String.concat "." (List.map segment_to_string path) in
   let str_path = String.concat Fpath.dir_sep (dir @ [ path ]) in
-  Fpath.(v str_path + ".3o")
+  if add_ext then Fpath.(v str_path + ".3o") else Fpath.v str_path
 
 let rec is_class_or_module_path (url : Url.Path.t) =
   match url.kind with
