@@ -1,3 +1,5 @@
+open Odoc_utils
+
 module Path = Paths.Path
 module Reference = Paths.Reference
 module Identifier = Paths.Identifier
@@ -8,6 +10,8 @@ type style = [ `Bold | `Italic | `Emphasis | `Superscript | `Subscript ]
 
 type alignment = [ `Left | `Center | `Right ]
 
+type media = [ `Image | `Audio | `Video ]
+
 type raw_markup_target = string
 
 type leaf_inline_element =
@@ -52,6 +56,10 @@ type 'a abstract_table = {
   align : alignment option list option;
 }
 
+type media_href = [ `Link of string | `Reference of Reference.Asset.t ]
+
+type media_element = [ `Media of media_href * media * string ]
+
 type nestable_block_element =
   [ `Paragraph of paragraph
   | `Code_block of
@@ -64,7 +72,7 @@ type nestable_block_element =
   | `Table of nestable_block_element abstract_table
   | `List of
     [ `Unordered | `Ordered ] * nestable_block_element with_location list list
-  ]
+  | media_element ]
 
 type tag =
   [ `Author of string
@@ -105,7 +113,9 @@ type block_element =
     heading_attrs * Identifier.Label.t * inline_element with_location list
   | `Tag of tag ]
 
-type docs = block_element with_location list
+type elements = block_element with_location list
+
+type docs = { elements : elements; warnings_tag : string option }
 
 type docs_or_stop = [ `Docs of docs | `Stop ]
 
@@ -128,3 +138,29 @@ let rec link_content_of_inline_element :
 
 and link_content_of_inline_elements l =
   l |> List.map link_content_of_inline_element |> List.concat
+
+let find_zero_heading docs : link_content option =
+  List.find_map
+    (fun doc ->
+      match doc.Location_.value with
+      | `Heading ({ heading_level = `Title; _ }, _, h_content) ->
+          Some (link_content_of_inline_elements h_content)
+      | _ -> None)
+    docs
+
+(* Used in particular to sort the title names *)
+let to_string (l : link_content) =
+  let rec s_of_i (i : non_link_inline_element) =
+    match i with
+    | `Code_span s -> s
+    | `Word w -> w
+    | `Math_span m -> m
+    | `Space -> " "
+    | `Styled (_, is) -> s_of_is is
+    | `Raw_markup (_, r) -> r
+  and s_of_is is =
+    is
+    |> List.map (fun { Location_.value; _ } -> s_of_i value)
+    |> String.concat ~sep:""
+  in
+  s_of_is l
