@@ -17,17 +17,6 @@
 open Paths
 
 (** {3 Modules} *)
-module Source_info = struct
-  type annotation =
-    | Definition of Paths.Identifier.SourceLocation.t
-    | Value of Paths.Identifier.SourceLocation.t
-
-  type 'a with_pos = 'a * (int * int)
-
-  type infos = annotation with_pos list
-
-  type t = { id : Identifier.SourcePage.t; infos : infos }
-end
 
 module rec Module : sig
   type decl =
@@ -36,8 +25,9 @@ module rec Module : sig
 
   type t = {
     id : Identifier.Module.t;
-    locs : Identifier.SourceLocation.t option;
-        (** Identifier.SourceLocation might not be set when the module is artificially constructed from a functor argument. *)
+    source_loc : Identifier.SourceLocation.t option;
+        (** Identifier.SourceLocation might not be set when the module is
+            artificially constructed from a functor argument. *)
     doc : Comment.docs;
     type_ : decl;
     canonical : Path.Module.t option;
@@ -81,17 +71,18 @@ and ModuleType : sig
 
   type typeof_t = {
     t_desc : type_of_desc;
+    t_original_path : Path.Module.t;
     t_expansion : simple_expansion option;
   }
 
   module U : sig
+    (* Unexpanded (aside from Signature, obviously) *)
+
     type expr =
       | Path of Path.ModuleType.t
       | Signature of Signature.t
       | With of substitution list * expr
-      | TypeOf of typeof_t
-
-    (* Nb. this may have an expansion! *)
+      | TypeOf of type_of_desc * Path.Module.t
   end
 
   type path_t = {
@@ -114,7 +105,7 @@ and ModuleType : sig
 
   type t = {
     id : Identifier.ModuleType.t;
-    locs : Identifier.SourceLocation.t option;
+    source_loc : Identifier.SourceLocation.t option;
         (** Can be [None] for module types created by a type substitution. *)
     doc : Comment.docs;
     canonical : Path.ModuleType.t option;
@@ -162,9 +153,15 @@ and Signature : sig
     | Include of Include.t
     | Comment of Comment.docs_or_stop
 
+  type removed_item =
+    | RModule of Names.ModuleName.t * Path.Module.t
+    | RType of Names.TypeName.t * TypeExpr.t * TypeDecl.Equation.t
+    | RModuleType of Names.ModuleTypeName.t * ModuleType.expr
+
   type t = {
     items : item list;
     compiled : bool;
+    removed : removed_item list;
     doc : Comment.docs;  (** The top comment. *)
   }
 end =
@@ -179,12 +176,12 @@ end =
 
 and Include : sig
   type shadowed = {
-    s_modules : string list;
-    s_module_types : string list;
-    s_values : string list;
-    s_types : string list;
-    s_classes : string list;
-    s_class_types : string list;
+    s_modules : (string * Names.ModuleName.t) list;
+    s_module_types : (string * Names.ModuleTypeName.t) list;
+    s_values : (string * Names.ValueName.t) list;
+    s_types : (string * Names.TypeName.t) list;
+    s_classes : (string * Names.TypeName.t) list;
+    s_class_types : (string * Names.TypeName.t) list;
   }
 
   type expansion = { shadowed : shadowed; content : Signature.t }
@@ -255,7 +252,7 @@ and TypeDecl : sig
 
   type t = {
     id : Identifier.Type.t;
-    locs : Identifier.SourceLocation.t option;
+    source_loc : Identifier.SourceLocation.t option;
     doc : Comment.docs;
     canonical : Path.Type.t option;
     equation : Equation.t;
@@ -270,7 +267,7 @@ and Extension : sig
   module Constructor : sig
     type t = {
       id : Identifier.Extension.t;
-      locs : Identifier.SourceLocation.t option;
+      source_loc : Identifier.SourceLocation.t option;
       doc : Comment.docs;
       args : TypeDecl.Constructor.argument;
       res : TypeExpr.t option;
@@ -292,7 +289,7 @@ end =
 and Exception : sig
   type t = {
     id : Identifier.Exception.t;
-    locs : Identifier.SourceLocation.t option;
+    source_loc : Identifier.SourceLocation.t option;
     doc : Comment.docs;
     args : TypeDecl.Constructor.argument;
     res : TypeExpr.t option;
@@ -307,7 +304,7 @@ and Value : sig
 
   type t = {
     id : Identifier.Value.t;
-    locs : Identifier.SourceLocation.t option;
+    source_loc : Identifier.SourceLocation.t option;
     value : value;
     doc : Comment.docs;
     type_ : TypeExpr.t;
@@ -324,7 +321,7 @@ and Class : sig
 
   type t = {
     id : Identifier.Class.t;
-    locs : Identifier.SourceLocation.t option;
+    source_loc : Identifier.SourceLocation.t option;
     doc : Comment.docs;
     virtual_ : bool;
     params : TypeDecl.param list;
@@ -343,7 +340,7 @@ and ClassType : sig
 
   type t = {
     id : Identifier.ClassType.t;
-    locs : Identifier.SourceLocation.t option;
+    source_loc : Identifier.SourceLocation.t option;
     doc : Comment.docs;
     virtual_ : bool;
     params : TypeDecl.param list;
@@ -435,7 +432,7 @@ and TypeExpr : sig
     type t = { path : Path.ModuleType.t; substitutions : substitution list }
   end
 
-  type label = Label of string | Optional of string
+  type label = Label of string | RawOptional of string | Optional of string
 
   type t =
     | Var of string
@@ -484,57 +481,97 @@ module rec Compilation_unit : sig
     content : content;
     expansion : Signature.t option;
     linked : bool;  (** Whether this unit has been linked. *)
+    source_loc : Identifier.SourceLocation.t option;
     canonical : Path.Module.t option;
-    source_info : Source_info.t option;
+  }
+end =
+  Compilation_unit
+
+module rec Source_info : sig
+  type 'a jump_to_impl =
+    | Unresolved of 'a
+    | Resolved of Identifier.SourceLocation.t
+
+  type 'a jump_to = {
+    documentation : 'a option;
+    implementation : 'a jump_to_impl option;
+  }
+
+  type annotation =
+    | Definition of Paths.Identifier.SourceLocation.t
+    | Value of Path.Value.t jump_to
+    | Module of Path.Module.t jump_to
+    | ModuleType of Path.ModuleType.t jump_to
+    | Type of Path.Type.t jump_to
+
+  type 'a with_pos = 'a * (int * int)
+
+  type t = annotation with_pos list
+end =
+  Source_info
+
+module rec Implementation : sig
+  type t = {
+    id : Identifier.SourcePage.t option;
+    digest : Digest.t;
+    root : Root.t;
+    linked : bool;  (** Whether this unit has been linked. *)
+    imports : Compilation_unit.Import.t list;
+    source_info : Source_info.t;
     shape_info :
       (Compat.shape * Paths.Identifier.SourceLocation.t Compat.shape_uid_map)
       option;
   }
 end =
-  Compilation_unit
+  Implementation
 
 module rec Page : sig
-  type child =
-    | Page_child of string
-    | Module_child of string
-    | Source_tree_child of string
-    | Asset_child of string
+  type child = Page_child of string | Module_child of string
 
   type t = {
     name : Identifier.Page.t;
     root : Root.t;
     content : Comment.docs;
     children : child list;
+    frontmatter : Frontmatter.t;
     digest : Digest.t;
     linked : bool;
   }
 end =
   Page
 
-module rec SourceTree : sig
-  type t = {
-    name : Identifier.Page.t;
-    root : Root.t;
-    source_children : Identifier.SourcePage.t list;
-    digest : Digest.t;
-  }
+module rec Asset : sig
+  type t = { name : Identifier.AssetFile.t; root : Root.t }
 end =
-  SourceTree
+  Asset
 
 let umty_of_mty : ModuleType.expr -> ModuleType.U.expr option = function
   | Signature sg -> Some (Signature sg)
   | Path { p_path; _ } -> Some (Path p_path)
   | Functor _ -> None
-  | TypeOf t -> Some (TypeOf t)
+  | TypeOf { t_desc; t_original_path; _ } ->
+      Some (TypeOf (t_desc, t_original_path))
   | With { w_substitutions; w_expr; _ } -> Some (With (w_substitutions, w_expr))
 
 (** Query the top-comment of a signature. This is [s.doc] most of the time with
     an exception for signature starting with an inline includes. *)
 let extract_signature_doc (s : Signature.t) =
+  let rec uexpr_considered_hidden = function
+    | ModuleType.U.Path p -> Path.is_hidden (p :> Path.t)
+    | Signature _ ->
+        true (* Hidden in some sense, we certainly want its top comment *)
+    | With (_, e) -> uexpr_considered_hidden e
+    | TypeOf (ModPath p, _) | TypeOf (StructInclude p, _) ->
+        Path.is_hidden (p :> Path.t)
+  in
+  let should_take_top = function
+    (* A signature that starts with an include may inherits the
+       top-comment from the expansion. *)
+    | { Include.status = `Inline; _ } -> true
+    | { decl = Alias p; _ } -> Paths.Path.is_hidden (p :> Path.t)
+    | { decl = ModuleType expr; _ } -> uexpr_considered_hidden expr
+  in
   match (s.doc, s.items) with
-  | [], Include { expansion; status = `Inline; _ } :: _ ->
-      (* A signature that starts with an [@inline] include inherits the
-         top-comment from the expansion. This comment is not rendered for
-         [include] items. *)
-      expansion.content.doc
+  | { elements = []; _ }, Include inc :: _ when should_take_top inc ->
+      inc.expansion.content.doc
   | doc, _ -> doc
