@@ -58,8 +58,8 @@ module Identifier : sig
 
     val functor_arg_pos : t -> int
     (** Gets the index in which the functor argument is, in the argument list.
-      Useful to turn identifiers into unique anchors, since multiple arguments
-      can have the same name. *)
+        Useful to turn identifiers into unique anchors, since multiple arguments
+        can have the same name. *)
   end
 
   module ModuleType :
@@ -67,9 +67,6 @@ module Identifier : sig
 
   module Type : IdSig with type t = Id.type_ and type t_pv = Id.type_pv
 
-  module SourceDir :
-    IdSig with type t = Id.source_dir and type t_pv = Id.source_dir_pv
-
   module Class : IdSig with type t = Id.class_ and type t_pv = Id.class_pv
 
   module ClassType :
@@ -142,6 +139,11 @@ module Identifier : sig
     type t_pv = Id.page_pv
   end
 
+  module LeafPage : sig
+    type t = Id.leaf_page
+    type t_pv = Id.leaf_page_pv
+  end
+
   module ContainerPage : sig
     type t = Id.container_page
     type t_pv = Id.container_page_pv
@@ -150,6 +152,8 @@ module Identifier : sig
   module NonSrc : sig
     type t = Id.non_src
     type t_pv = Id.non_src_pv
+    val hash : t -> int
+    val equal : ([< t_pv ] id as 'a) -> 'a -> bool
   end
 
   module SourcePage : sig
@@ -182,12 +186,6 @@ module Identifier : sig
     module Type :
       IdSig with type t = Id.path_type and type t_pv = Id.path_type_pv
 
-    module DataType :
-      IdSig with type t = Id.path_datatype and type t_pv = Id.path_datatype_pv
-
-    module Constructor :
-      IdSig with type t = Id.path_constructor and type t_pv = Id.constructor_pv
-
     module Value : IdSig with type t = Id.path_value and type t_pv = Id.value_pv
 
     module ClassType :
@@ -206,13 +204,11 @@ module Identifier : sig
 
   val name : [< t_pv ] id -> string
 
-  (* val root : [< t_pv ] id -> RootModule.t_pv id option *)
-
   val fullname : [< t_pv ] id -> string list
   (** The fullname of value [x] in module [M] is [M.x], whereas the regular name
       is [x]. *)
 
-  val is_internal : [< t_pv ] id -> bool
+  val is_hidden : [< t_pv ] id -> bool
 
   val compare : t -> t -> int
 
@@ -244,6 +240,14 @@ module Identifier : sig
     end
   end
 
+  module Hashtbl : sig
+    module Any : Hashtbl.S with type key = Any.t
+    module ContainerPage : Hashtbl.S with type key = ContainerPage.t
+    module LeafPage : Hashtbl.S with type key = LeafPage.t
+    module RootModule : Hashtbl.S with type key = RootModule.t
+    module SourcePage : Hashtbl.S with type key = SourcePage.t
+  end
+
   module Mk : sig
     open Names
 
@@ -255,14 +259,16 @@ module Identifier : sig
       ContainerPage.t option * PageName.t ->
       [> `LeafPage of ContainerPage.t option * PageName.t ] id
 
-    val source_page : ContainerPage.t * string list -> SourcePage.t
+    val source_page : ContainerPage.t * string -> SourcePage.t
 
-    val asset_file : Page.t * string -> AssetFile.t
+    val asset_file : Page.t * AssetName.t -> AssetFile.t
 
     val root :
       ContainerPage.t option * ModuleName.t ->
       [> `Root of ContainerPage.t option * ModuleName.t ] id
 
+    val implementation : string -> [> `Implementation of ModuleName.t ] id
+
     val module_ :
       Signature.t * ModuleName.t ->
       [> `Module of Signature.t * ModuleName.t ] id
@@ -278,11 +284,10 @@ module Identifier : sig
       [> `ModuleType of Signature.t * ModuleTypeName.t ] id
 
     val class_ :
-      Signature.t * ClassName.t -> [> `Class of Signature.t * ClassName.t ] id
+      Signature.t * TypeName.t -> [> `Class of Signature.t * TypeName.t ] id
 
     val class_type :
-      Signature.t * ClassTypeName.t ->
-      [> `ClassType of Signature.t * ClassTypeName.t ] id
+      Signature.t * TypeName.t -> [> `ClassType of Signature.t * TypeName.t ] id
 
     val type_ :
       Signature.t * TypeName.t -> [> `Type of Signature.t * TypeName.t ] id
@@ -304,17 +309,16 @@ module Identifier : sig
     val extension_decl :
       Signature.t * (ExtensionName.t * ExtensionName.t) ->
       [> `ExtensionDecl of Signature.t * ExtensionName.t * ExtensionName.t ] id
-    (** [extension_decl (sg, e1, eN)] defines an extension declaration where [sg] is the parent,
-        [e1] is the first constructor of the extension, and [eN] is the constructor the Id is created for.
-        [e1] will be used for the url, and [eN] will be the one displayed.
-        The first constructor of the extension will always be used to reference the extension point. *)
+    (** [extension_decl (sg, e1, eN)] defines an extension declaration where
+        [sg] is the parent, [e1] is the first constructor of the extension, and
+        [eN] is the constructor the Id is created for. [e1] will be used for the
+        url, and [eN] will be the one displayed. The first constructor of the
+        extension will always be used to reference the extension point. *)
 
     val exception_ :
       Signature.t * ExceptionName.t ->
       [> `Exception of Signature.t * ExceptionName.t ] id
 
-    val core_exception : string -> [> `CoreException of ExceptionName.t ] id
-
     val value :
       Signature.t * ValueName.t -> [> `Value of Signature.t * ValueName.t ] id
 
@@ -374,22 +378,8 @@ module rec Path : sig
       (* val identifier : t -> Identifier.Path.Type.t *)
     end
 
-    module DataType : sig
-      type t = Paths_types.Resolved_path.datatype
-    end
-
-    module Constructor : sig
-      type t = Paths_types.Resolved_path.constructor
-    end
-
     module Value : sig
       type t = Paths_types.Resolved_path.value
-
-      (* val of_ident : Identifier.Path.Value.t -> t *)
-
-      (* val is_hidden : t -> bool *)
-
-      (* val identifier : t -> Identifier.Path.Type.t *)
     end
 
     module ClassType : sig
@@ -402,7 +392,8 @@ module rec Path : sig
 
     type t = Paths_types.Resolved_path.any
 
-    val identifier : t -> Identifier.t
+    val identifier : t -> Identifier.t option
+    (** If the path points to a core type, no identifier can be generated *)
 
     val is_hidden : t -> bool
   end
@@ -421,14 +412,6 @@ module rec Path : sig
     type t = Paths_types.Path.type_
   end
 
-  module DataType : sig
-    type t = Paths_types.Path.datatype
-  end
-
-  module Constructor : sig
-    type t = Paths_types.Path.constructor
-  end
-
   module Value : sig
     type t = Paths_types.Path.value
   end
@@ -576,6 +559,12 @@ module rec Reference : sig
       type t = Paths_types.Resolved_reference.page
     end
 
+    module Asset : sig
+      type t = Paths_types.Resolved_reference.asset
+
+      val identifier : t -> Identifier.AssetFile.t
+    end
+
     type t = Paths_types.Resolved_reference.any
 
     val identifier : t -> Identifier.t
@@ -661,7 +650,16 @@ module rec Reference : sig
     type t = Paths_types.Reference.page
   end
 
+  module Asset : sig
+    type t = Paths_types.Reference.asset
+  end
+
+  module Hierarchy : sig
+    type t = Paths_types.Reference.hierarchy
+  end
+
   type t = Paths_types.Reference.any
 
   type tag_any = Paths_types.Reference.tag_any
+  type tag_hierarchy = Paths_types.Reference.tag_hierarchy
 end
