@@ -11,31 +11,30 @@ module Identifier = struct
   and container_page = container_page_pv id
   (** @canonical Odoc_model.Paths.Identifier.ContainerPage.t *)
 
-  type page_pv =
-    [ container_page_pv | `LeafPage of container_page option * PageName.t ]
+  type leaf_page_pv = [ `LeafPage of container_page option * PageName.t ]
+  (** @canonical Odoc_model.Paths.Identifier.LeafPage.t_pv *)
+
+  and leaf_page = leaf_page_pv id
+  (** @canonical Odoc_model.Paths.Identifier.LeafPage.t *)
+
+  type page_pv = [ container_page_pv | leaf_page_pv ]
   (** @canonical Odoc_model.Paths.Identifier.Page.t_pv *)
 
   and page = page_pv id
   (** @canonical Odoc_model.Paths.Identifier.Page.t *)
 
-  type source_dir_pv = [ container_page_pv | `SourceDir of source_dir * string ]
-  (** @canonical Odoc_model.Paths.Identifier.SourceDir.t_pv *)
-
-  and source_dir = source_dir_pv id
-  (** @canonical Odoc_model.Paths.Identifier.SourceDir.t *)
-
-  type source_page_pv = [ `SourcePage of source_dir * string ]
+  type source_page_pv = [ `SourcePage of container_page * string ]
   (** The second argument is the filename.
 
-     @canonical Odoc_model.Paths.Identifier.SourcePage.t_pv *)
+      @canonical Odoc_model.Paths.Identifier.SourcePage.t_pv *)
 
   type source_page = source_page_pv id
   (** @canonical Odoc_model.Paths.Identifier.SourcePage.t *)
 
-  type asset_file_pv = [ `AssetFile of page * string ]
+  type asset_file_pv = [ `AssetFile of page * AssetName.t ]
   (** The second argument is the filename.
 
-    @canonical Odoc_model.Paths.Identifier.AssetFile.t_pv *)
+      @canonical Odoc_model.Paths.Identifier.AssetFile.t_pv *)
 
   type asset_file = asset_file_pv id
   (** @canonical Odoc_model.Paths.Identifier.AssetFile.t *)
@@ -49,7 +48,12 @@ module Identifier = struct
   and source_location = source_location_pv id
   (** @canonical Odoc_model.Paths.Identifier.SourceLocation.t_pv *)
 
-  type odoc_id_pv = [ page_pv | `Root of container_page option * ModuleName.t ]
+  type odoc_id_pv =
+    [ page_pv
+    | source_page_pv
+    | asset_file_pv
+    | `Root of container_page option * ModuleName.t
+    | `Implementation of ModuleName.t ]
   (** @canonical Odoc_model.Paths.Identifier.OdocId.t_pv *)
 
   and odoc_id = odoc_id_pv id
@@ -67,15 +71,13 @@ module Identifier = struct
   (** @canonical Odoc_model.Paths.Identifier.Signature.t *)
 
   type class_signature_pv =
-    [ `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t ]
+    [ `Class of signature * TypeName.t | `ClassType of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Identifier.ClassSignature.t_pv *)
 
   and class_signature = class_signature_pv id
   (** @canonical Odoc_model.Paths.Identifier.ClassSignature.t *)
 
-  type datatype_pv =
-    [ `Type of signature * TypeName.t | `CoreType of TypeName.t ]
+  type datatype_pv = [ `Type of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Identifier.DataType.t_pv *)
 
   and datatype = datatype_pv id
@@ -129,7 +131,7 @@ module Identifier = struct
   and module_type = module_type_pv id
   (** @canonical Odoc_model.Paths.Identifier.ModuleType.t *)
 
-  type type_pv = [ `Type of signature * TypeName.t | `CoreType of TypeName.t ]
+  type type_pv = [ `Type of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Identifier.Type.t_pv *)
 
   and type_ = type_pv id
@@ -160,9 +162,7 @@ module Identifier = struct
   and extension_decl = extension_decl_pv id
   (** @canonical Odoc_model.Paths.Identifier.ExtensionDecl.t *)
 
-  type exception_pv =
-    [ `Exception of signature * ExceptionName.t
-    | `CoreException of ExceptionName.t ]
+  type exception_pv = [ `Exception of signature * ExceptionName.t ]
   (** @canonical Odoc_model.Paths.Identifier.Exception.t_pv *)
 
   and exception_ = exception_pv id
@@ -174,13 +174,13 @@ module Identifier = struct
   and value = value_pv id
   (** @canonical Odoc_model.Paths.Identifier.Value.t *)
 
-  type class_pv = [ `Class of signature * ClassName.t ]
+  type class_pv = [ `Class of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Identifier.Class.t_pv *)
 
   and class_ = class_pv id
   (** @canonical Odoc_model.Paths.Identifier.Class.t *)
 
-  type class_type_pv = [ `ClassType of signature * ClassTypeName.t ]
+  type class_type_pv = [ `ClassType of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Identifier.ClassType.t_pv *)
 
   and class_type = class_type_pv id
@@ -234,11 +234,7 @@ module Identifier = struct
   (** @canonical Odoc_model.Paths.Identifier.NonSrc.t *)
 
   type any_pv =
-    [ non_src_pv
-    | source_page_pv
-    | source_dir_pv
-    | source_location_pv
-    | asset_file_pv ]
+    [ non_src_pv | source_page_pv | source_location_pv | asset_file_pv ]
   (** @canonical Odoc_model.Paths.Identifier.t_pv *)
 
   and any = any_pv id
@@ -259,14 +255,6 @@ module Identifier = struct
   and path_type = path_type_pv id
   (** @canonical Odoc_model.Paths.Identifier.Path.Type.t *)
 
-  type path_datatype_pv = type_pv
-  (** @canonical Odoc_model.Paths.Identifier.Path.DataType.t_pv *)
-
-  and path_datatype = path_datatype_pv id
-  (** @canonical Odoc_model.Paths.Identifier.Path.DataType.t *)
-
-  type path_constructor = constructor
-
   type path_value = value
 
   type path_class_type_pv = [ class_pv | class_type_pv ]
@@ -324,49 +312,53 @@ module rec Path : sig
   type module_ =
     [ `Resolved of Resolved_path.module_
     | `Identifier of Identifier.path_module * bool
-    | `Root of string
+    | `Substituted of module_
+    | `Root of ModuleName.t
     | `Forward of string
-    | `Dot of module_ * string
+    | `Dot of module_ * ModuleName.t
     | `Apply of module_ * module_ ]
   (** @canonical Odoc_model.Paths.Path.Module.t *)
 
   type module_type =
     [ `Resolved of Resolved_path.module_type
+    | `SubstitutedMT of module_type
     | `Identifier of Identifier.path_module_type * bool
-    | `Dot of module_ * string ]
+    | `DotMT of module_ * ModuleTypeName.t ]
   (** @canonical Odoc_model.Paths.Path.ModuleType.t *)
 
   type type_ =
     [ `Resolved of Resolved_path.type_
+    | `SubstitutedT of type_
     | `Identifier of Identifier.path_type * bool
-    | `Dot of module_ * string ]
+    | `DotT of module_ * TypeName.t ]
   (** @canonical Odoc_model.Paths.Path.Type.t *)
 
-  type datatype =
-    [ `Resolved of Resolved_path.datatype
-    | `Identifier of Identifier.path_datatype * bool
-    | `Dot of module_ * string ]
-  (** @canonical Odoc_model.Paths.Path.DataType.t *)
-
-  type constructor =
-    [ `Resolved of Resolved_path.constructor | `Dot of datatype * string ]
-  (** @canonical Odoc_model.Paths.Path.Constructor.t *)
-
-  type value = [ `Resolved of Resolved_path.value | `Dot of module_ * string ]
+  type value =
+    [ `Resolved of Resolved_path.value
+    | `Identifier of Identifier.path_value * bool
+    | `DotV of module_ * ValueName.t ]
   (** @canonical Odoc_model.Paths.Path.Value.t *)
 
   type class_type =
     [ `Resolved of Resolved_path.class_type
+    | `SubstitutedCT of class_type
     | `Identifier of Identifier.path_class_type * bool
-    | `Dot of module_ * string ]
+    | `DotT of module_ * TypeName.t ]
   (** @canonical Odoc_model.Paths.Path.ClassType.t *)
 
   type any =
     [ `Resolved of Resolved_path.any
+    | `SubstitutedT of type_
+    | `SubstitutedMT of module_type
+    | `Substituted of module_
+    | `SubstitutedCT of class_type
     | `Identifier of Identifier.path_any * bool
-    | `Root of string
+    | `Root of ModuleName.t
     | `Forward of string
-    | `Dot of module_ * string
+    | `Dot of module_ * ModuleName.t
+    | `DotT of module_ * TypeName.t
+    | `DotMT of module_ * ModuleTypeName.t
+    | `DotV of module_ * ValueName.t
     | `Apply of module_ * module_ ]
   (** @canonical Odoc_model.Paths.Path.t *)
 end =
@@ -376,6 +368,7 @@ and Resolved_path : sig
   type module_ =
     [ `Identifier of Identifier.path_module
     | `Subst of module_type * module_
+    | `Substituted of module_
     | `Hidden of module_
     | `Module of module_ * ModuleName.t
     | `Canonical of module_ * Path.module_  (** [`Canonical (mod, canonical)] *)
@@ -387,39 +380,40 @@ and Resolved_path : sig
   and module_type =
     [ `Identifier of Identifier.path_module_type
     | `SubstT of module_type * module_type
+    | `SubstitutedMT of module_type
     | `CanonicalModuleType of module_type * Path.module_type
     | `AliasModuleType of module_type * module_type
     | `ModuleType of module_ * ModuleTypeName.t
     | `OpaqueModuleType of module_type ]
   (** @canonical Odoc_model.Paths.Path.Resolved.ModuleType.t *)
 
-  type type_ =
-    [ `Identifier of Identifier.path_type
-    | `CanonicalType of type_ * Path.type_
-    | `Type of module_ * TypeName.t
-    | `Class of module_ * ClassName.t
-    | `ClassType of module_ * ClassTypeName.t ]
-  (** @canonical Odoc_model.Paths.Path.Resolved.Type.t *)
-
-  type datatype =
-    [ `Identifier of Identifier.datatype
-    | `CanonicalDataType of datatype * Path.datatype
-    | `Type of module_ * TypeName.t ]
-  (** @canonical Odoc_model.Paths.Path.Resolved.DataType.t *)
-
-  type constructor = [ `Constructor of datatype * ConstructorName.t ]
-  (** @canonical Odoc_model.Paths.Path.Resolved.Constructor.t *)
-
-  type value = [ `Value of module_ * ValueName.t ]
+  type value =
+    [ `Identifier of Identifier.path_value | `Value of module_ * ValueName.t ]
   (** @canonical Odoc_model.Paths.Path.Resolved.Value.t *)
 
   type class_type =
     [ `Identifier of Identifier.path_class_type
-    | `Class of module_ * ClassName.t
-    | `ClassType of module_ * ClassTypeName.t ]
+    | `SubstitutedCT of class_type
+    | `Class of module_ * TypeName.t
+    | `ClassType of module_ * TypeName.t ]
+
+  type type_ =
+    [ `Identifier of Identifier.path_type
+    | `SubstitutedT of type_
+    | `SubstitutedCT of class_type
+    | `CanonicalType of type_ * Path.type_
+    | `Type of module_ * TypeName.t
+    | `Class of module_ * TypeName.t
+    | `ClassType of module_ * TypeName.t
+    | `CoreType of TypeName.t ]
+  (** @canonical Odoc_model.Paths.Path.Resolved.Type.t *)
 
   type any =
     [ `Identifier of Identifier.any
+    | `SubstitutedCT of class_type
+    | `SubstitutedT of type_
+    | `SubstitutedMT of module_type
+    | `Substituted of module_
     | `Subst of module_type * module_
     | `Hidden of module_
     | `Module of module_ * ModuleName.t
@@ -433,14 +427,13 @@ and Resolved_path : sig
     | `SubstT of module_type * module_type
     | `OpaqueModuleType of module_type
     | `CanonicalType of type_ * Path.type_
-    | `CanonicalDataType of datatype * Path.datatype
     | `Type of module_ * TypeName.t
-    | `Constructor of datatype * ConstructorName.t
-    | `Class of module_ * ClassName.t
-    | `ClassType of module_ * ClassTypeName.t
-    | `Class of module_ * ClassName.t
+    | `Class of module_ * TypeName.t
+    | `ClassType of module_ * TypeName.t
+    | `Class of module_ * TypeName.t
     | `Value of module_ * ValueName.t
-    | `ClassType of module_ * ClassTypeName.t ]
+    | `ClassType of module_ * TypeName.t
+    | `CoreType of TypeName.t ]
   (** @canonical Odoc_model.Paths.Path.Resolved.t *)
 end =
   Resolved_path
@@ -497,8 +490,8 @@ and Resolved_fragment : sig
 
   type type_ =
     [ `Type of signature * TypeName.t
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t ]
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Fragment.Resolved.Type.t *)
 
   and module_type = [ `Module_type of signature * ModuleTypeName.t ]
@@ -515,8 +508,8 @@ and Resolved_fragment : sig
     | `Module of signature * ModuleName.t
     | `Module_type of signature * ModuleTypeName.t
     | `Type of signature * TypeName.t
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t
     | `OpaqueModule of module_ ]
   (** @canonical Odoc_model.Paths.Fragment.Resolved.t *)
 end =
@@ -557,6 +550,12 @@ module rec Reference : sig
 
   type tag_only_child_module = [ `TChildModule ]
 
+  type tag_hierarchy =
+    [ `TRelativePath  (** [{!identifier/}] *)
+    | `TAbsolutePath  (** [{!/identifier}] *)
+    | `TCurrentPackage  (** [{!//identifier}] *) ]
+  (** @canonical Odoc_model.Paths.Reference.tag_hierarchy *)
+
   type tag_any =
     [ `TModule
     | `TModuleType
@@ -573,9 +572,11 @@ module rec Reference : sig
     | `TInstanceVariable
     | `TLabel
     | `TPage
+    | `TAsset
     | `TChildPage
     | `TChildModule
     | `TUnknown ]
+  (** @canonical Odoc_model.Paths.Reference.tag_any *)
 
   type tag_signature = [ `TUnknown | `TModule | `TModuleType ]
 
@@ -596,10 +597,14 @@ module rec Reference : sig
     | `TChildPage
     | `TChildModule ]
 
+  type hierarchy = tag_hierarchy * string list
+  (** @canonical Odoc_model.Paths.Reference.Hierarchy.t *)
+
   type signature =
     [ `Resolved of Resolved_reference.signature
     | `Root of string * tag_signature
     | `Dot of label_parent * string
+    | `Module_path of hierarchy
     | `Module of signature * ModuleName.t
     | `ModuleType of signature * ModuleTypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.Signature.t *)
@@ -608,8 +613,8 @@ module rec Reference : sig
     [ `Resolved of Resolved_reference.class_signature
     | `Root of string * tag_class_signature
     | `Dot of label_parent * string
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t ]
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.ClassSignature.t *)
 
   and datatype =
@@ -624,6 +629,7 @@ module rec Reference : sig
     [ `Resolved of Resolved_reference.field_parent
     | `Root of string * tag_parent
     | `Dot of label_parent * string
+    | `Module_path of hierarchy
     | `Module of signature * ModuleName.t
     | `ModuleType of signature * ModuleTypeName.t
     | `Type of signature * TypeName.t ]
@@ -633,17 +639,24 @@ module rec Reference : sig
     [ `Resolved of Resolved_reference.label_parent
     | `Root of string * tag_label_parent
     | `Dot of label_parent * string
+    | `Page_path of hierarchy
+    | `Module_path of hierarchy
+    | `Any_path of hierarchy
     | `Module of signature * ModuleName.t
     | `ModuleType of signature * ModuleTypeName.t
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t
     | `Type of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.LabelParent.t *)
 
+  type asset =
+    [ `Resolved of Resolved_reference.asset | `Asset_path of hierarchy ]
+
   type module_ =
     [ `Resolved of Resolved_reference.module_
     | `Root of string * [ `TModule | `TUnknown ]
     | `Dot of label_parent * string
+    | `Module_path of hierarchy
     | `Module of signature * ModuleName.t ]
   (** @canonical Odoc_model.Paths.Reference.Module.t *)
 
@@ -658,8 +671,8 @@ module rec Reference : sig
     [ `Resolved of Resolved_reference.type_
     | `Root of string * [ `TType | `TClass | `TClassType | `TUnknown ]
     | `Dot of label_parent * string
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t
     | `Type of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.Type.t *)
 
@@ -712,15 +725,15 @@ module rec Reference : sig
     [ `Resolved of Resolved_reference.class_
     | `Root of string * [ `TClass | `TUnknown ]
     | `Dot of label_parent * string
-    | `Class of signature * ClassName.t ]
+    | `Class of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.Class.t *)
 
   type class_type =
     [ `Resolved of Resolved_reference.class_type
     | `Root of string * [ `TClass | `TClassType | `TUnknown ]
     | `Dot of label_parent * string
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t ]
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.ClassType.t *)
 
   type method_ =
@@ -747,13 +760,17 @@ module rec Reference : sig
   type page =
     [ `Resolved of Resolved_reference.page
     | `Root of string * [ `TPage | `TUnknown ]
-    | `Dot of label_parent * string ]
+    | `Page_path of hierarchy ]
   (** @canonical Odoc_model.Paths.Reference.Page.t *)
 
   type any =
     [ `Resolved of Resolved_reference.any
     | `Root of string * tag_any
     | `Dot of label_parent * string
+    | `Page_path of hierarchy
+    | `Module_path of hierarchy
+    | `Asset_path of hierarchy
+    | `Any_path of hierarchy
     | `Module of signature * ModuleName.t
     | `ModuleType of signature * ModuleTypeName.t
     | `Type of signature * TypeName.t
@@ -763,8 +780,8 @@ module rec Reference : sig
     | `ExtensionDecl of signature * ExtensionName.t
     | `Exception of signature * ExceptionName.t
     | `Value of signature * ValueName.t
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t
     | `Method of class_signature * MethodName.t
     | `InstanceVariable of class_signature * InstanceVariableName.t
     | `Label of label_parent * LabelName.t ]
@@ -802,8 +819,8 @@ and Resolved_reference : sig
 
   and class_signature =
     [ `Identifier of Identifier.class_signature
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t ]
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.Resolved.ClassSignature.t *)
 
   (* fragment_type_parent in resolved references is for record fields parent.
@@ -828,8 +845,8 @@ and Resolved_reference : sig
     | `Module of signature * ModuleName.t
     | `Hidden of module_
     | `ModuleType of signature * ModuleTypeName.t
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t
     | `Type of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.Resolved.LabelParent.t *)
 
@@ -842,8 +859,8 @@ and Resolved_reference : sig
   type type_ =
     [ `Identifier of Identifier.reference_type
     | `Type of signature * TypeName.t
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t ]
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.Resolved.Type.t *)
 
   type constructor =
@@ -886,13 +903,13 @@ and Resolved_reference : sig
 
   type class_ =
     [ `Identifier of Identifier.reference_class
-    | `Class of signature * ClassName.t ]
+    | `Class of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.Resolved.Class.t *)
 
   type class_type =
     [ `Identifier of Identifier.reference_class_type
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t ]
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t ]
   (** @canonical Odoc_model.Paths.Reference.Resolved.ClassType.t *)
 
   type method_ =
@@ -913,6 +930,9 @@ and Resolved_reference : sig
   type page = [ `Identifier of Identifier.reference_page ]
   (** @canonical Odoc_model.Paths.Reference.Resolved.Page.t *)
 
+  type asset = [ `Identifier of Identifier.asset_file ]
+  (** @canonical Odoc_model.Paths.Reference.Resolved.Asset.t *)
+
   type any =
     [ `Identifier of Identifier.any
     | `Alias of Resolved_path.module_ * module_
@@ -922,13 +942,14 @@ and Resolved_reference : sig
     | `ModuleType of signature * ModuleTypeName.t
     | `Type of signature * TypeName.t
     | `Constructor of datatype * ConstructorName.t
+    | `PolyConstructor of datatype * ConstructorName.t
     | `Field of field_parent * FieldName.t
     | `Extension of signature * ExtensionName.t
     | `ExtensionDecl of signature * ExtensionName.t * ExtensionName.t
     | `Exception of signature * ExceptionName.t
     | `Value of signature * ValueName.t
-    | `Class of signature * ClassName.t
-    | `ClassType of signature * ClassTypeName.t
+    | `Class of signature * TypeName.t
+    | `ClassType of signature * TypeName.t
     | `Method of class_signature * MethodName.t
     | `InstanceVariable of class_signature * InstanceVariableName.t
     | `Label of label_parent * LabelName.t ]
