@@ -1,18 +1,22 @@
+open Odoc_utils
+
 module Location = Location_
 module Ast = Odoc_parser.Ast
 
 type internal_tags_removed =
   [ `Tag of Ast.ocamldoc_tag
   | `Heading of Ast.heading
+  | `Media of
+    Ast.reference_kind * Ast.media_href Ast.with_location * string * Ast.media
   | Ast.nestable_block_element ]
 (** {!Ast.block_element} without internal tags. *)
 
 type _ handle_internal_tags =
-  | Expect_status
-      : [ `Default | `Inline | `Open | `Closed ] handle_internal_tags
-  | Expect_canonical
-      : [ `Dot of Paths.Path.Module.t * string ] option handle_internal_tags
+  | Expect_status :
+      [ `Default | `Inline | `Open | `Closed ] handle_internal_tags
+  | Expect_canonical : Reference.path option handle_internal_tags
   | Expect_none : unit handle_internal_tags
+  | Expect_page_tags : Frontmatter.t handle_internal_tags
 
 let describe_internal_tag = function
   | `Canonical _ -> "@canonical"
@@ -20,6 +24,10 @@ let describe_internal_tag = function
   | `Open -> "@open"
   | `Closed -> "@closed"
   | `Hidden -> "@hidden"
+  | `Children_order _ -> "@children_order"
+  | `Toc_status _ -> "@toc_status"
+  | `Short_title _ -> "@short_title"
+  | `Order_category _ -> "@order_category"
 
 let warn_unexpected_tag { Location.value; location } =
   Error.raise_warning
@@ -31,35 +39,23 @@ let warn_root_canonical location =
   Error.raise_warning
   @@ Error.make "Canonical paths must contain a dot, eg. X.Y." location
 
-let rec find_tag f = function
+let rec find_tag ~filter = function
   | [] -> None
   | hd :: tl -> (
-      match f hd.Location.value with
+      match filter hd.Location.value with
       | Some x -> Some (x, hd.location)
       | None ->
           warn_unexpected_tag hd;
-          find_tag f tl)
+          find_tag ~filter tl)
 
-let handle_internal_tags (type a) tags : a handle_internal_tags -> a = function
-  | Expect_status -> (
-      match
-        find_tag
-          (function (`Inline | `Open | `Closed) as t -> Some t | _ -> None)
-          tags
-      with
-      | Some (status, _) -> status
-      | None -> `Default)
-  | Expect_canonical -> (
-      match find_tag (function `Canonical p -> Some p | _ -> None) tags with
-      | Some (`Root _, location) ->
-          warn_root_canonical location;
-          None
-      | Some ((`Dot _ as p), _) -> Some p
-      | None -> None)
-  | Expect_none ->
-      (* Will raise warnings. *)
-      ignore (find_tag (fun _ -> None) tags);
-      ()
+let rec find_tags acc ~filter = function
+  | [] -> List.rev acc
+  | hd :: tl -> (
+      match filter hd.Location.value with
+      | Some x -> find_tags ((x, hd.location) :: acc) ~filter tl
+      | None ->
+          warn_unexpected_tag hd;
+          find_tags acc ~filter tl)
 
 (* Errors *)
 let invalid_raw_markup_target : string -> Location.span -> Error.t =
@@ -70,12 +66,6 @@ let default_raw_markup_target_not_supported : Location.span -> Error.t =
   Error.make ~suggestion:"try '{%html:...%}'."
     "'{%%...%%}' (raw markup) needs a target language."
 
-let headings_not_allowed : Location.span -> Error.t =
-  Error.make "Headings not allowed in this comment."
-
-let titles_not_allowed : Location.span -> Error.t =
-  Error.make "Title-level headings {0 ...} are only allowed in pages."
-
 let bad_heading_level : int -> Location.span -> Error.t =
   Error.make "'%d': bad heading level (0-5 allowed)."
 
@@ -106,9 +96,10 @@ let not_allowed :
 let describe_element = function
   | `Reference (`Simple, _, _) -> "'{!...}' (cross-reference)"
   | `Reference (`With_text, _, _) -> "'{{!...} ...}' (cross-reference)"
-  | `Link _ -> "'{{:...} ...}' (external link)"
+  | `Link (_, _) -> "'{{:...} ...}' (external link)"
   | `Heading (level, _, _) ->
       Printf.sprintf "'{%i ...}' (section heading)" level
+  | `Specific s -> s
 
 (* End of errors *)
 
@@ -127,7 +118,6 @@ type alerts =
   [ `Tag of [ `Alert of string * string option ] ] Location_.with_location list
 
 type status = {
-  sections_allowed : sections_allowed;
   tags_allowed : bool;
   parent_of_sections : Paths.Identifier.LabelParent.t;
 }
@@ -143,8 +133,9 @@ let leaf_inline_element :
       match target with
       | Some invalid_target
         when String.trim invalid_target = ""
-             || String.contains invalid_target '%'
-             || String.contains invalid_target '}' ->
+             || String.exists
+                  (function '%' | '}' -> true | _ -> false)
+                  invalid_target ->
           Error.raise_warning
             (invalid_raw_markup_target invalid_target location);
 
@@ -160,20 +151,20 @@ type surrounding =
   | `Reference of
     [ `Simple | `With_text ]
     * string Location_.with_location
-    * Odoc_parser.Ast.inline_element Location_.with_location list ]
+    * Odoc_parser.Ast.inline_element Location_.with_location list
+  | `Specific of string ]
 
 let rec non_link_inline_element :
-    status ->
     surrounding:surrounding ->
     Odoc_parser.Ast.inline_element with_location ->
     Comment.non_link_inline_element with_location =
- fun status ~surrounding element ->
+ fun ~surrounding element ->
   match element with
   | { value = #ast_leaf_inline_element; _ } as element ->
       (leaf_inline_element element
         :> Comment.non_link_inline_element with_location)
   | { value = `Styled (style, content); _ } ->
-      `Styled (style, non_link_inline_elements status ~surrounding content)
+      `Styled (style, non_link_inline_elements ~surrounding content)
       |> Location.same element
   | ( { value = `Reference (_, _, content); _ }
     | { value = `Link (_, content); _ } ) as element ->
@@ -183,29 +174,26 @@ let rec non_link_inline_element :
         element.location
       |> Error.raise_warning;
 
-      `Styled (`Emphasis, non_link_inline_elements status ~surrounding content)
+      `Styled (`Emphasis, non_link_inline_elements ~surrounding content)
       |> Location.same element
 
-and non_link_inline_elements status ~surrounding elements =
-  List.map (non_link_inline_element status ~surrounding) elements
+and non_link_inline_elements ~surrounding elements =
+  List.map (non_link_inline_element ~surrounding) elements
 
 let rec inline_element :
-    status ->
     Odoc_parser.Ast.inline_element with_location ->
     Comment.inline_element with_location =
- fun status element ->
+ fun element ->
   match element with
   | { value = #ast_leaf_inline_element; _ } as element ->
       (leaf_inline_element element :> Comment.inline_element with_location)
   | { value = `Styled (style, content); location } ->
-      `Styled (style, inline_elements status content) |> Location.at location
+      `Styled (style, inline_elements content) |> Location.at location
   | { value = `Reference (kind, target, content) as value; location } -> (
       let { Location.value = target; location = target_location } = target in
       match Error.raise_warnings (Reference.parse target_location target) with
       | Result.Ok target ->
-          let content =
-            non_link_inline_elements status ~surrounding:value content
-          in
+          let content = non_link_inline_elements ~surrounding:value content in
           Location.at location (`Reference (target, content))
       | Result.Error error ->
           Error.raise_warning error;
@@ -214,21 +202,20 @@ let rec inline_element :
             | `Simple -> `Code_span target
             | `With_text -> `Styled (`Emphasis, content)
           in
-          inline_element status (Location.at location placeholder))
+          inline_element (Location.at location placeholder))
   | { value = `Link (target, content) as value; location } ->
-      `Link (target, non_link_inline_elements status ~surrounding:value content)
+      `Link (target, non_link_inline_elements ~surrounding:value content)
       |> Location.at location
 
-and inline_elements status elements = List.map (inline_element status) elements
+and inline_elements elements = List.map inline_element elements
 
 let rec nestable_block_element :
-    status ->
     Odoc_parser.Ast.nestable_block_element with_location ->
     Comment.nestable_block_element with_location =
- fun status element ->
+ fun element ->
   match element with
   | { value = `Paragraph content; location } ->
-      Location.at location (`Paragraph (inline_elements status content))
+      Location.at location (`Paragraph (inline_elements content))
   | { value = `Code_block { meta; delimiter = _; content; output }; location }
     ->
       let lang_tag =
@@ -239,11 +226,21 @@ let rec nestable_block_element :
       let outputs =
         match output with
         | None -> None
-        | Some l -> Some (List.map (nestable_block_element status) l)
+        | Some l -> Some (List.map nestable_block_element l)
       in
+      let trimmed_content, warnings =
+        Odoc_parser.codeblock_content location content.value
+      in
+      let warnings = List.map Error.t_of_parser_t warnings in
+      List.iter (Error.raise_warning ~non_fatal:true) warnings;
+      let content = Location.at content.location trimmed_content in
       Location.at location (`Code_block (lang_tag, content, outputs))
   | { value = `Math_block s; location } -> Location.at location (`Math_block s)
-  | { value = `Verbatim _; _ } as element -> element
+  | { value = `Verbatim v; location } ->
+      let v, warnings = Odoc_parser.codeblock_content location v in
+      let warnings = List.map Error.t_of_parser_t warnings in
+      List.iter (Error.raise_warning ~non_fatal:true) warnings;
+      Location.at location (`Verbatim v)
   | { value = `Modules modules; location } ->
       let modules =
         List.fold_left
@@ -261,19 +258,41 @@ let rec nestable_block_element :
       in
       Location.at location (`Modules modules)
   | { value = `List (kind, _syntax, items); location } ->
-      `List (kind, List.map (nestable_block_elements status) items)
+      `List (kind, List.map nestable_block_elements items)
       |> Location.at location
   | { value = `Table ((grid, align), (`Heavy | `Light)); location } ->
       let data =
         List.map
           (List.map (fun (cell, cell_type) ->
-               (nestable_block_elements status cell, cell_type)))
+               (nestable_block_elements cell, cell_type)))
           grid
       in
       `Table { Comment.data; align } |> Location.at location
+  | { value = `Media (_, { value = `Link href; _ }, content, m); location } ->
+      `Media (`Link href, m, content) |> Location.at location
+  | {
+   value =
+     `Media
+       (kind, { value = `Reference href; location = href_location }, content, m);
+   location;
+  } -> (
+      let fallback error =
+        Error.raise_warning error;
+        let placeholder =
+          match kind with
+          | `Simple -> `Code_span href
+          | `With_text ->
+              `Styled (`Emphasis, [ `Word content |> Location.at location ])
+        in
+        `Paragraph (inline_elements [ placeholder |> Location.at location ])
+        |> Location.at location
+      in
+      match Error.raise_warnings (Reference.parse_asset href_location href) with
+      | Result.Ok target ->
+          `Media (`Reference target, m, content) |> Location.at location
+      | Result.Error error -> fallback error)
 
-and nestable_block_elements status elements =
-  List.map (nestable_block_element status) elements
+and nestable_block_elements elements = List.map nestable_block_element elements
 
 let tag :
     location:Location.span ->
@@ -290,26 +309,23 @@ let tag :
   let ok t = Result.Ok (Location.at location (`Tag t)) in
   match tag with
   | (`Author _ | `Since _ | `Version _) as tag -> ok tag
-  | `Deprecated content ->
-      ok (`Deprecated (nestable_block_elements status content))
+  | `Deprecated content -> ok (`Deprecated (nestable_block_elements content))
   | `Param (name, content) ->
-      ok (`Param (name, nestable_block_elements status content))
+      ok (`Param (name, nestable_block_elements content))
   | `Raise (name, content) -> (
       match Error.raise_warnings (Reference.parse location name) with
       (* TODO: location for just name *)
       | Result.Ok target ->
-          ok
-            (`Raise
-              (`Reference (target, []), nestable_block_elements status content))
+          ok (`Raise (`Reference (target, []), nestable_block_elements content))
       | Result.Error error ->
           Error.raise_warning error;
           let placeholder = `Code_span name in
-          ok (`Raise (placeholder, nestable_block_elements status content)))
-  | `Return content -> ok (`Return (nestable_block_elements status content))
+          ok (`Raise (placeholder, nestable_block_elements content)))
+  | `Return content -> ok (`Return (nestable_block_elements content))
   | `See (kind, target, content) ->
-      ok (`See (kind, target, nestable_block_elements status content))
+      ok (`See (kind, target, nestable_block_elements content))
   | `Before (version, content) ->
-      ok (`Before (version, nestable_block_elements status content))
+      ok (`Before (version, nestable_block_elements content))
 
 (* When the user does not give a section heading a label (anchor), we generate
    one from the text in the heading. This is the common case. This involves
@@ -355,12 +371,7 @@ let generate_heading_label : Comment.inline_element with_location list -> string
               anchor
           | `Styled (_, content) ->
               content |> strip_locs |> scan_inline_elements anchor
-          | `Reference (_, content) ->
-              content |> strip_locs
-              |> List.map (fun (ele : Comment.non_link_inline_element) ->
-                     (ele :> Comment.inline_element))
-              |> scan_inline_elements anchor
-          | `Link (_, content) ->
+          | `Reference (_, content) | `Link (_, content) ->
               content |> strip_locs
               |> List.map (fun (ele : Comment.non_link_inline_element) ->
                      (ele :> Comment.inline_element))
@@ -379,7 +390,7 @@ let section_heading :
  fun status ~top_heading_level location heading ->
   let (`Heading (level, label, content)) = heading in
 
-  let text = inline_elements status content in
+  let text = inline_elements content in
 
   let heading_label_explicit, label =
     match label with
@@ -399,42 +410,27 @@ let section_heading :
     in
     (top_heading_level, element)
   in
-
-  match (status.sections_allowed, level) with
-  | `None, _any_level ->
-      Error.raise_warning (headings_not_allowed location);
-      let text = (text :> Comment.inline_element with_location list) in
-      let element =
-        Location.at location
-          (`Paragraph [ Location.at location (`Styled (`Bold, text)) ])
-      in
-      (top_heading_level, element)
-  | `No_titles, 0 ->
-      Error.raise_warning (titles_not_allowed location);
-      mk_heading `Title
-  | _, level ->
-      let level' =
-        match level with
-        | 0 -> `Title
-        | 1 -> `Section
-        | 2 -> `Subsection
-        | 3 -> `Subsubsection
-        | 4 -> `Paragraph
-        | 5 -> `Subparagraph
-        | _ ->
-            Error.raise_warning (bad_heading_level level location);
-            (* Implicitly promote to level-5. *)
-            `Subparagraph
-      in
-      (match top_heading_level with
-      | Some top_level
-        when status.sections_allowed = `All && level <= top_level && level <= 5
-        ->
-          Error.raise_warning
-            (heading_level_should_be_lower_than_top_level level top_level
-               location)
-      | _ -> ());
-      mk_heading level'
+  let level' =
+    match level with
+    | 0 -> `Title
+    | 1 -> `Section
+    | 2 -> `Subsection
+    | 3 -> `Subsubsection
+    | 4 -> `Paragraph
+    | 5 -> `Subparagraph
+    | _ ->
+        Error.raise_warning (bad_heading_level level location);
+        (* Implicitly promote to level-5. *)
+        `Subparagraph
+  in
+  let () =
+    match top_heading_level with
+    | Some top_level when level <= top_level && level <= 5 ->
+        Error.raise_warning
+          (heading_level_should_be_lower_than_top_level level top_level location)
+    | _ -> ()
+  in
+  mk_heading level'
 
 let validate_first_page_heading status ast_element =
   match status.parent_of_sections.iv with
@@ -457,12 +453,12 @@ let top_level_block_elements status ast_elements =
     | [] -> List.rev comment_elements_acc
     | ast_element :: ast_elements -> (
         (* The first [ast_element] in pages must be a title or section heading. *)
-        if status.sections_allowed = `All && top_heading_level = None then
+        if top_heading_level = None then
           validate_first_page_heading status ast_element;
 
         match ast_element with
         | { value = #Odoc_parser.Ast.nestable_block_element; _ } as element ->
-            let element = nestable_block_element status element in
+            let element = nestable_block_element element in
             let element = (element :> Comment.block_element with_location) in
             traverse ~top_heading_level
               (element :: comment_elements_acc)
@@ -494,12 +490,22 @@ let top_level_block_elements status ast_elements =
   traverse ~top_heading_level [] ast_elements
 
 let strip_internal_tags ast : internal_tags_removed with_location list * _ =
-  let rec loop tags ast' = function
+  let rec loop ~start tags ast' = function
     | ({ Location.value = `Tag (#Ast.internal_tag as tag); _ } as wloc) :: tl
       -> (
-        let next tag = loop ({ wloc with value = tag } :: tags) ast' tl in
+        let next tag =
+          loop ~start ({ wloc with value = tag } :: tags) ast' tl
+        in
         match tag with
         | (`Inline | `Open | `Closed | `Hidden) as tag -> next tag
+        | ( `Children_order _ | `Short_title _ | `Toc_status _
+          | `Order_category _ ) as tag ->
+            let tag_name = describe_internal_tag tag in
+            if not start then
+              Error.raise_warning
+                (Error.make "%s tag has to be before any content" tag_name
+                   wloc.location);
+            next tag
         | `Canonical { Location.value = s; location = r_location } -> (
             match
               Error.raise_warnings (Reference.read_path_longident r_location s)
@@ -507,19 +513,22 @@ let strip_internal_tags ast : internal_tags_removed with_location list * _ =
             | Result.Ok path -> next (`Canonical path)
             | Result.Error e ->
                 Error.raise_warning e;
-                loop tags ast' tl))
+                loop ~start tags ast' tl))
     | ({
          value =
-           `Tag #Ast.ocamldoc_tag | `Heading _ | #Ast.nestable_block_element;
+           ( `Tag #Ast.ocamldoc_tag
+           | `Heading _ | `Media _
+           | #Ast.nestable_block_element );
          _;
        } as hd)
       :: tl ->
-        loop tags (hd :: ast') tl
+        loop ~start:false tags (hd :: ast') tl
     | [] -> (List.rev ast', List.rev tags)
   in
-  loop [] [] ast
+  loop ~start:true [] [] ast
 
-(** Append alerts at the end of the comment. Tags are favoured in case of alerts of the same name. *)
+(** Append alerts at the end of the comment. Tags are favoured in case of alerts
+    of the same name. *)
 let append_alerts_to_comment alerts
     (comment : Comment.block_element with_location list) =
   let alerts =
@@ -534,25 +543,79 @@ let append_alerts_to_comment alerts
           comment)
       alerts
   in
-  comment @ (alerts : alerts :> Comment.docs)
+  comment @ (alerts :> Comment.elements)
 
-let ast_to_comment ~internal_tags ~sections_allowed ~tags_allowed
-    ~parent_of_sections ast alerts =
+let handle_internal_tags (type a) tags : a handle_internal_tags -> a = function
+  | Expect_status -> (
+      match
+        find_tag
+          ~filter:(function
+            | (`Inline | `Open | `Closed) as t -> Some t | _ -> None)
+          tags
+      with
+      | Some (status, _) -> status
+      | None -> `Default)
+  | Expect_canonical -> (
+      match
+        find_tag ~filter:(function `Canonical p -> Some p | _ -> None) tags
+      with
+      | Some (`Root _, location) ->
+          warn_root_canonical location;
+          None
+      | Some ((`Dot _ as p), _) -> Some p
+      | None -> None)
+  | Expect_page_tags ->
+      let unparsed_lines =
+        find_tags []
+          ~filter:(function
+            | ( `Children_order _ | `Toc_status _ | `Short_title _
+              | `Order_category _ ) as p ->
+                Some p
+            | _ -> None)
+          tags
+      in
+      let lines =
+        let do_ parse loc els =
+          let els = nestable_block_elements els in
+          match parse loc els with
+          | Ok res -> Some res
+          | Error e ->
+              Error.raise_warning e;
+              None
+        in
+        List.filter_map
+          (function
+            | `Children_order co, loc ->
+                do_ Frontmatter.parse_children_order loc co
+            | `Toc_status co, loc -> do_ Frontmatter.parse_toc_status loc co
+            | `Short_title t, loc -> do_ Frontmatter.parse_short_title loc t
+            | `Order_category t, loc ->
+                do_ Frontmatter.parse_order_category loc t)
+          unparsed_lines
+      in
+      Frontmatter.of_lines lines |> Error.raise_warnings
+  | Expect_none ->
+      (* Will raise warnings. *)
+      ignore (find_tag ~filter:(fun _ -> None) tags);
+      ()
+
+let ast_to_comment ~internal_tags ~tags_allowed ~parent_of_sections
+    (ast : Ast.t) alerts =
   Error.catch_warnings (fun () ->
-      let status = { sections_allowed; tags_allowed; parent_of_sections } in
+      let status = { tags_allowed; parent_of_sections } in
       let ast, tags = strip_internal_tags ast in
       let elts =
         top_level_block_elements status ast |> append_alerts_to_comment alerts
       in
       (elts, handle_internal_tags tags internal_tags))
 
-let parse_comment ~internal_tags ~sections_allowed ~tags_allowed
-    ~containing_definition ~location ~text =
+let parse_comment ~internal_tags ~tags_allowed ~containing_definition ~location
+    ~text =
   Error.catch_warnings (fun () ->
       let ast =
         Odoc_parser.parse_comment ~location ~text |> Error.raise_parser_warnings
       in
-      ast_to_comment ~internal_tags ~sections_allowed ~tags_allowed
+      ast_to_comment ~internal_tags ~tags_allowed
         ~parent_of_sections:containing_definition ast []
       |> Error.raise_warnings)
 
@@ -566,3 +629,11 @@ let parse_reference text =
       }
   in
   Reference.parse location text
+
+let non_link_inline_element :
+    context:string ->
+    Odoc_parser.Ast.inline_element with_location list ->
+    Comment.non_link_inline_element with_location list =
+ fun ~context elements ->
+  let surrounding = `Specific context in
+  non_link_inline_elements ~surrounding elements
