@@ -1,5 +1,6 @@
 open Type_desc
 open Odoc_model
+open Odoc_utils
 open Paths_desc
 open Comment_desc
 module T = Type_desc
@@ -16,10 +17,6 @@ let inline_status =
     | `Closed -> C0 "`Closed"
     | `Inline -> C0 "`Inline")
 
-let source_info =
-  let open Lang.Source_info in
-  Record [ F ("id", (fun t -> t.id), identifier) ]
-
 (** {3 Module} *)
 
 let rec module_decl =
@@ -38,7 +35,7 @@ and module_t =
   Record
     [
       F ("id", (fun t -> t.id), identifier);
-      F ("locs", (fun t -> t.locs), Option identifier);
+      F ("source_loc", (fun t -> t.source_loc), Option identifier);
       F ("doc", (fun t -> t.doc), docs);
       F ("type_", (fun t -> t.type_), module_decl);
       F
@@ -61,7 +58,8 @@ and functorparameter_t =
   let open Lang.FunctorParameter in
   Variant
     (function
-    | Unit -> C0 "Unit" | Named x -> C ("Named", x, functorparameter_parameter))
+    | Unit -> C0 "Unit"
+    | Named x -> C ("Named", x, functorparameter_parameter))
 
 (** {3 ModuleType} *)
 
@@ -140,6 +138,10 @@ and moduletype_typeof_t =
   Record
     [
       F ("t_desc", (fun t -> t.t_desc), moduletype_type_of_desc);
+      F
+        ( "t_original_path",
+          (fun t -> (t.t_original_path :> Odoc_model.Paths.Path.t)),
+          path );
       F ("t_expansion", (fun t -> t.t_expansion), Option simple_expansion);
     ]
 
@@ -165,14 +167,18 @@ and moduletype_u_expr =
           ( "With",
             (t, e),
             Pair (List moduletype_substitution, moduletype_u_expr) )
-    | TypeOf x -> C ("TypeOf", x, moduletype_typeof_t))
+    | TypeOf (t, o) ->
+        C
+          ( "TypeOf",
+            (t, (o :> Paths.Path.t)),
+            Pair (moduletype_type_of_desc, path) ))
 
 and moduletype_t =
   let open Lang.ModuleType in
   Record
     [
       F ("id", (fun t -> t.id), identifier);
-      F ("locs", (fun t -> t.locs), Option identifier);
+      F ("source_loc", (fun t -> t.source_loc), Option identifier);
       F ("doc", (fun t -> t.doc), docs);
       F
         ( "canonical",
@@ -257,20 +263,13 @@ and include_shadowed =
   let open Lang.Include in
   Record
     [
-      F ("s_modules", (fun t -> t.s_modules), List string);
-      F ("s_module_types", (fun t -> t.s_module_types), List string);
-      F ("s_values", (fun t -> t.s_values), List string);
-      F ("s_types", (fun t -> t.s_types), List string);
-      F ("s_classes", (fun t -> t.s_classes), List string);
-      F ("s_class_types", (fun t -> t.s_class_types), List string);
-    ]
-
-and include_expansion =
-  let open Lang.Include in
-  Record
-    [
-      F ("shadowed", (fun t -> t.shadowed), include_shadowed);
-      F ("content", (fun t -> t.content), signature_t);
+      (* TODO: fixme *)
+      F ("s_modules", (fun t -> List.map fst t.s_modules), List string);
+      F ("s_module_types", (fun t -> List.map fst t.s_module_types), List string);
+      F ("s_values", (fun t -> List.map fst t.s_values), List string);
+      F ("s_types", (fun t -> List.map fst t.s_types), List string);
+      F ("s_classes", (fun t -> List.map fst t.s_classes), List string);
+      F ("s_class_types", (fun t -> List.map fst t.s_class_types), List string);
     ]
 
 and include_decl =
@@ -280,6 +279,14 @@ and include_decl =
     | Alias p -> C ("Alias", (p :> Paths.Path.t), path)
     | ModuleType e -> C ("ModuleType", e, moduletype_u_expr))
 
+and include_expansion =
+  let open Lang.Include in
+  Record
+    [
+      F ("shadowed", (fun t -> t.shadowed), include_shadowed);
+      F ("content", (fun t -> t.content), signature_t);
+    ]
+
 and include_t =
   let open Lang.Include in
   Record
@@ -362,7 +369,7 @@ and typedecl_t =
   Record
     [
       F ("id", (fun t -> t.id), identifier);
-      F ("locs", (fun t -> t.locs), Option identifier);
+      F ("source_loc", (fun t -> t.source_loc), Option identifier);
       F ("doc", (fun t -> t.doc), docs);
       F ("equation", (fun t -> t.equation), typedecl_equation);
       F
@@ -377,7 +384,7 @@ and extension_constructor =
   Record
     [
       F ("id", (fun t -> t.id), identifier);
-      F ("locs", (fun t -> t.locs), Option identifier);
+      F ("source_loc", (fun t -> t.source_loc), Option identifier);
       F ("doc", (fun t -> t.doc), docs);
       F ("args", (fun t -> t.args), typedecl_constructor_argument);
       F ("res", (fun t -> t.res), Option typeexpr_t);
@@ -401,7 +408,7 @@ and exception_t =
   Record
     [
       F ("id", (fun t -> t.id), identifier);
-      F ("locs", (fun t -> t.locs), Option identifier);
+      F ("source_loc", (fun t -> t.source_loc), Option identifier);
       F ("doc", (fun t -> t.doc), docs);
       F ("args", (fun t -> t.args), typedecl_constructor_argument);
       F ("res", (fun t -> t.res), Option typeexpr_t);
@@ -414,12 +421,13 @@ and value_t =
   let value_value_t =
     Variant
       (function
-      | Abstract -> C0 "Abstract" | External x -> C ("External", x, List string))
+      | Abstract -> C0 "Abstract"
+      | External x -> C ("External", x, List string))
   in
   Record
     [
       F ("id", (fun t -> t.id), identifier);
-      F ("locs", (fun t -> t.locs), Option identifier);
+      F ("source_loc", (fun t -> t.source_loc), Option identifier);
       F ("doc", (fun t -> t.doc), docs);
       F ("type_", (fun t -> t.type_), typeexpr_t);
       F ("value", (fun t -> t.value), value_value_t);
@@ -443,7 +451,7 @@ and class_t =
   Record
     [
       F ("id", (fun t -> t.id), identifier);
-      F ("locs", (fun t -> t.locs), Option identifier);
+      F ("source_loc", (fun t -> t.source_loc), Option identifier);
       F ("doc", (fun t -> t.doc), docs);
       F ("virtual_", (fun t -> t.virtual_), bool);
       F ("params", (fun t -> t.params), List typedecl_param);
@@ -466,7 +474,7 @@ and classtype_t =
   Record
     [
       F ("id", (fun t -> t.id), identifier);
-      F ("locs", (fun t -> t.locs), Option identifier);
+      F ("source_loc", (fun t -> t.source_loc), Option identifier);
       F ("doc", (fun t -> t.doc), docs);
       F ("virtual_", (fun t -> t.virtual_), bool);
       F ("params", (fun t -> t.params), List typedecl_param);
@@ -604,7 +612,9 @@ and typeexpr_label =
   let open Lang.TypeExpr in
   Variant
     (function
-    | Label x -> C ("Label", x, string) | Optional x -> C ("Optional", x, string))
+    | Label x -> C ("Label", x, string)
+    | RawOptional x -> C ("RawOptional", x, string)
+    | Optional x -> C ("Optional", x, string))
 
 and typeexpr_t =
   let open Lang.TypeExpr in
@@ -682,7 +692,6 @@ and compilation_unit_t =
         ( "canonical",
           (fun t -> (t.canonical :> Paths.Path.t option)),
           Option path );
-      F ("sources", (fun t -> t.source_info), Option source_info);
     ]
 
 (** {3 Page} *)
@@ -693,16 +702,57 @@ and page_t =
     [
       F ("name", (fun t -> t.name), identifier);
       F ("root", (fun t -> t.root), root);
+      F ("frontmatter", (fun t -> t.frontmatter), frontmatter);
       F ("content", (fun t -> t.content), docs);
       F ("digest", (fun t -> t.digest), Digest.t);
     ]
 
-and source_tree_page_t =
-  let open Lang.SourceTree in
+and frontmatter =
+  let open Odoc_model.Frontmatter in
+  let ignore_loc x = x.Location_.value in
+  Record
+    [
+      F
+        ( "children",
+          (fun t -> Option.map ignore_loc t.children_order),
+          Option (List child) );
+      F
+        ( "short_title",
+          (fun t ->
+            (t.short_title
+              :> Comment.inline_element Location_.with_location list option)),
+          Option Comment_desc.inline_element );
+      F
+        ( "toc_status",
+          (fun t ->
+            Option.map
+              (function `Hidden -> "hidden" | `Open -> "open")
+              t.toc_status),
+          Option string );
+      F ("order_category", (fun t -> t.order_category), Option string);
+    ]
+
+and child =
+  let open Odoc_model.Frontmatter in
+  Variant
+    (function
+    | { Location_.value = Page s; _ } -> C ("Page", s, string)
+    | { Location_.value = Dir s; _ } -> C ("Dir", s, string)
+    | { Location_.value = Module s; _ } -> C ("Module", s, string))
+
+and implementation_t =
+  let open Lang.Implementation in
+  Record
+    [
+      F ("id", (fun t -> t.id), Option identifier);
+      F ("digest", (fun t -> t.digest), Digest.t);
+      F ("root", (fun t -> t.root), root);
+    ]
+
+and asset_t =
+  let open Lang.Asset in
   Record
     [
       F ("name", (fun t -> t.name), identifier);
       F ("root", (fun t -> t.root), root);
-      F ("digest", (fun t -> t.digest), Digest.t);
-      F ("source_children", (fun t -> t.source_children), List identifier);
     ]
