@@ -3,10 +3,13 @@ open Result
 let resolve url_to_string directories reference =
   let resolver =
     Resolver.create ~important_digests:false ~directories ~open_modules:[]
+      ~roots:None
   in
   let reference =
     let open Odoc_model in
-    let warnings_options = { Error.warn_error = true; print_warnings = true } in
+    let warnings_options =
+      { Error.warn_error = true; print_warnings = true; warnings_tag = None }
+    in
     Semantics.parse_reference reference
     |> Error.handle_errors_and_warnings ~warnings_options
   in
@@ -25,19 +28,16 @@ let resolve url_to_string directories reference =
               Odoc_xref2.Errors.Tools_error.pp_reference_lookup_error e
           in
           Error (`Msg error)
-      | Ok resolved_reference -> (
+      | Ok (resolved_reference, _) ->
           let identifier =
             Odoc_model.Paths.Reference.Resolved.identifier resolved_reference
           in
           let url =
             Odoc_document.Url.from_identifier ~stop_before:false identifier
           in
-          match url with
-          | Error e -> Error (`Msg (Odoc_document.Url.Error.to_string e))
-          | Ok url ->
-              let href = url_to_string url in
-              print_endline href;
-              Ok ()))
+          let href = url_to_string url in
+          print_endline href;
+          Ok ())
 
 let reference_to_url_html { Html_page.html_config = config; _ } root_url =
   let url_to_string url =
