@@ -325,7 +325,8 @@ and delimited_inline_element_list :
     requires_leading_whitespace:bool ->
     input ->
     Ast.inline_element with_location list * Loc.span =
- fun ~parent_markup ~parent_markup_location ~requires_leading_whitespace input ->
+ fun ~parent_markup ~parent_markup_location ~requires_leading_whitespace
+     input ->
   (* [~at_start_of_line] is used to interpret [`Minus] and [`Plus]. These are
      word tokens if not the first non-whitespace tokens on their line. Then,
      they are allowed in a non-link element list. *)
@@ -533,7 +534,7 @@ type where_in_line =
 
    When it is called inside a shorthand list item ([- foo]), it stops on end of
    input, right brace, a blank line (indicating end of shorthand list), plus or
-   minus (indicating the start of the next liste item), or a section heading or
+   minus (indicating the start of the next list item), or a section heading or
    tag, which cannot be nested in list markup.
 
    The block parser [block_element_list] explicitly returns the token that
@@ -554,6 +555,7 @@ type stopped_implicitly =
   | `Minus
   | `Plus
   | Token.section_heading
+  | Token.media_markup
   | Token.tag ]
 
 (* Ensure that the above two types are really subsets of [Token.t]. *)
@@ -575,27 +577,26 @@ let _check_subset : stopped_implicitly -> Token.t = fun t -> (t :> Token.t)
    - The type of token that the block parser stops at. See discussion above. *)
 type ('block, 'stops_at_which_tokens) context =
   | Top_level : (Ast.block_element, stops_at_delimiters) context
-  | In_shorthand_list : (Ast.nestable_block_element, stopped_implicitly) context
+  | In_implicitly_ended :
+      [ `Tag | `Shorthand_list ]
+      -> (Ast.nestable_block_element, stopped_implicitly) context
   | In_explicit_list : (Ast.nestable_block_element, stops_at_delimiters) context
   | In_table_cell : (Ast.nestable_block_element, stops_at_delimiters) context
   | In_code_results : (Ast.nestable_block_element, code_stop) context
-  | In_tag : (Ast.nestable_block_element, Token.t) context
 
 (* This is a no-op. It is needed to prove to the type system that nestable block
    elements are acceptable block elements in all contexts. *)
-let accepted_in_all_contexts :
-    type block stops_at_which_tokens.
+let accepted_in_all_contexts : type block stops_at_which_tokens.
     (block, stops_at_which_tokens) context ->
     Ast.nestable_block_element ->
     block =
  fun context block ->
   match context with
   | Top_level -> (block :> Ast.block_element)
-  | In_shorthand_list -> block
+  | In_implicitly_ended (`Tag | `Shorthand_list) -> block
   | In_explicit_list -> block
   | In_table_cell -> block
   | In_code_results -> block
-  | In_tag -> block
 
 (* Converts a tag to a series of words. This is used in error recovery, when a
    tag cannot be generated. *)
@@ -616,6 +617,10 @@ let tag_to_words = function
   | `See (`Url, s) -> [ `Word "@see"; `Space " "; `Word ("<" ^ s ^ ">") ]
   | `Since s -> [ `Word "@since"; `Space " "; `Word s ]
   | `Version s -> [ `Word "@version"; `Space " "; `Word s ]
+  | `Children_order -> [ `Word "@children_order" ]
+  | `Toc_status -> [ `Word "@toc_status" ]
+  | `Order_category -> [ `Word "@order_category" ]
+  | `Short_title -> [ `Word "@short_title" ]
 
 (* {3 Block element lists} *)
 
@@ -627,8 +632,7 @@ let tag_to_words = function
    - tables,
    - lists, and
    - section headings. *)
-let rec block_element_list :
-    type block stops_at_which_tokens.
+let rec block_element_list : type block stops_at_which_tokens.
     (block, stops_at_which_tokens) context ->
     parent_markup:[< Token.t | `Comment ] ->
     input ->
@@ -637,13 +641,12 @@ let rec block_element_list :
     * where_in_line =
  fun context ~parent_markup input ->
   let rec consume_block_elements :
-      parsed_a_tag:bool ->
       where_in_line ->
       block with_location list ->
       block with_location list
       * stops_at_which_tokens with_location
       * where_in_line =
-   fun ~parsed_a_tag where_in_line acc ->
+   fun where_in_line acc ->
     let describe token =
       match token with
       | #token_that_always_begins_an_inline_element -> "paragraph"
@@ -656,16 +659,6 @@ let rec block_element_list :
         |> add_warning input
     in
 
-    let warn_if_after_tags { Loc.location; value = token } =
-      if parsed_a_tag then
-        let suggestion =
-          Printf.sprintf "move %s before any tags." (Token.describe token)
-        in
-        Parse_error.not_allowed ~what:(describe token)
-          ~in_what:"the tags section" ~suggestion location
-        |> add_warning input
-    in
-
     let warn_because_not_at_top_level { Loc.location; value = token } =
       let suggestion =
         Printf.sprintf "move %s outside of any other markup."
@@ -682,9 +675,9 @@ let rec block_element_list :
     | { value = `End; _ } as next_token -> (
         match context with
         | Top_level -> (List.rev acc, next_token, where_in_line)
-        | In_shorthand_list -> (List.rev acc, next_token, where_in_line)
+        | In_implicitly_ended (`Tag | `Shorthand_list) ->
+            (List.rev acc, next_token, where_in_line)
         | In_explicit_list -> (List.rev acc, next_token, where_in_line)
-        | In_tag -> (List.rev acc, next_token, where_in_line)
         | In_table_cell -> (List.rev acc, next_token, where_in_line)
         | In_code_results -> (List.rev acc, next_token, where_in_line))
     | { value = `Right_brace; _ } as next_token -> (
@@ -693,37 +686,39 @@ let rec block_element_list :
            possible values of [context]. *)
         match context with
         | Top_level -> (List.rev acc, next_token, where_in_line)
-        | In_shorthand_list -> (List.rev acc, next_token, where_in_line)
+        | In_implicitly_ended (`Tag | `Shorthand_list) ->
+            (List.rev acc, next_token, where_in_line)
         | In_explicit_list -> (List.rev acc, next_token, where_in_line)
         | In_table_cell -> (List.rev acc, next_token, where_in_line)
-        | In_tag -> (List.rev acc, next_token, where_in_line)
         | In_code_results ->
             junk input;
-            consume_block_elements ~parsed_a_tag where_in_line acc)
+            consume_block_elements where_in_line acc)
     | { value = `Right_code_delimiter; _ } as next_token -> (
         match context with
         | In_code_results -> (List.rev acc, next_token, where_in_line)
         | _ ->
             junk input;
-            consume_block_elements ~parsed_a_tag where_in_line acc)
+            consume_block_elements where_in_line acc)
     (* Whitespace. This can terminate some kinds of block elements. It is also
        necessary to track it to interpret [`Minus] and [`Plus] correctly, as
        well as to ensure that all block elements begin on their own line. *)
     | { value = `Space _; _ } ->
         junk input;
-        consume_block_elements ~parsed_a_tag where_in_line acc
+        consume_block_elements where_in_line acc
     | { value = `Single_newline _; _ } ->
         junk input;
-        consume_block_elements ~parsed_a_tag `At_start_of_line acc
+        consume_block_elements `At_start_of_line acc
     | { value = `Blank_line _; _ } as next_token -> (
         match context with
-        (* Blank lines terminate shorthand lists ([- foo]). They also terminate
-           paragraphs, but the paragraph parser is aware of that internally. *)
-        | In_shorthand_list -> (List.rev acc, next_token, where_in_line)
+        (* Blank lines terminate shorthand lists ([- foo]) and tags. They also
+           terminate paragraphs, but the paragraph parser is aware of that
+           internally. *)
+        | In_implicitly_ended (`Tag | `Shorthand_list) ->
+            (List.rev acc, next_token, where_in_line)
         (* Otherwise, blank lines are pretty much like single newlines. *)
         | _ ->
             junk input;
-            consume_block_elements ~parsed_a_tag `At_start_of_line acc)
+            consume_block_elements `At_start_of_line acc)
     (* Explicit list items ([{li ...}] and [{- ...}]) can never appear directly
        in block content. They can only appear inside [{ul ...}] and [{ol ...}].
        So, catch those. *)
@@ -739,7 +734,7 @@ let rec block_element_list :
         |> add_warning input;
 
         junk input;
-        consume_block_elements ~parsed_a_tag where_in_line acc
+        consume_block_elements where_in_line acc
     (* Table rows ([{tr ...}]) can never appear directly
        in block content. They can only appear inside [{table ...}]. *)
     | { value = `Begin_table_row as token; location } ->
@@ -752,7 +747,7 @@ let rec block_element_list :
           ~suggestion location
         |> add_warning input;
         junk input;
-        consume_block_elements ~parsed_a_tag where_in_line acc
+        consume_block_elements where_in_line acc
     (* Table cells ([{th ...}] and [{td ...}]) can never appear directly
        in block content. They can only appear inside [{tr ...}]. *)
     | { value = `Begin_table_cell _ as token; location } ->
@@ -765,9 +760,8 @@ let rec block_element_list :
           ~suggestion location
         |> add_warning input;
         junk input;
-        consume_block_elements ~parsed_a_tag where_in_line acc
-    (* Tags. These can appear at the top level only. Also, once one tag is seen,
-       the only top-level elements allowed are more tags. *)
+        consume_block_elements where_in_line acc
+    (* Tags. These can appear at the top level only. *)
     | { value = `Tag tag as token; location } as next_token -> (
         let recover_when_not_at_top_level context =
           warn_because_not_at_top_level next_token;
@@ -778,8 +772,7 @@ let rec block_element_list :
             |> accepted_in_all_contexts context
             |> Loc.at location
           in
-          consume_block_elements ~parsed_a_tag `At_start_of_line
-            (paragraph :: acc)
+          consume_block_elements `At_start_of_line (paragraph :: acc)
         in
 
         match context with
@@ -788,15 +781,11 @@ let rec block_element_list :
         (* If a tag starts at the beginning of a line, it terminates the preceding
            tag and/or the current shorthand list. In this case, return to the
            caller, and let the caller decide how to interpret the tag token. *)
-        | In_shorthand_list ->
+        | In_implicitly_ended (`Tag | `Shorthand_list) ->
             if where_in_line = `At_start_of_line then
               (List.rev acc, next_token, where_in_line)
             else recover_when_not_at_top_level context
         | In_table_cell -> recover_when_not_at_top_level context
-        | In_tag ->
-            if where_in_line = `At_start_of_line then
-              (List.rev acc, next_token, where_in_line)
-            else recover_when_not_at_top_level context
         | In_code_results -> recover_when_not_at_top_level context
         (* If this is the top-level call to [block_element_list], parse the
            tag. *)
@@ -830,26 +819,31 @@ let rec block_element_list :
                 in
 
                 let tag = Loc.at location (`Tag tag) in
-                consume_block_elements ~parsed_a_tag:true `After_text
-                  (tag :: acc)
-            | (`Deprecated | `Return) as tag ->
+                consume_block_elements `After_text (tag :: acc)
+            | ( `Deprecated | `Return | `Children_order | `Short_title
+              | `Toc_status | `Order_category ) as tag ->
                 let content, _stream_head, where_in_line =
-                  block_element_list In_tag ~parent_markup:token input
+                  block_element_list (In_implicitly_ended `Tag)
+                    ~parent_markup:token input
                 in
                 let tag =
                   match tag with
                   | `Deprecated -> `Deprecated content
+                  | `Toc_status -> `Toc_status content
                   | `Return -> `Return content
+                  | `Children_order -> `Children_order content
+                  | `Short_title -> `Short_title content
+                  | `Order_category -> `Order_category content
                 in
                 let location =
                   location :: List.map Loc.location content |> Loc.span
                 in
                 let tag = Loc.at location (`Tag tag) in
-                consume_block_elements ~parsed_a_tag:true where_in_line
-                  (tag :: acc)
+                consume_block_elements where_in_line (tag :: acc)
             | (`Param _ | `Raise _ | `Before _) as tag ->
                 let content, _stream_head, where_in_line =
-                  block_element_list In_tag ~parent_markup:token input
+                  block_element_list (In_implicitly_ended `Tag)
+                    ~parent_markup:token input
                 in
                 let tag =
                   match tag with
@@ -861,34 +855,30 @@ let rec block_element_list :
                   location :: List.map Loc.location content |> Loc.span
                 in
                 let tag = Loc.at location (`Tag tag) in
-                consume_block_elements ~parsed_a_tag:true where_in_line
-                  (tag :: acc)
+                consume_block_elements where_in_line (tag :: acc)
             | `See (kind, target) ->
                 let content, _next_token, where_in_line =
-                  block_element_list In_tag ~parent_markup:token input
+                  block_element_list (In_implicitly_ended `Tag)
+                    ~parent_markup:token input
                 in
                 let location =
                   location :: List.map Loc.location content |> Loc.span
                 in
                 let tag = `Tag (`See (kind, target, content)) in
                 let tag = Loc.at location tag in
-                consume_block_elements ~parsed_a_tag:true where_in_line
-                  (tag :: acc)
+                consume_block_elements where_in_line (tag :: acc)
             | (`Inline | `Open | `Closed | `Hidden) as tag ->
                 let tag = Loc.at location (`Tag tag) in
-                consume_block_elements ~parsed_a_tag:true `After_text
-                  (tag :: acc)))
+                consume_block_elements `After_text (tag :: acc)))
     | ( { value = #token_that_always_begins_an_inline_element; _ }
       | { value = `Bar; _ } ) as next_token ->
-        warn_if_after_tags next_token;
         warn_if_after_text next_token;
 
         let block = paragraph input in
         let block = Loc.map (accepted_in_all_contexts context) block in
         let acc = block :: acc in
-        consume_block_elements ~parsed_a_tag `After_text acc
+        consume_block_elements `After_text acc
     | { value = `Verbatim s as token; location } as next_token ->
-        warn_if_after_tags next_token;
         warn_if_after_text next_token;
         if s = "" then
           Parse_error.should_not_be_empty ~what:(Token.describe token) location
@@ -898,9 +888,8 @@ let rec block_element_list :
         let block = accepted_in_all_contexts context token in
         let block = Loc.at location block in
         let acc = block :: acc in
-        consume_block_elements ~parsed_a_tag `After_text acc
+        consume_block_elements `After_text acc
     | { value = `Math_block s as token; location } as next_token ->
-        warn_if_after_tags next_token;
         warn_if_after_text next_token;
         if s = "" then
           Parse_error.should_not_be_empty ~what:(Token.describe token) location
@@ -910,14 +899,13 @@ let rec block_element_list :
         let block = accepted_in_all_contexts context token in
         let block = Loc.at location block in
         let acc = block :: acc in
-        consume_block_elements ~parsed_a_tag `After_text acc
+        consume_block_elements `After_text acc
     | {
         value =
           `Code_block (meta, delim, { value = s; location = v_loc }, has_outputs)
           as token;
         location;
       } as next_token ->
-        warn_if_after_tags next_token;
         warn_if_after_text next_token;
         junk input;
         let delimiter = if delim = "" then None else Some delim in
@@ -948,18 +936,17 @@ let rec block_element_list :
         let block =
           accepted_in_all_contexts context
             (`Code_block
-              {
-                Ast.meta;
-                delimiter;
-                content = { value = s; location = v_loc };
-                output;
-              })
+               {
+                 Ast.meta;
+                 delimiter;
+                 content = { value = s; location = v_loc };
+                 output;
+               })
         in
         let block = Loc.at location block in
         let acc = block :: acc in
-        consume_block_elements ~parsed_a_tag `After_text acc
+        consume_block_elements `After_text acc
     | { value = `Modules s as token; location } as next_token ->
-        warn_if_after_tags next_token;
         warn_if_after_text next_token;
 
         junk input;
@@ -998,9 +985,8 @@ let rec block_element_list :
         let block = accepted_in_all_contexts context (`Modules modules) in
         let block = Loc.at location block in
         let acc = block :: acc in
-        consume_block_elements ~parsed_a_tag `After_text acc
+        consume_block_elements `After_text acc
     | { value = `Begin_list kind as token; location } as next_token ->
-        warn_if_after_tags next_token;
         warn_if_after_text next_token;
 
         junk input;
@@ -1017,10 +1003,9 @@ let rec block_element_list :
         let block = accepted_in_all_contexts context block in
         let block = Loc.at location block in
         let acc = block :: acc in
-        consume_block_elements ~parsed_a_tag `After_text acc
+        consume_block_elements `After_text acc
     | { value = (`Begin_table_light | `Begin_table_heavy) as token; location }
       as next_token ->
-        warn_if_after_tags next_token;
         warn_if_after_text next_token;
         junk input;
         let block, brace_location =
@@ -1036,7 +1021,7 @@ let rec block_element_list :
         let block = accepted_in_all_contexts context (`Table block) in
         let block = Loc.at location block in
         let acc = block :: acc in
-        consume_block_elements ~parsed_a_tag `After_text acc
+        consume_block_elements `After_text acc
     | { value = (`Minus | `Plus) as token; location } as next_token -> (
         (match where_in_line with
         | `After_text | `After_shorthand_bullet ->
@@ -1045,10 +1030,9 @@ let rec block_element_list :
             |> add_warning input
         | _ -> ());
 
-        warn_if_after_tags next_token;
-
         match context with
-        | In_shorthand_list -> (List.rev acc, next_token, where_in_line)
+        | In_implicitly_ended `Shorthand_list ->
+            (List.rev acc, next_token, where_in_line)
         | _ ->
             let items, where_in_line =
               shorthand_list_items next_token where_in_line input
@@ -1063,11 +1047,9 @@ let rec block_element_list :
             let block = accepted_in_all_contexts context block in
             let block = Loc.at location block in
             let acc = block :: acc in
-            consume_block_elements ~parsed_a_tag where_in_line acc)
+            consume_block_elements where_in_line acc)
     | { value = `Begin_section_heading (level, label) as token; location } as
       next_token -> (
-        warn_if_after_tags next_token;
-
         let recover_when_not_at_top_level context =
           warn_because_not_at_top_level next_token;
           junk input;
@@ -1082,18 +1064,16 @@ let rec block_element_list :
             |> accepted_in_all_contexts context
             |> Loc.at location
           in
-          consume_block_elements ~parsed_a_tag `At_start_of_line
-            (paragraph :: acc)
+          consume_block_elements `At_start_of_line (paragraph :: acc)
         in
 
         match context with
-        | In_shorthand_list ->
+        | In_implicitly_ended (`Tag | `Shorthand_list) ->
             if where_in_line = `At_start_of_line then
               (List.rev acc, next_token, where_in_line)
             else recover_when_not_at_top_level context
         | In_explicit_list -> recover_when_not_at_top_level context
         | In_table_cell -> recover_when_not_at_top_level context
-        | In_tag -> recover_when_not_at_top_level context
         | In_code_results -> recover_when_not_at_top_level context
         | Top_level ->
             if where_in_line <> `At_start_of_line then
@@ -1126,7 +1106,7 @@ let rec block_element_list :
             let heading = `Heading (level, label, content) in
             let heading = Loc.at location heading in
             let acc = heading :: acc in
-            consume_block_elements ~parsed_a_tag `After_text acc)
+            consume_block_elements `After_text acc)
     | { value = `Begin_paragraph_style _ as token; location } ->
         junk input;
         let content, brace_location =
@@ -1145,20 +1125,68 @@ let rec block_element_list :
           |> accepted_in_all_contexts context
           |> Loc.at location
         in
-        consume_block_elements ~parsed_a_tag `At_start_of_line (paragraph :: acc)
+        consume_block_elements `At_start_of_line (paragraph :: acc)
+    | {
+     location;
+     value = `Media_with_replacement_text (href, media, content) as token;
+    } ->
+        junk input;
+
+        let r_location =
+          Loc.nudge_start
+            (String.length @@ Token.s_of_media `Replaced media)
+            location
+          |> Loc.nudge_end (String.length content + 1)
+          (* +1 for closing character *)
+        in
+        let c_location =
+          Loc.nudge_start
+            (String.length (Token.s_of_media `Replaced media)
+            + String.length (match href with `Reference s | `Link s -> s))
+            location
+          |> Loc.nudge_end 1
+        in
+        let content = String.trim content in
+        let href = href |> Loc.at r_location in
+
+        if content = "" then
+          Parse_error.should_not_be_empty ~what:(Token.describe token)
+            c_location
+          |> add_warning input;
+
+        let block = `Media (`Simple, href, content, media) in
+        let block = accepted_in_all_contexts context block in
+        let block = Loc.at location block in
+        let acc = block :: acc in
+        consume_block_elements `After_text acc
+    | { location; value = `Simple_media (href, media) } ->
+        junk input;
+
+        let r_location =
+          Loc.nudge_start
+            (String.length @@ Token.s_of_media `Simple media)
+            location
+          |> Loc.nudge_end 1
+        in
+        let href = href |> Loc.at r_location in
+        let block = `Media (`Simple, href, "", media) in
+        let block = accepted_in_all_contexts context block in
+        let block = Loc.at location block in
+        let acc = block :: acc in
+        consume_block_elements `After_text acc
   in
 
   let where_in_line =
     match context with
     | Top_level -> `At_start_of_line
-    | In_shorthand_list -> `After_shorthand_bullet
+    | In_implicitly_ended `Shorthand_list -> `After_shorthand_bullet
     | In_explicit_list -> `After_explicit_list_bullet
     | In_table_cell -> `After_table_cell
     | In_code_results -> `After_tag
-    | In_tag -> `After_tag
+    | In_implicitly_ended `Tag -> `After_tag
   in
 
-  consume_block_elements ~parsed_a_tag:false where_in_line []
+  consume_block_elements where_in_line []
 
 (* {3 Lists} *)
 
@@ -1187,14 +1215,16 @@ and shorthand_list_items :
       Ast.nestable_block_element with_location list list * where_in_line =
    fun next_token where_in_line acc ->
     match next_token.value with
-    | `End | `Right_brace | `Blank_line _ | `Tag _ | `Begin_section_heading _ ->
+    | `End | `Right_brace | `Blank_line _ | `Tag _ | `Begin_section_heading _
+    | `Simple_media _ | `Media_with_replacement_text _ ->
         (List.rev acc, where_in_line)
     | (`Minus | `Plus) as bullet ->
         if bullet = bullet_token then (
           junk input;
 
           let content, stream_head, where_in_line =
-            block_element_list In_shorthand_list ~parent_markup:bullet input
+            block_element_list (In_implicitly_ended `Shorthand_list)
+              ~parent_markup:bullet input
           in
           if content = [] then
             Parse_error.should_not_be_empty ~what:(Token.describe bullet)
