@@ -21,26 +21,15 @@ let type_path : Env.t -> Paths.Path.Type.t -> Paths.Path.Type.t =
       | Ok p' -> `Resolved Lang_of.(Path.resolved_type (empty ()) p')
       | Error _ -> p)
 
-(* and value_path : Env.t -> Paths.Path.Value.t -> Paths.Path.Value.t = *)
-(*  fun env p -> *)
-(*   match p with *)
-(*   | `Resolved _ -> p *)
-(*   | _ -> ( *)
-(*       let cp = Component.Of_Lang.(value_path (empty ()) p) in *)
-(*       match Tools.resolve_value_path env cp with *)
-(*       | Ok p' -> `Resolved Lang_of.(Path.resolved_value (empty ()) p') *)
-(*       | Error _ -> p) *)
-
-(* and constructor_path : *)
-(*     Env.t -> Paths.Path.Constructor.t -> Paths.Path.Constructor.t = *)
-(*  fun env p -> *)
-(*   match p with *)
-(*   | `Resolved _ -> p *)
-(*   | _ -> ( *)
-(*       let cp = Component.Of_Lang.(constructor_path (empty ()) p) in *)
-(*       match Tools.resolve_constructor_path env cp with *)
-(*       | Ok p' -> `Resolved Lang_of.(Path.resolved_constructor (empty ()) p') *)
-(*       | Error _ -> p) *)
+and value_path : Env.t -> Paths.Path.Value.t -> Paths.Path.Value.t =
+ fun env p ->
+  match p with
+  | `Resolved _ -> p
+  | _ -> (
+      let cp = Component.Of_Lang.(value_path (empty ()) p) in
+      match Tools.resolve_value_path env cp with
+      | Ok p' -> `Resolved Lang_of.(Path.resolved_value (empty ()) p')
+      | Error _ -> p)
 
 and module_type_path :
     Env.t -> Paths.Path.ModuleType.t -> Paths.Path.ModuleType.t =
@@ -76,22 +65,38 @@ and class_type_path : Env.t -> Paths.Path.ClassType.t -> Paths.Path.ClassType.t
 
 let rec unit env t =
   let open Compilation_unit in
-  let source_info =
-    match t.source_info with
-    | Some si -> Some (source_info env si)
-    | None -> None
+  { t with content = content env t.id t.content }
+
+and implementation env sp =
+  let open Implementation in
+  { sp with source_info = source_info_infos env sp.source_info }
+
+and source_info_infos env infos =
+  let open Source_info in
+  let map_doc f v =
+    let documentation =
+      match v.documentation with Some p -> Some (f p) | None -> None
+    in
+    { v with documentation }
   in
-  { t with content = content env t.id t.content; source_info }
-
-and source_info env si = { si with infos = source_info_infos env si.infos }
-
-and source_info_infos _env infos = infos
+  List.map
+    (function
+      | v, pos ->
+          let v =
+            match v with
+            | Value v -> Value (map_doc (value_path env) v)
+            | Module v -> Module (map_doc (module_path env) v)
+            | ModuleType v -> ModuleType (map_doc (module_type_path env) v)
+            | Type v -> Type (map_doc (type_path env) v)
+            | Definition _ as d -> d
+          in
+          (v, pos))
+    infos
 
 and content env id =
   let open Compilation_unit in
   function
-  | Module m ->
-      let sg = Type_of.signature env m in
+  | Module sg ->
       let sg = signature env (id :> Id.Signature.t) sg in
       Module sg
   | Pack p -> Pack p
@@ -140,7 +145,7 @@ and class_type env c =
   let open ClassType in
   let expansion =
     match
-      let open Utils.OptionMonad in
+      let open Odoc_utils.OptionMonad in
       Env.(lookup_by_id s_class_type) c.id env >>= fun (`ClassType (_, c')) ->
       Tools.class_signature_of_class_type env c' >>= fun sg ->
       let cs =
@@ -206,7 +211,7 @@ and class_ env parent c =
   let container = (parent :> Id.LabelParent.t) in
   let expansion =
     match
-      let open Utils.OptionMonad in
+      let open Odoc_utils.OptionMonad in
       Env.(lookup_by_id s_class) c.id env >>= fun (`Class (_, c')) ->
       Tools.class_signature_of_class env c' >>= fun sg ->
       let cs =
@@ -248,7 +253,11 @@ and signature_items : Env.t -> Id.Signature.t -> Signature.item list -> _ =
                 Component.Delayed.(
                   put (fun () -> Component.Of_Lang.(module_ (empty ()) m)))
               in
-              Env.add_module (m.id :> Paths.Identifier.Path.Module.t) ty [] env
+              Env.add_module
+                (m.id :> Paths.Identifier.Path.Module.t)
+                ty
+                { elements = []; warnings_tag = None }
+                env
             in
             let env =
               match r with
@@ -356,14 +365,14 @@ and signature : Env.t -> Id.Signature.t -> Signature.t -> _ =
     {
       Signature.items;
       compiled = true;
+      removed = s.removed;
       doc = s.doc (* comments are ignored while compiling *);
     }
 
 and module_ : Env.t -> Module.t -> Module.t =
  fun env m ->
   let open Module in
-  if m.hidden then m
-  else { m with type_ = module_decl env (m.id :> Id.Signature.t) m.type_ }
+  { m with type_ = module_decl env (m.id :> Id.Signature.t) m.type_ }
 
 and module_decl : Env.t -> Id.Signature.t -> Module.decl -> Module.decl =
  fun env id decl ->
@@ -396,13 +405,12 @@ and include_ : Env.t -> Include.t -> Include.t * Env.t =
   let decl = Component.Of_Lang.(include_decl (empty ()) i.decl) in
   let get_expansion () =
     match
-      let open Utils.ResultMonad in
+      let open Odoc_utils.ResultMonad in
       match decl with
       | Alias p ->
           Tools.expansion_of_module_path env ~strengthen:true p >>= fun exp ->
           Tools.assert_not_functor exp
-      | ModuleType mty ->
-          Tools.signature_of_u_module_type_expr ~mark_substituted:false env mty
+      | ModuleType mty -> Tools.signature_of_u_module_type_expr env mty
     with
     | Error e ->
         Errors.report ~what:(`Include decl) ~tools_error:e `Expand;
@@ -462,7 +470,7 @@ and functor_parameter_parameter :
   { a with expr = module_type_expr env (a.id :> Id.Signature.t) a.expr }
 
 and module_type_expr_sub id ~fragment_root (sg_res, env, subs) lsub =
-  let open Utils.ResultMonad in
+  let open Odoc_utils.ResultMonad in
   match sg_res with
   | Error _ -> (sg_res, env, lsub :: subs)
   | Ok sg -> (
@@ -490,8 +498,7 @@ and module_type_expr_sub id ~fragment_root (sg_res, env, subs) lsub =
             let resolved_csub =
               Component.ModuleType.ModuleEq (cfrag', cdecl')
             in
-            Tools.fragmap ~mark_substituted:true env resolved_csub sg
-            >>= fun sg' ->
+            Tools.fragmap env resolved_csub sg >>= fun sg' ->
             Ok (sg', Odoc_model.Lang.ModuleType.ModuleEq (frag', decl'))
         | TypeEq (frag, eqn) ->
             let cfrag = Component.Of_Lang.(type_fragment (empty ()) frag) in
@@ -510,9 +517,7 @@ and module_type_expr_sub id ~fragment_root (sg_res, env, subs) lsub =
             in
             let eqn' = type_decl_equation env (id :> Id.LabelParent.t) eqn in
             let ceqn' = Component.Of_Lang.(type_equation (empty ()) eqn') in
-            Tools.fragmap ~mark_substituted:true env
-              (Component.ModuleType.TypeEq (cfrag', ceqn'))
-              sg
+            Tools.fragmap env (Component.ModuleType.TypeEq (cfrag', ceqn')) sg
             >>= fun sg' ->
             Ok (sg', Odoc_model.Lang.ModuleType.TypeEq (frag', eqn'))
         | ModuleSubst (frag, mpath) ->
@@ -532,7 +537,7 @@ and module_type_expr_sub id ~fragment_root (sg_res, env, subs) lsub =
             in
             let mpath' = module_path env mpath in
             let cmpath' = Component.Of_Lang.(module_path (empty ()) mpath') in
-            Tools.fragmap ~mark_substituted:true env
+            Tools.fragmap env
               (Component.ModuleType.ModuleSubst (cfrag', cmpath'))
               sg
             >>= fun sg' ->
@@ -553,7 +558,7 @@ and module_type_expr_sub id ~fragment_root (sg_res, env, subs) lsub =
             in
             let eqn' = type_decl_equation env (id :> Id.LabelParent.t) eqn in
             let ceqn' = Component.Of_Lang.(type_equation (empty ()) eqn') in
-            Tools.fragmap ~mark_substituted:true env
+            Tools.fragmap env
               (Component.ModuleType.TypeSubst (cfrag', ceqn'))
               sg
             >>= fun sg' ->
@@ -580,8 +585,7 @@ and module_type_expr_sub id ~fragment_root (sg_res, env, subs) lsub =
             let resolved_csub =
               Component.ModuleType.ModuleTypeEq (cfrag', mty')
             in
-            Tools.fragmap ~mark_substituted:true env resolved_csub sg
-            >>= fun sg' ->
+            Tools.fragmap env resolved_csub sg >>= fun sg' ->
             Ok (sg', Odoc_model.Lang.ModuleType.ModuleTypeEq (frag', mty))
         | Odoc_model.Lang.ModuleType.ModuleTypeSubst (frag, mty) ->
             let cfrag =
@@ -605,8 +609,7 @@ and module_type_expr_sub id ~fragment_root (sg_res, env, subs) lsub =
             let resolved_csub =
               Component.ModuleType.ModuleTypeSubst (cfrag', mty')
             in
-            Tools.fragmap ~mark_substituted:true env resolved_csub sg
-            >>= fun sg' ->
+            Tools.fragmap env resolved_csub sg >>= fun sg' ->
             Ok (sg', Odoc_model.Lang.ModuleType.ModuleTypeSubst (frag', mty))
       in
 
@@ -622,17 +625,15 @@ and module_type_map_subs env id cexpr subs =
     | Path (`Resolved p) -> Some (`ModuleType p)
     | Path _ -> None
     | With (_, e) -> find_parent e
-    | TypeOf { t_desc = ModPath (`Resolved p); _ }
-    | TypeOf { t_desc = StructInclude (`Resolved p); _ } ->
+    | TypeOf (ModPath (`Resolved p), _) | TypeOf (StructInclude (`Resolved p), _)
+      ->
         Some (`Module p)
     | TypeOf _ -> None
   in
   match find_parent cexpr with
   | None -> None
   | Some parent -> (
-      match
-        Tools.signature_of_u_module_type_expr ~mark_substituted:true env cexpr
-      with
+      match Tools.signature_of_u_module_type_expr env cexpr with
       | Error e ->
           Errors.report ~what:(`Module_type id) ~tools_error:e `Lookup;
           None
@@ -665,13 +666,13 @@ and u_module_type_expr :
         in
         let result : ModuleType.U.expr = With (subs', expr') in
         result
-    | TypeOf { t_desc; t_expansion } ->
+    | TypeOf (t_desc, t_original_path) ->
         let t_desc =
           match t_desc with
           | ModPath p -> ModPath (module_path env p)
           | StructInclude p -> StructInclude (module_path env p)
         in
-        TypeOf { t_desc; t_expansion }
+        TypeOf (t_desc, t_original_path)
   in
   inner expr
 
@@ -682,14 +683,14 @@ and module_type_expr :
     ModuleType.expr ->
     ModuleType.expr =
  fun env id ?(expand_paths = true) expr ->
-  let open Utils.ResultMonad in
+  let open Odoc_utils.ResultMonad in
   let get_expansion cur e =
     match cur with
     | Some e -> Some (simple_expansion env id e)
     | None -> (
         let ce = Component.Of_Lang.(module_type_expr (empty ()) e) in
         match
-          Tools.expansion_of_module_type_expr ~mark_substituted:false env ce
+          Tools.expansion_of_module_type_expr env ce
           >>= Expand_tools.handle_expansion env id
         with
         | Ok (_, ce) ->
@@ -709,33 +710,41 @@ and module_type_expr :
       Path { p_path = module_type_path env p_path; p_expansion }
   | With { w_substitutions; w_expansion; w_expr } as e -> (
       let w_expansion = get_expansion w_expansion e in
-      let w_expr = u_module_type_expr env id w_expr in
-      let cexpr = Component.Of_Lang.(u_module_type_expr (empty ()) w_expr) in
-      let subs' = module_type_map_subs env id cexpr w_substitutions in
-      match subs' with
-      | None -> With { w_substitutions; w_expansion; w_expr }
-      | Some s -> With { w_substitutions = s; w_expansion; w_expr })
+      let rec all_withs = function
+        | ModuleType.U.With (_, e) -> all_withs e
+        | Signature _ -> true
+        | _ -> false
+      in
+      match (all_withs w_expr, w_expansion) with
+      | true, Some (Signature e) -> Signature e
+      | _ -> (
+          let w_expr = u_module_type_expr env id w_expr in
+          let cexpr =
+            Component.Of_Lang.(u_module_type_expr (empty ()) w_expr)
+          in
+          let subs' = module_type_map_subs env id cexpr w_substitutions in
+          match subs' with
+          | None -> With { w_substitutions; w_expansion; w_expr }
+          | Some s -> With { w_substitutions = s; w_expansion; w_expr }))
   | Functor (param, res) ->
       let param' = functor_parameter env param in
       let env' = Env.add_functor_parameter param env in
       let res' = module_type_expr env' (Paths.Identifier.Mk.result id) res in
       Functor (param', res')
-  | TypeOf { t_desc; t_expansion } as e ->
+  | TypeOf { t_desc; t_original_path; t_expansion } as e ->
       let t_expansion = get_expansion t_expansion e in
       let t_desc =
         match t_desc with
         | ModPath p -> ModuleType.ModPath (module_path env p)
         | StructInclude p -> StructInclude (module_path env p)
       in
-      TypeOf { t_desc; t_expansion }
+      TypeOf { t_desc; t_original_path; t_expansion }
 
 and type_decl : Env.t -> TypeDecl.t -> TypeDecl.t =
  fun env t ->
   let open TypeDecl in
   let container =
-    match t.id.iv with
-    | `Type (parent, _) -> (parent :> Id.LabelParent.t)
-    | `CoreType _ -> assert false
+    match t.id.iv with `Type (parent, _) -> (parent :> Id.LabelParent.t)
   in
   let equation = type_decl_equation env container t.equation in
   let representation =
@@ -813,9 +822,7 @@ and type_expression_object env parent o =
 and type_expression_package env parent p =
   let open TypeExpr.Package in
   let cp = Component.Of_Lang.(module_type_path (empty ()) p.path) in
-  match
-    Tools.resolve_module_type ~mark_substituted:true ~add_canonical:true env cp
-  with
+  match Tools.resolve_module_type env cp with
   | Ok (path, mt) -> (
       match p.substitutions with
       | [] ->
@@ -851,26 +858,65 @@ and type_expression_package env parent p =
               }))
   | Error _ -> { p with path = Lang_of.(Path.module_type (empty ()) cp) }
 
+and handle_arrow :
+    Env.t ->
+    Id.Id.label_parent ->
+    TypeExpr.label option ->
+    TypeExpr.t ->
+    TypeExpr.t ->
+    TypeExpr.t =
+ fun env parent lbl t1 t2 ->
+  let t2' = type_expression env parent t2 in
+  match lbl with
+  | Some (Optional _ | Label _) | None ->
+      Arrow (lbl, type_expression env parent t1, t2')
+  | Some (RawOptional s) -> (
+      (* s is definitely an option type, but not _obviously_ so. *)
+      match Component.Of_Lang.(type_expression (empty ()) t1) with
+      | Constr (p, _ts) -> (
+          (* This handles only the simplest case *)
+          let find_option t =
+            match Tools.resolve_type env t with
+            | Ok (_, `FType (_n, decl)) -> (
+                match decl.equation.manifest with
+                | Some (Constr (`Resolved (`CoreType n), [ t ]))
+                  when Names.TypeName.to_string n = "option" ->
+                    let t = Lang_of.(type_expr (empty ()) parent t) in
+                    Some t
+                | Some _ -> None
+                | None -> None)
+            | Ok (_, `CoreType _) -> None
+            | Ok (_, (`FClass _ | `FClassType _ | `FType_removed _)) -> None
+            | Error _ -> None
+          in
+          match find_option p with
+          | Some t1 ->
+              Arrow (Some (Optional s), type_expression env parent t1, t2')
+          | None ->
+              Arrow (Some (RawOptional s), type_expression env parent t1, t2'))
+      | _ -> Arrow (Some (RawOptional s), type_expression env parent t1, t2'))
+
 and type_expression : Env.t -> Id.LabelParent.t -> _ -> _ =
  fun env parent texpr ->
   let open TypeExpr in
   match texpr with
   | Var _ | Any -> texpr
   | Alias (t, str) -> Alias (type_expression env parent t, str)
-  | Arrow (lbl, t1, t2) ->
-      Arrow (lbl, type_expression env parent t1, type_expression env parent t2)
+  | Arrow (lbl, t1, t2) -> handle_arrow env parent lbl t1 t2
   | Tuple ts -> Tuple (List.map (type_expression env parent) ts)
   | Constr (path, ts') -> (
       let cp = Component.Of_Lang.(type_path (empty ()) path) in
       let ts = List.map (type_expression env parent) ts' in
-      match Tools.resolve_type env ~add_canonical:true cp with
-      | Ok (cp, (`FType _ | `FClass _ | `FClassType _)) ->
+      match Tools.resolve_type env cp with
+      | Ok (cp, (`FType _ | `FClass _ | `FClassType _ | `CoreType _)) ->
           let p = Lang_of.(Path.resolved_type (empty ()) cp) in
           Constr (`Resolved p, ts)
       | Ok (_cp, `FType_removed (_, x, _eq)) ->
           (* Substitute type variables ? *)
           Lang_of.(type_expr (empty ()) parent x)
-      | Error _e -> Constr (Lang_of.(Path.type_ (empty ()) cp), ts))
+      | Error _e ->
+          Constr ((Lang_of.(Path.type_ (empty ()) cp) :> Paths.Path.Type.t), ts)
+      )
   | Polymorphic_variant v ->
       Polymorphic_variant (type_expression_polyvar env parent v)
   | Object o -> Object (type_expression_object env parent o)
@@ -888,4 +934,7 @@ and type_expression : Env.t -> Id.LabelParent.t -> _ -> _ =
 let compile ~filename env compilation_unit =
   Lookup_failures.catch_failures ~filename (fun () -> unit env compilation_unit)
 
+let compile_impl ~filename env i =
+  Lookup_failures.catch_failures ~filename (fun () -> implementation env i)
+
 let resolve_page _resolver y = y
