@@ -4,25 +4,18 @@ module ModuleMap : Map.S with type key = Ident.module_
 
 module TypeMap : Map.S with type key = Ident.type_
 
-module PathModuleMap : Map.S with type key = Ident.path_module
-(** Useful maps *)
-
 module ModuleTypeMap : Map.S with type key = Ident.module_type
 
-module PathTypeMap : Map.S with type key = Ident.path_type
-
-module PathValueMap : Map.S with type key = Ident.path_value
-
-module PathClassTypeMap : Map.S with type key = Ident.path_class_type
+module ValueMap : Map.S with type key = Ident.value
 
 module IdentMap : Map.S with type key = Ident.any
 
-(** Delayed is a bit like Lazy.t but may in the future offer the chance to peek inside
-    to be able to optimize the calculation *)
+(** Delayed is a bit like Lazy.t but may in the future offer the chance to peek
+    inside to be able to optimize the calculation *)
 module Delayed : sig
   val eager : bool ref
   (** If [eager] is true then no delaying is done. Most useful for testing and
-        documentation *)
+      documentation *)
 
   type 'a t = { mutable v : 'a option; mutable get : (unit -> 'a) option }
 
@@ -38,29 +31,30 @@ module Opt : sig
 end
 
 (** {2 Components}
-  
-    The modules/types here are very similar to those in {!module:Odoc_model.Lang},
-    in most cases the only difference being that we have {{!module:Ident}Idents} 
-    (which are local) rather than {{!module:Odoc_model.Paths.Identifier}Identifiers}
-    (which are global), {{!module:Cpath}Cpaths} instead of {{!module:Odoc_model.Paths.Path}Paths},
-    and {{!module:Cfrag}Cfrags} rather than {{!module:Odoc_model.Paths.Fragment}Fragments}.
-    All of these are to support the use of local idents.
-    
-    Note that we still use global {{!module:Odoc_model.Paths.Reference}References}
-    rather than local ones - this is because at the point these components are being
-    used all the references are unresolved, and hence do not contain any resolved
-    global identifiers. When resolving references, we construct paths in parallel,
+
+    The modules/types here are very similar to those in
+    {!module:Odoc_model.Lang}, in most cases the only difference being that we
+    have {{!module:Ident}Idents} (which are local) rather than
+    {{!module:Odoc_model.Paths.Identifier}Identifiers} (which are global),
+    {{!module:Cpath}Cpaths} instead of {{!module:Odoc_model.Paths.Path}Paths},
+    and {{!module:Cfrag}Cfrags} rather than
+    {{!module:Odoc_model.Paths.Fragment}Fragments}. All of these are to support
+    the use of local idents.
+
+    Note that we still use global
+    {{!module:Odoc_model.Paths.Reference}References} rather than local ones -
+    this is because at the point these components are being used all the
+    references are unresolved, and hence do not contain any resolved global
+    identifiers. When resolving references, we construct paths in parallel,
     which also helps avoid the need for local references.
 
-    These idents of items are kept outside of the types themselves in
-    order to help with laziness.
+    These idents of items are kept outside of the types themselves in order to
+    help with laziness.
 
     There are a few other minor differences:
-    
-    - Signatures keep track of removed items. These items are removed during
-      destructive substitution.
 
-*)
+    - Signatures keep track of removed items. These items are removed during
+      destructive substitution. *)
 
 module rec Module : sig
   type decl =
@@ -68,7 +62,7 @@ module rec Module : sig
     | ModuleType of ModuleType.expr
 
   type t = {
-    locs : Odoc_model.Paths.Identifier.SourceLocation.t option;
+    source_loc : Odoc_model.Paths.Identifier.SourceLocation.t option;
     doc : CComment.docs;
     type_ : decl;
     canonical : Odoc_model.Paths.Path.Module.t option;
@@ -136,7 +130,7 @@ and Extension : sig
   module Constructor : sig
     type t = {
       name : string;
-      locs : Odoc_model.Paths.Identifier.SourceLocation.t option;
+      source_loc : Odoc_model.Paths.Identifier.SourceLocation.t option;
       doc : CComment.docs;
       args : TypeDecl.Constructor.argument;
       res : TypeExpr.t option;
@@ -154,7 +148,7 @@ end
 
 and Exception : sig
   type t = {
-    locs : Odoc_model.Paths.Identifier.SourceLocation.t option;
+    source_loc : Odoc_model.Paths.Identifier.SourceLocation.t option;
     doc : CComment.docs;
     args : TypeDecl.Constructor.argument;
     res : TypeExpr.t option;
@@ -162,7 +156,7 @@ and Exception : sig
 end
 
 and FunctorParameter : sig
-  type parameter = { id : Ident.functor_parameter; expr : ModuleType.expr }
+  type parameter = { id : Ident.module_; expr : ModuleType.expr }
 
   type t = Named of parameter | Unit
 end
@@ -186,6 +180,7 @@ and ModuleType : sig
 
   type typeof_t = {
     t_desc : type_of_desc;
+    t_original_path : Cpath.module_;
     t_expansion : simple_expansion option;
   }
 
@@ -194,7 +189,7 @@ and ModuleType : sig
       | Path of Cpath.module_type
       | Signature of Signature.t
       | With of substitution list * expr
-      | TypeOf of typeof_t
+      | TypeOf of type_of_desc * Cpath.module_
   end
 
   type path_t = {
@@ -216,7 +211,7 @@ and ModuleType : sig
     | TypeOf of typeof_t
 
   type t = {
-    locs : Odoc_model.Paths.Identifier.SourceLocation.t option;
+    source_loc : Odoc_model.Paths.Identifier.SourceLocation.t option;
     doc : CComment.docs;
     canonical : Odoc_model.Paths.Path.ModuleType.t option;
     expr : expr option;
@@ -263,7 +258,7 @@ and TypeDecl : sig
   end
 
   type t = {
-    locs : Odoc_model.Paths.Identifier.SourceLocation.t option;
+    source_loc : Odoc_model.Paths.Identifier.SourceLocation.t option;
     doc : CComment.docs;
     canonical : Odoc_model.Paths.Path.Type.t option;
     equation : Equation.t;
@@ -284,8 +279,8 @@ and Signature : sig
     | Exception of Ident.exception_ * Exception.t
     | TypExt of Extension.t
     | Value of Ident.value * Value.t Delayed.t
-    | Class of Ident.class_ * recursive * Class.t
-    | ClassType of Ident.class_type * recursive * ClassType.t
+    | Class of Ident.type_ * recursive * Class.t
+    | ClassType of Ident.type_ * recursive * ClassType.t
     | Include of Include.t
     | Open of Open.t
     | Comment of CComment.docs_or_stop
@@ -293,9 +288,9 @@ and Signature : sig
   (* When doing destructive substitution we keep track of the items that have been removed,
        and the path they've been substituted with *)
   type removed_item =
-    | RModule of Ident.module_ * Cpath.Resolved.module_
-    | RType of Ident.type_ * TypeExpr.t * TypeDecl.Equation.t
-    | RModuleType of Ident.module_type * ModuleType.expr
+    | RModule of Odoc_model.Names.ModuleName.t * Cpath.module_
+    | RType of Odoc_model.Names.TypeName.t * TypeExpr.t * TypeDecl.Equation.t
+    | RModuleType of Odoc_model.Names.ModuleTypeName.t * ModuleType.expr
 
   type t = {
     items : item list;
@@ -328,7 +323,7 @@ and Value : sig
   type value = Odoc_model.Lang.Value.value
 
   type t = {
-    locs : Odoc_model.Paths.Identifier.SourceLocation.t option;
+    source_loc : Odoc_model.Paths.Identifier.SourceLocation.t option;
     doc : CComment.docs;
     type_ : TypeExpr.t;
     value : value;
@@ -341,7 +336,7 @@ and Class : sig
     | Arrow of TypeExpr.label option * TypeExpr.t * decl
 
   type t = {
-    locs : Odoc_model.Paths.Identifier.SourceLocation.t option;
+    source_loc : Odoc_model.Paths.Identifier.SourceLocation.t option;
     doc : CComment.docs;
     virtual_ : bool;
     params : TypeDecl.param list;
@@ -356,7 +351,7 @@ and ClassType : sig
     | Signature of ClassSignature.t
 
   type t = {
-    locs : Odoc_model.Paths.Identifier.SourceLocation.t option;
+    source_loc : Odoc_model.Paths.Identifier.SourceLocation.t option;
     doc : CComment.docs;
     virtual_ : bool;
     params : TypeDecl.param list;
@@ -406,29 +401,27 @@ and Substitution : sig
   type subst_module =
     [ `Prefixed of Cpath.module_ * Cpath.Resolved.module_
     | `Substituted
-    | `Renamed of Ident.path_module ]
+    | `Renamed of Ident.module_ ]
 
   type subst_module_type =
     [ `Prefixed of Cpath.module_type * Cpath.Resolved.module_type
     | `Renamed of Ident.module_type ]
 
   type subst_type =
-    [ `Prefixed of Cpath.type_ * Cpath.Resolved.type_
-    | `Renamed of Ident.path_type ]
+    [ `Prefixed of Cpath.type_ * Cpath.Resolved.type_ | `Renamed of Ident.type_ ]
 
   type subst_class_type =
     [ `Prefixed of Cpath.class_type * Cpath.Resolved.class_type
-    | `Renamed of Ident.path_class_type ]
+    | `Renamed of Ident.type_ ]
 
   type t = {
-    module_ : subst_module PathModuleMap.t;
+    module_ : subst_module ModuleMap.t;
     module_type : subst_module_type ModuleTypeMap.t;
-    type_ : subst_type PathTypeMap.t;
-    class_type : subst_class_type PathClassTypeMap.t;
-    type_replacement : (TypeExpr.t * TypeDecl.Equation.t) PathTypeMap.t;
+    type_ : subst_type TypeMap.t;
+    class_type : subst_class_type TypeMap.t;
+    type_replacement : (TypeExpr.t * TypeDecl.Equation.t) TypeMap.t;
     module_type_replacement : ModuleType.expr ModuleTypeMap.t;
-    path_invalidating_modules : Ident.path_module list;
-    module_type_of_invalidating_modules : Ident.path_module list;
+    path_invalidating_modules : Ident.module_ list;
     unresolve_opaque_paths : bool;
   }
 end
@@ -437,9 +430,14 @@ and CComment : sig
   type block_element =
     [ Odoc_model.Comment.nestable_block_element
     | `Heading of Label.t
-    | `Tag of Odoc_model.Comment.tag ]
+    | `Tag of Odoc_model.Comment.tag
+    | `Media of
+      Odoc_model.Comment.media_href * Odoc_model.Comment.media * string ]
 
-  type docs = block_element Odoc_model.Comment.with_location list
+  type docs = {
+    elements : block_element Odoc_model.Comment.with_location list;
+    warnings_tag : string option;
+  }
 
   type docs_or_stop = [ `Docs of docs | `Stop ]
 end
@@ -514,131 +512,152 @@ end
 
 (** Formatting functions for components *)
 module Fmt : sig
-  val signature : Format.formatter -> Signature.t -> unit
+  type config = {
+    short_paths : bool;
+    show_canonical : bool;
+    show_removed : bool;
+    show_expansions : bool;
+    show_include_expansions : bool;
+  }
 
-  val removed_item : Format.formatter -> Signature.removed_item -> unit
+  val default : config
+
+  val signature : config -> Format.formatter -> Signature.t -> unit
+
+  val removed_item :
+    config -> Format.formatter -> Signature.removed_item -> unit
 
   val removed_item_list :
-    Format.formatter -> Signature.removed_item list -> unit
+    config -> Format.formatter -> Signature.removed_item list -> unit
 
-  val class_ : Format.formatter -> Class.t -> unit
+  val class_ : config -> Format.formatter -> Class.t -> unit
 
-  val class_type : Format.formatter -> ClassType.t -> unit
+  val class_type : config -> Format.formatter -> ClassType.t -> unit
 
-  val include_ : Format.formatter -> Include.t -> unit
+  val include_ : config -> Format.formatter -> Include.t -> unit
 
-  val value : Format.formatter -> Value.t -> unit
+  val value : config -> Format.formatter -> Value.t -> unit
 
-  val module_decl : Format.formatter -> Module.decl -> unit
+  val module_decl : config -> Format.formatter -> Module.decl -> unit
 
-  val include_decl : Format.formatter -> Include.decl -> unit
+  val include_decl : config -> Format.formatter -> Include.decl -> unit
 
-  val module_ : Format.formatter -> Module.t -> unit
+  val module_ : config -> Format.formatter -> Module.t -> unit
 
-  val module_type : Format.formatter -> ModuleType.t -> unit
+  val module_type : config -> Format.formatter -> ModuleType.t -> unit
 
-  val simple_expansion : Format.formatter -> ModuleType.simple_expansion -> unit
+  val simple_expansion :
+    config -> bool -> Format.formatter -> ModuleType.simple_expansion -> unit
 
   val module_type_type_of_desc :
-    Format.formatter -> ModuleType.type_of_desc -> unit
+    config -> Format.formatter -> ModuleType.type_of_desc -> unit
 
-  val u_module_type_expr : Format.formatter -> ModuleType.U.expr -> unit
+  val u_module_type_expr :
+    config -> Format.formatter -> ModuleType.U.expr -> unit
 
-  val module_type_expr : Format.formatter -> ModuleType.expr -> unit
+  val module_type_expr : config -> Format.formatter -> ModuleType.expr -> unit
 
-  val functor_parameter : Format.formatter -> FunctorParameter.t -> unit
+  val functor_parameter :
+    config -> Format.formatter -> FunctorParameter.t -> unit
 
   val functor_parameter_parameter :
-    Format.formatter -> FunctorParameter.parameter -> unit
+    config -> Format.formatter -> FunctorParameter.parameter -> unit
 
-  val type_decl : Format.formatter -> TypeDecl.t -> unit
+  val type_decl : config -> Format.formatter -> TypeDecl.t -> unit
 
-  val type_equation : Format.formatter -> TypeDecl.Equation.t -> unit
+  val type_equation : config -> Format.formatter -> TypeDecl.Equation.t -> unit
 
-  val exception_ : Format.formatter -> Exception.t -> unit
+  val exception_ : config -> Format.formatter -> Exception.t -> unit
 
-  val extension : Format.formatter -> Extension.t -> unit
+  val extension : config -> Format.formatter -> Extension.t -> unit
 
-  val substitution : Format.formatter -> ModuleType.substitution -> unit
+  val substitution :
+    config -> Format.formatter -> ModuleType.substitution -> unit
 
   val substitution_list :
-    Format.formatter -> ModuleType.substitution list -> unit
+    config -> Format.formatter -> ModuleType.substitution list -> unit
 
-  val type_expr_list : Format.formatter -> TypeExpr.t list -> unit
+  val type_expr_list : config -> Format.formatter -> TypeExpr.t list -> unit
 
-  val type_object : Format.formatter -> TypeExpr.Object.t -> unit
+  val type_object : config -> Format.formatter -> TypeExpr.Object.t -> unit
 
   val type_class :
-    Format.formatter -> Cpath.class_type * TypeExpr.t list -> unit
+    config -> Format.formatter -> Cpath.class_type * TypeExpr.t list -> unit
 
-  val type_package : Format.formatter -> TypeExpr.Package.t -> unit
+  val type_package : config -> Format.formatter -> TypeExpr.Package.t -> unit
 
   val type_expr_polymorphic_variant :
-    Format.formatter -> TypeExpr.Polymorphic_variant.t -> unit
+    config -> Format.formatter -> TypeExpr.Polymorphic_variant.t -> unit
 
-  val type_expr : Format.formatter -> TypeExpr.t -> unit
+  val type_expr : config -> Format.formatter -> TypeExpr.t -> unit
 
-  val resolved_module_path : Format.formatter -> Cpath.Resolved.module_ -> unit
+  val resolved_module_path :
+    config -> Format.formatter -> Cpath.Resolved.module_ -> unit
 
-  val module_path : Format.formatter -> Cpath.module_ -> unit
+  val module_path : config -> Format.formatter -> Cpath.module_ -> unit
 
   val resolved_module_type_path :
-    Format.formatter -> Cpath.Resolved.module_type -> unit
+    config -> Format.formatter -> Cpath.Resolved.module_type -> unit
 
-  val module_type_path : Format.formatter -> Cpath.module_type -> unit
+  val module_type_path : config -> Format.formatter -> Cpath.module_type -> unit
 
-  val resolved_type_path : Format.formatter -> Cpath.Resolved.type_ -> unit
+  val resolved_type_path :
+    config -> Format.formatter -> Cpath.Resolved.type_ -> unit
 
-  val resolved_value_path : Format.formatter -> Cpath.Resolved.value -> unit
+  val resolved_value_path :
+    config -> Format.formatter -> Cpath.Resolved.value -> unit
 
-  val resolved_parent_path : Format.formatter -> Cpath.Resolved.parent -> unit
+  val resolved_parent_path :
+    config -> Format.formatter -> Cpath.Resolved.parent -> unit
 
-  val type_path : Format.formatter -> Cpath.type_ -> unit
+  val type_path : config -> Format.formatter -> Cpath.type_ -> unit
 
-  val value_path : Format.formatter -> Cpath.value -> unit
-
-  val constructor_path : Format.formatter -> Cpath.constructor -> unit
+  val value_path : config -> Format.formatter -> Cpath.value -> unit
 
   val resolved_class_type_path :
-    Format.formatter -> Cpath.Resolved.class_type -> unit
+    config -> Format.formatter -> Cpath.Resolved.class_type -> unit
 
-  val class_type_path : Format.formatter -> Cpath.class_type -> unit
+  val class_type_path : config -> Format.formatter -> Cpath.class_type -> unit
 
-  val model_path : Format.formatter -> Odoc_model.Paths.Path.t -> unit
+  val model_path : config -> Format.formatter -> Odoc_model.Paths.Path.t -> unit
 
   val model_resolved_path :
-    Format.formatter -> Odoc_model.Paths.Path.Resolved.t -> unit
+    config -> Format.formatter -> Odoc_model.Paths.Path.Resolved.t -> unit
 
   val model_identifier :
-    Format.formatter -> Odoc_model.Paths.Identifier.t -> unit
+    config -> Format.formatter -> Odoc_model.Paths.Identifier.t -> unit
 
-  val model_fragment : Format.formatter -> Odoc_model.Paths.Fragment.t -> unit
+  val model_fragment :
+    config -> Format.formatter -> Odoc_model.Paths.Fragment.t -> unit
 
   val model_resolved_fragment :
-    Format.formatter -> Odoc_model.Paths.Fragment.Resolved.t -> unit
+    config -> Format.formatter -> Odoc_model.Paths.Fragment.Resolved.t -> unit
 
-  val resolved_root_fragment : Format.formatter -> Cfrag.root -> unit
+  val resolved_root_fragment : config -> Format.formatter -> Cfrag.root -> unit
 
   val resolved_signature_fragment :
-    Format.formatter -> Cfrag.resolved_signature -> unit
+    config -> Format.formatter -> Cfrag.resolved_signature -> unit
 
   val resolved_module_fragment :
-    Format.formatter -> Cfrag.resolved_module -> unit
+    config -> Format.formatter -> Cfrag.resolved_module -> unit
 
-  val resolved_type_fragment : Format.formatter -> Cfrag.resolved_type -> unit
+  val resolved_type_fragment :
+    config -> Format.formatter -> Cfrag.resolved_type -> unit
 
-  val signature_fragment : Format.formatter -> Cfrag.signature -> unit
+  val signature_fragment : config -> Format.formatter -> Cfrag.signature -> unit
 
-  val module_fragment : Format.formatter -> Cfrag.module_ -> unit
+  val module_fragment : config -> Format.formatter -> Cfrag.module_ -> unit
 
-  val module_type_fragment : Format.formatter -> Cfrag.module_type -> unit
+  val module_type_fragment :
+    config -> Format.formatter -> Cfrag.module_type -> unit
 
-  val type_fragment : Format.formatter -> Cfrag.type_ -> unit
+  val type_fragment : config -> Format.formatter -> Cfrag.type_ -> unit
 
   val model_resolved_reference :
-    Format.formatter -> Odoc_model.Paths.Reference.Resolved.t -> unit
+    config -> Format.formatter -> Odoc_model.Paths.Reference.Resolved.t -> unit
 
-  val model_reference : Format.formatter -> Odoc_model.Paths.Reference.t -> unit
+  val model_reference :
+    config -> Format.formatter -> Odoc_model.Paths.Reference.t -> unit
 end
 
 module Of_Lang : sig
@@ -663,11 +682,6 @@ module Of_Lang : sig
   val resolved_value_path :
     map -> Odoc_model.Paths.Path.Resolved.Value.t -> Cpath.Resolved.value
 
-  val resolved_constructor_path :
-    map ->
-    Odoc_model.Paths.Path.Resolved.Constructor.t ->
-    Cpath.Resolved.constructor
-
   val resolved_class_type_path :
     map ->
     Odoc_model.Paths.Path.Resolved.ClassType.t ->
@@ -680,13 +694,8 @@ module Of_Lang : sig
 
   val type_path : map -> Odoc_model.Paths.Path.Type.t -> Cpath.type_
 
-  val datatype : map -> Odoc_model.Paths.Path.DataType.t -> Cpath.datatype
-
   val value_path : map -> Odoc_model.Paths.Path.Value.t -> Cpath.value
 
-  val constructor_path :
-    map -> Odoc_model.Paths.Path.Constructor.t -> Cpath.constructor
-
   val class_type_path :
     map -> Odoc_model.Paths.Path.ClassType.t -> Cpath.class_type
 
@@ -762,7 +771,7 @@ module Of_Lang : sig
 
   val functor_parameter :
     map ->
-    Ident.functor_parameter ->
+    Ident.module_ ->
     Odoc_model.Lang.FunctorParameter.parameter ->
     FunctorParameter.parameter
 
