@@ -2,24 +2,26 @@
 open Odoc_model
 open Odoc_model.Names
 open Odoc_model.Paths
+open Odoc_utils
 
-type lookup_unit_result =
-  | Forward_reference
-  | Found of Lang.Compilation_unit.t
-  | Not_found
+type lookup_unit_result = Forward_reference | Found of Lang.Compilation_unit.t
 
-type lookup_page_result = Lang.Page.t option
+type path_query = [ `Path of Reference.Hierarchy.t | `Name of string ]
 
-type root =
-  | Resolved of (Odoc_model.Root.t * Identifier.Module.t * Component.Module.t)
-  | Forward
+type lookup_error = [ `Not_found ]
 
 type resolver = {
   open_units : string list;
-  lookup_unit : string -> lookup_unit_result;
-  lookup_page : string -> lookup_page_result;
+  lookup_unit : path_query -> (lookup_unit_result, lookup_error) result;
+  lookup_page : path_query -> (Lang.Page.t, lookup_error) result;
+  lookup_asset : path_query -> (Lang.Asset.t, lookup_error) result;
+  lookup_impl : string -> Lang.Implementation.t option;
 }
 
+type root =
+  | Resolved of (Root.t * Identifier.Module.t * Component.Module.t)
+  | Forward
+
 let unique_id =
   let i = ref 0 in
   fun () ->
@@ -29,7 +31,7 @@ let unique_id =
 type lookup_type =
   | Module of Paths.Identifier.Path.Module.t
   | ModuleType of Paths.Identifier.Path.ModuleType.t
-  | RootModule of string * [ `Forward | `Resolved of Digest.t ] option
+  | RootModule of ModuleName.t * [ `Forward | `Resolved of Digest.t ] option
   | ModuleByName of string * Paths.Identifier.Path.Module.t
   | FragmentRoot of int
 
@@ -39,16 +41,21 @@ let pp_lookup_type fmt =
     | Some (`Resolved digest) -> Format.fprintf fmt "Some (Resolved %s)" digest
     | None -> Format.fprintf fmt "None"
   in
+  let c = Component.Fmt.default in
   function
   | Module r ->
-      Format.fprintf fmt "Module %a" Component.Fmt.model_identifier
+      Format.fprintf fmt "Module %a"
+        (Component.Fmt.model_identifier c)
         (r :> Identifier.t)
   | ModuleType r ->
-      Format.fprintf fmt "ModuleType %a" Component.Fmt.model_identifier
+      Format.fprintf fmt "ModuleType %a"
+        (Component.Fmt.model_identifier c)
         (r :> Identifier.t)
-  | RootModule (str, res) -> Format.fprintf fmt "RootModule %s %a" str fmtrm res
+  | RootModule (n, res) ->
+      Format.fprintf fmt "RootModule %a %a" ModuleName.fmt n fmtrm res
   | ModuleByName (n, r) ->
-      Format.fprintf fmt "ModuleByName %s, %a" n Component.Fmt.model_identifier
+      Format.fprintf fmt "ModuleByName %s, %a" n
+        (Component.Fmt.model_identifier c)
         (r :> Identifier.t)
   | FragmentRoot i -> Format.fprintf fmt "FragmentRoot %d" i
 
@@ -162,9 +169,16 @@ type t = {
   ambiguous_labels : Component.Element.label amb_err Identifier.Maps.Label.t;
   resolver : resolver option;
   recorder : recorder option;
+  warnings_tags : string list;
   fragmentroot : (int * Component.Signature.t) option;
 }
 
+let should_suppress_warnings env opt =
+  match opt with None -> false | Some x -> not (List.mem x env.warnings_tags)
+(* Suppress warnings unless the tag is in the list *)
+
+let set_warnings_tags env tags = { env with warnings_tags = tags }
+
 let is_linking env = env.linking
 
 let set_resolver t resolver = { t with resolver = Some resolver }
@@ -198,6 +212,7 @@ let empty =
     resolver = None;
     recorder = None;
     ambiguous_labels = Identifier.Maps.Label.empty;
+    warnings_tags = [];
     fragmentroot = None;
   }
 
@@ -266,7 +281,7 @@ let add_docs (docs : Comment.docs) env =
           let label = Ident.Of_Identifier.label id in
           add_label id { Component.Label.attrs; label; text; location } env
       | _ -> env)
-    env docs
+    env docs.elements
 
 let add_comment (com : Comment.docs_or_stop) env =
   match com with `Docs doc -> add_docs doc env | `Stop -> env
@@ -282,7 +297,7 @@ let add_cdocs p (docs : Component.CComment.docs) env =
           in
           add_label label h env
       | _ -> env)
-    env docs
+    env docs.elements
 
 let add_module identifier m docs env =
   let env' = add_to_elts Kind_Module identifier (`Module (identifier, m)) env in
@@ -359,19 +374,14 @@ let add_extension_constructor identifier
 let module_of_unit : Lang.Compilation_unit.t -> Component.Module.t =
  fun unit ->
   let id = (unit.id :> Paths.Identifier.Module.t) in
-  let locs =
-    match unit.source_info with
-    | Some src -> Some (Identifier.Mk.source_location_mod src.id)
-    | None -> None
-  in
   match unit.content with
   | Module s ->
       let m =
         Lang.Module.
           {
             id;
-            locs;
-            doc = [];
+            source_loc = None;
+            doc = { elements = []; warnings_tag = None };
             type_ = ModuleType (Signature s);
             canonical = unit.canonical;
             hidden = unit.hidden;
@@ -384,10 +394,17 @@ let module_of_unit : Lang.Compilation_unit.t -> Component.Module.t =
         Lang.Module.
           {
             id;
-            locs;
-            doc = [];
+            source_loc = None;
+            doc = { elements = []; warnings_tag = None };
             type_ =
-              ModuleType (Signature { items = []; compiled = true; doc = [] });
+              ModuleType
+                (Signature
+                   {
+                     items = [];
+                     compiled = true;
+                     removed = [];
+                     doc = { elements = []; warnings_tag = None };
+                   });
             canonical = unit.canonical;
             hidden = unit.hidden;
           }
@@ -400,10 +417,10 @@ let lookup_root_module name env =
     match env.resolver with
     | None -> None
     | Some r -> (
-        match r.lookup_unit name with
-        | Forward_reference -> Some Forward
-        | Not_found -> None
-        | Found u ->
+        match r.lookup_unit (`Name (ModuleName.to_string name)) with
+        | Ok Forward_reference -> Some Forward
+        | Error `Not_found -> None
+        | Ok (Found u) ->
             let ({ Odoc_model.Paths.Identifier.iv = `Root _; _ } as id) =
               u.id
             in
@@ -424,11 +441,37 @@ let lookup_root_module name env =
   | None, _ -> ());
   result
 
-let lookup_page name env =
-  match env.resolver with None -> None | Some r -> r.lookup_page name
+let lookup_page query env =
+  match env.resolver with
+  | None -> Error `Not_found
+  | Some r -> r.lookup_page query
 
-let lookup_unit name env =
-  match env.resolver with None -> None | Some r -> Some (r.lookup_unit name)
+let lookup_asset query env =
+  match env.resolver with
+  | None -> Error `Not_found
+  | Some r -> r.lookup_asset query
+
+let lookup_unit query env =
+  match env.resolver with
+  | None -> Error `Not_found
+  | Some r -> r.lookup_unit query
+
+let lookup_impl name env =
+  match env.resolver with None -> None | Some r -> r.lookup_impl name
+
+let lookup_page_by_name n env = lookup_page (`Name n) env
+let lookup_page_by_path p env = lookup_page (`Path p) env
+
+let lookup_asset_by_name p env = lookup_asset (`Name p) env
+let lookup_asset_by_path p env = lookup_asset (`Path p) env
+
+let lookup_unit_by_path p env =
+  match lookup_unit (`Path p) env with
+  | Ok (Found u) ->
+      let m = Component.Delayed.put_val (module_of_unit u) in
+      Ok (`Module ((u.id :> Identifier.Path.Module.t), m))
+  | Ok Forward_reference -> Error `Not_found (* TODO: Remove this case *)
+  | Error _ as e -> e
 
 type 'a scope = {
   filter : Component.Element.any -> ([< Component.Element.any ] as 'a) option;
@@ -496,20 +539,19 @@ let lookup_by_id (scope : 'a scope) id env : 'a option =
       | _ -> None)
 
 let lookup_root_module_fallback name t =
-  match lookup_root_module name t with
+  match lookup_root_module (ModuleName.make_std name) t with
   | Some (Resolved (_, id, m)) ->
       Some
-        (`Module
-          ((id :> Identifier.Path.Module.t), Component.Delayed.put_val m))
+        (`Module ((id :> Identifier.Path.Module.t), Component.Delayed.put_val m))
   | Some Forward | None -> None
 
 let lookup_page_or_root_module_fallback name t =
   match lookup_root_module_fallback name t with
   | Some _ as x -> x
   | None -> (
-      match lookup_page name t with
-      | Some page -> Some (`Page (page.Lang.Page.name, page))
-      | None -> None)
+      match lookup_page_by_name name t with
+      | Ok page -> Some (`Page (page.Lang.Page.name, page))
+      | Error `Not_found -> None)
 
 let s_signature : Component.Element.signature scope =
   make_scope ~root:lookup_root_module_fallback (function
@@ -614,7 +656,13 @@ let lookup_fragment_root env =
 let mk_functor_parameter module_type =
   let type_ = Component.Module.ModuleType module_type in
   Component.Module.
-    { locs = None; doc = []; type_; canonical = None; hidden = false }
+    {
+      source_loc = None;
+      doc = { elements = []; warnings_tag = None };
+      type_;
+      canonical = None;
+      hidden = false;
+    }
 
 let add_functor_parameter : Lang.FunctorParameter.t -> t -> t =
  fun p t ->
@@ -626,7 +674,10 @@ let add_functor_parameter : Lang.FunctorParameter.t -> t -> t =
         let open Component.Of_Lang in
         mk_functor_parameter (module_type_expr (empty ()) n.expr)
       in
-      add_module id (Component.Delayed.put_val m) [] t
+      add_module id
+        (Component.Delayed.put_val m)
+        { elements = []; warnings_tag = None }
+        t
 
 let add_functor_args' :
     Paths.Identifier.Signature.t -> Component.ModuleType.expr -> t -> t =
@@ -637,9 +688,7 @@ let add_functor_args' :
       | ModuleType.Functor (Named arg, res) ->
           ( arg.Component.FunctorParameter.id,
             Paths.Identifier.Mk.parameter
-              ( parent,
-                Ident.Name.typed_functor_parameter
-                  arg.Component.FunctorParameter.id ),
+              (parent, Ident.Name.typed_module arg.Component.FunctorParameter.id),
             mk_functor_parameter arg.expr )
           :: find_args (Paths.Identifier.Mk.result parent) res
       | ModuleType.Functor (Unit, res) ->
@@ -652,7 +701,7 @@ let add_functor_args' :
        themselves *)
     let fold_fn (env, subst) (ident, identifier, m) =
       let ident, identifier =
-        ((ident, identifier) :> Ident.path_module * Identifier.Path.Module.t)
+        ((ident, identifier) :> Ident.module_ * Identifier.Path.Module.t)
       in
       let doc = m.Component.Module.doc in
       let m = Component.Delayed.put_val (Subst.module_ subst m) in
@@ -782,7 +831,7 @@ let open_module_type_substitution : Lang.ModuleTypeSubstitution.t -> t -> t =
     module_type (empty ())
       {
         id = t.id;
-        locs = None;
+        source_loc = None;
         doc = t.doc;
         expr = Some t.manifest;
         canonical = None;
@@ -790,20 +839,24 @@ let open_module_type_substitution : Lang.ModuleTypeSubstitution.t -> t -> t =
   in
   add_module_type t.Lang.ModuleTypeSubstitution.id ty env
 
-let inherit_resolver env =
-  match env.resolver with Some r -> set_resolver empty r | None -> empty
-
 let open_units resolver env =
   List.fold_left
     (fun env m ->
-      match resolver.lookup_unit m with
-      | Found unit -> (
+      match resolver.lookup_unit (`Name m) with
+      | Ok (Found unit) -> (
           match unit.content with
           | Module sg -> open_signature sg env
           | _ -> env)
       | _ -> env)
     env resolver.open_units
 
+let inherit_resolver env =
+  match env.resolver with
+  | Some r ->
+      let e = set_resolver empty r in
+      open_units r e
+  | None -> empty
+
 let env_of_unit t ~linking resolver =
   let open Lang.Compilation_unit in
   let initial_env =
@@ -820,6 +873,9 @@ let env_of_page page resolver =
   let initial_env = open_page page empty in
   set_resolver initial_env resolver |> open_units resolver
 
+let env_of_impl _impl resolver =
+  set_resolver empty resolver |> open_units resolver
+
 let env_for_reference resolver =
   set_resolver empty resolver |> open_units resolver
 
@@ -839,10 +895,10 @@ let verify_lookups env lookups =
           match env.resolver with
           | None -> None
           | Some r -> (
-              match r.lookup_unit name with
-              | Forward_reference -> Some `Forward
-              | Not_found -> None
-              | Found u -> Some (`Resolved u.root.digest))
+              match r.lookup_unit (`Name (ModuleName.to_string name)) with
+              | Ok Forward_reference -> Some `Forward
+              | Ok (Found u) -> Some (`Resolved u.root.digest)
+              | Error `Not_found -> None)
         in
         match (res, actual_result) with
         | None, None -> false
