@@ -2,29 +2,32 @@
 
 open Odoc_model
 open Odoc_model.Paths
+open Odoc_utils
 
-type lookup_unit_result =
-  | Forward_reference
-  | Found of Lang.Compilation_unit.t
-  | Not_found
+type lookup_unit_result = Forward_reference | Found of Lang.Compilation_unit.t
 
-type lookup_page_result = Lang.Page.t option
+type path_query = [ `Path of Reference.Hierarchy.t | `Name of string ]
 
-type root =
-  | Resolved of
-      (Root.t * Odoc_model.Paths.Identifier.Module.t * Component.Module.t)
-  | Forward
+type lookup_error = [ `Not_found ]
 
 type resolver = {
   open_units : string list;
-  lookup_unit : string -> lookup_unit_result;
-  lookup_page : string -> lookup_page_result;
+  lookup_unit : path_query -> (lookup_unit_result, lookup_error) result;
+  lookup_page : path_query -> (Lang.Page.t, lookup_error) result;
+  lookup_asset : path_query -> (Lang.Asset.t, lookup_error) result;
+  lookup_impl : string -> Lang.Implementation.t option;
 }
 
+type root =
+  | Resolved of (Root.t * Identifier.Module.t * Component.Module.t)
+  | Forward
+
 type lookup_type =
   | Module of Identifier.Path.Module.t
   | ModuleType of Identifier.ModuleType.t
-  | RootModule of string * [ `Forward | `Resolved of Digest.t ] option
+  | RootModule of
+      Odoc_model.Names.ModuleName.t
+      * [ `Forward | `Resolved of Digest.t ] option
   | ModuleByName of string * Identifier.Path.Module.t
   | FragmentRoot of int
 
@@ -91,13 +94,23 @@ val add_module_type_functor_args :
 
 val lookup_fragment_root : t -> (int * Component.Signature.t) option
 
-val lookup_page : string -> t -> Lang.Page.t option
+val lookup_page_by_name : string -> t -> (Lang.Page.t, lookup_error) result
+val lookup_page_by_path :
+  Reference.Hierarchy.t -> t -> (Lang.Page.t, lookup_error) result
 
-val lookup_unit : string -> t -> lookup_unit_result option
+val lookup_asset_by_path :
+  Reference.Hierarchy.t -> t -> (Lang.Asset.t, lookup_error) result
+
+val lookup_asset_by_name : string -> t -> (Lang.Asset.t, lookup_error) result
+
+val lookup_impl : string -> t -> Lang.Implementation.t option
+
+val lookup_unit_by_path :
+  Reference.Hierarchy.t -> t -> (Component.Element.module_, lookup_error) result
 
 val module_of_unit : Lang.Compilation_unit.t -> Component.Module.t
 
-val lookup_root_module : string -> t -> root option
+val lookup_root_module : Odoc_model.Names.ModuleName.t -> t -> root option
 
 type 'a scope constraint 'a = [< Component.Element.any ]
 (** Target of a lookup *)
@@ -170,6 +183,9 @@ val env_of_unit : Lang.Compilation_unit.t -> linking:bool -> resolver -> t
 val env_of_page : Lang.Page.t -> resolver -> t
 (** Create a new env for a page. *)
 
+val env_of_impl : Lang.Implementation.t -> resolver -> t
+(** Create a new env for an implementation. *)
+
 val env_for_reference : resolver -> t
 (** Create a new env for a reference. *)
 
@@ -183,4 +199,8 @@ val len : int ref
 
 val n : int ref
 
+val should_suppress_warnings : t -> string option -> bool
+
+val set_warnings_tags : t -> string list -> t
+
 val verify_lookups : t -> LookupTypeSet.t -> bool
