@@ -5,13 +5,12 @@ open Names
 type maps = {
   module_ : Identifier.Module.t Component.ModuleMap.t;
   module_type : Identifier.ModuleType.t Component.ModuleTypeMap.t;
-  functor_parameter :
-    (Ident.functor_parameter * Identifier.FunctorParameter.t) list;
+  functor_parameter : (Ident.module_ * Identifier.FunctorParameter.t) list;
   type_ : Identifier.Type.t Component.TypeMap.t;
-  path_type : Identifier.Path.Type.t Component.PathTypeMap.t;
-  class_ : (Ident.class_ * Identifier.Class.t) list;
-  class_type : (Ident.class_type * Identifier.ClassType.t) list;
-  path_class_type : Identifier.Path.ClassType.t Component.PathClassTypeMap.t;
+  path_type : Identifier.Path.Type.t Component.TypeMap.t;
+  class_ : (Ident.type_ * Identifier.Class.t) list;
+  class_type : (Ident.type_ * Identifier.ClassType.t) list;
+  path_class_type : Identifier.Path.ClassType.t Component.TypeMap.t;
   fragment_root : Cfrag.root option;
   (* Shadowed items *)
   shadowed : Lang.Include.shadowed;
@@ -34,10 +33,10 @@ let empty () =
     module_type = Component.ModuleTypeMap.empty;
     functor_parameter = [];
     type_ = Component.TypeMap.empty;
-    path_type = Component.PathTypeMap.empty;
+    path_type = Component.TypeMap.empty;
     class_ = [];
     class_type = [];
-    path_class_type = Component.PathClassTypeMap.empty;
+    path_class_type = Component.TypeMap.empty;
     fragment_root = None;
     shadowed = empty_shadow;
   }
@@ -47,19 +46,16 @@ let with_fragment_root r = { (empty ()) with fragment_root = Some r }
 let with_shadowed shadowed = { (empty ()) with shadowed }
 
 (** Raises [Not_found] *)
-let lookup_module map : Ident.path_module -> _ = function
-  | (`LRoot _ | `LModule _) as id ->
+let lookup_module map : Ident.module_ -> _ = function
+  | `LModule _ as id ->
       (Component.ModuleMap.find id map.module_ :> Identifier.Path.Module.t)
-  | #Ident.functor_parameter as id ->
-      (List.assoc id map.functor_parameter :> Identifier.Path.Module.t)
-  | _ -> raise Not_found
 
 module Opt = Component.Opt
 
 module Path = struct
   let rec module_ map (p : Cpath.module_) : Odoc_model.Paths.Path.Module.t =
     match p with
-    | `Substituted x -> module_ map x
+    | `Substituted x -> `Substituted (module_ map x)
     | `Local (id, b) ->
         let m =
           try lookup_module map id
@@ -70,7 +66,7 @@ module Path = struct
           b
           ||
           match m.iv with
-          | `Module (_, n) -> Odoc_model.Names.ModuleName.is_internal n
+          | `Module (_, n) -> Odoc_model.Names.ModuleName.is_hidden n
           | _ -> false
         in
         `Identifier (m, hidden)
@@ -80,14 +76,13 @@ module Path = struct
     | `Dot (p, s) -> `Dot (module_ map p, s)
     | `Forward s -> `Forward s
     | `Apply (m1, m2) -> `Apply (module_ map m1, module_ map m2)
-    | `Module (`Module p, n) ->
-        `Dot (`Resolved (resolved_module map p), ModuleName.to_string n)
+    | `Module (`Module p, n) -> `Dot (`Resolved (resolved_module map p), n)
     | `Module (_, _) -> failwith "Probably shouldn't happen"
 
   and module_type map (p : Cpath.module_type) :
       Odoc_model.Paths.Path.ModuleType.t =
     match p with
-    | `Substituted x -> module_type map x
+    | `Substituted x -> `SubstitutedMT (module_type map x)
     | `Identifier
         (({ iv = #Odoc_model.Paths.Identifier.ModuleType.t_pv; _ } as y), b) ->
         `Identifier (y, b)
@@ -98,45 +93,38 @@ module Path = struct
                failwith (Format.asprintf "Not_found: %a" Ident.fmt id)),
             b )
     | `Resolved x -> `Resolved (resolved_module_type map x)
-    | `Dot (p, n) -> `Dot (module_ map p, n)
-    | `ModuleType (`Module p, n) ->
-        `Dot (`Resolved (resolved_module map p), ModuleTypeName.to_string n)
+    | `DotMT (p, n) -> `DotMT (module_ map p, n)
+    | `ModuleType (`Module p, n) -> `DotMT (`Resolved (resolved_module map p), n)
     | `ModuleType (_, _) -> failwith "Probably shouldn't happen"
 
   and type_ map (p : Cpath.type_) : Odoc_model.Paths.Path.Type.t =
     match p with
-    | `Substituted x -> type_ map x
+    | `Substituted x -> `SubstitutedT (type_ map x)
     | `Identifier
         (({ iv = #Odoc_model.Paths.Identifier.Path.Type.t_pv; _ } as y), b) ->
         `Identifier (y, b)
-    | `Local (id, b) ->
-        `Identifier (Component.PathTypeMap.find id map.path_type, b)
+    | `Local (id, b) -> `Identifier (Component.TypeMap.find id map.path_type, b)
     | `Resolved x -> `Resolved (resolved_type map x)
-    | `Dot (p, n) -> `Dot (module_ map p, n)
-    | `Type (`Module p, n) ->
-        `Dot (`Resolved (resolved_module map p), TypeName.to_string n)
-    | `Class (`Module p, n) ->
-        `Dot (`Resolved (resolved_module map p), ClassName.to_string n)
-    | `ClassType (`Module p, n) ->
-        `Dot (`Resolved (resolved_module map p), ClassTypeName.to_string n)
+    | `DotT (p, n) -> `DotT (module_ map p, n)
+    | `Type (`Module p, n) -> `DotT (`Resolved (resolved_module map p), n)
+    | `Class (`Module p, n) -> `DotT (`Resolved (resolved_module map p), n)
+    | `ClassType (`Module p, n) -> `DotT (`Resolved (resolved_module map p), n)
     | `Type _ | `Class _ | `ClassType _ -> failwith "Probably shouldn't happen"
 
   and class_type map (p : Cpath.class_type) : Odoc_model.Paths.Path.ClassType.t
       =
     match p with
-    | `Substituted x -> class_type map x
+    | `Substituted x -> `SubstitutedCT (class_type map x)
     | `Identifier
         (({ iv = #Odoc_model.Paths.Identifier.Path.ClassType.t_pv; _ } as y), b)
       ->
         `Identifier (y, b)
     | `Local (id, b) ->
-        `Identifier (Component.PathClassTypeMap.find id map.path_class_type, b)
+        `Identifier (Component.TypeMap.find id map.path_class_type, b)
     | `Resolved x -> `Resolved (resolved_class_type map x)
-    | `Dot (p, n) -> `Dot (module_ map p, n)
-    | `Class (`Module p, n) ->
-        `Dot (`Resolved (resolved_module map p), ClassName.to_string n)
-    | `ClassType (`Module p, n) ->
-        `Dot (`Resolved (resolved_module map p), ClassTypeName.to_string n)
+    | `DotT (p, n) -> `DotT (module_ map p, n)
+    | `Class (`Module p, n) -> `DotT (`Resolved (resolved_module map p), n)
+    | `ClassType (`Module p, n) -> `DotT (`Resolved (resolved_module map p), n)
     | `Class _ | `ClassType _ -> failwith "Probably shouldn't happen"
 
   and resolved_module map (p : Cpath.Resolved.module_) :
@@ -147,7 +135,7 @@ module Path = struct
           (try lookup_module map id
            with Not_found ->
              failwith (Format.asprintf "Not_found: %a" Ident.fmt id))
-    | `Substituted x -> resolved_module map x
+    | `Substituted x -> `Substituted (resolved_module map x)
     | `Gpath y -> y
     | `Subst (mty, m) ->
         `Subst (resolved_module_type map mty, resolved_module map m)
@@ -177,7 +165,7 @@ module Path = struct
            with Not_found ->
              failwith (Format.asprintf "Not_found: %a" Ident.fmt id))
     | `ModuleType (p, name) -> `ModuleType (resolved_parent map p, name)
-    | `Substituted s -> resolved_module_type map s
+    | `Substituted s -> `SubstitutedMT (resolved_module_type map s)
     | `SubstT (p1, p2) ->
         `SubstT (resolved_module_type map p1, resolved_module_type map p2)
     | `AliasModuleType (p1, p2) ->
@@ -190,42 +178,29 @@ module Path = struct
   and resolved_type map (p : Cpath.Resolved.type_) :
       Odoc_model.Paths.Path.Resolved.Type.t =
     match p with
+    | `CoreType _ as c -> c
     | `Gpath y -> y
-    | `Local id -> `Identifier (Component.PathTypeMap.find id map.path_type)
+    | `Local id -> `Identifier (Component.TypeMap.find id map.path_type)
     | `CanonicalType (t1, t2) -> `CanonicalType (resolved_type map t1, t2)
     | `Type (p, name) -> `Type (resolved_parent map p, name)
     | `Class (p, name) -> `Class (resolved_parent map p, name)
     | `ClassType (p, name) -> `ClassType (resolved_parent map p, name)
-    | `Substituted s -> resolved_type map s
+    | `Substituted s -> `SubstitutedT (resolved_type map s)
 
-  and resolved_value map (`Value (p, name) : Cpath.Resolved.value) :
+  and resolved_value map (p : Cpath.Resolved.value) :
       Odoc_model.Paths.Path.Resolved.Value.t =
-    `Value (resolved_parent map p, name)
-
-  and resolved_datatype map (p : Cpath.Resolved.datatype) :
-      Odoc_model.Paths.Path.Resolved.DataType.t =
     match p with
+    | `Value (p, name) -> `Value (resolved_parent map p, name)
     | `Gpath y -> y
-    | `Local id -> `Identifier (Component.TypeMap.find id map.type_)
-    | `CanonicalDataType (t1, t2) ->
-        `CanonicalDataType (resolved_datatype map t1, t2)
-    | `Type (p, name) -> `Type (resolved_parent map p, name)
-    | `Substituted s -> resolved_datatype map s
-
-  and resolved_constructor map
-      (`Constructor (p, name) : Cpath.Resolved.constructor) :
-      Odoc_model.Paths.Path.Resolved.Constructor.t =
-    `Constructor (resolved_datatype map p, name)
 
   and resolved_class_type map (p : Cpath.Resolved.class_type) :
       Odoc_model.Paths.Path.Resolved.ClassType.t =
     match p with
     | `Gpath y -> y
-    | `Local id ->
-        `Identifier (Component.PathClassTypeMap.find id map.path_class_type)
+    | `Local id -> `Identifier (Component.TypeMap.find id map.path_class_type)
     | `Class (p, name) -> `Class (resolved_parent map p, name)
     | `ClassType (p, name) -> `ClassType (resolved_parent map p, name)
-    | `Substituted s -> resolved_class_type map s
+    | `Substituted s -> `SubstitutedCT (resolved_class_type map s)
 
   let rec module_fragment :
       maps -> Cfrag.module_ -> Odoc_model.Paths.Fragment.Module.t =
@@ -301,10 +276,10 @@ module ExtractIDs = struct
   open Component
 
   let rec type_decl parent map id =
-    let name = Ident.Name.unsafe_type id in
+    let name = Ident.Name.type_ id in
     let typed_name =
-      if List.mem name map.shadowed.s_types then
-        Odoc_model.Names.TypeName.internal_of_string name
+      if List.mem_assoc name map.shadowed.s_types then
+        List.assoc name map.shadowed.s_types
       else Ident.Name.typed_type id
     in
     let identifier = Identifier.Mk.type_ (parent, typed_name) in
@@ -312,27 +287,27 @@ module ExtractIDs = struct
       map with
       type_ = Component.TypeMap.add id identifier map.type_;
       path_type =
-        Component.PathTypeMap.add
-          (id :> Ident.path_type)
+        Component.TypeMap.add
+          (id :> Ident.type_)
           (identifier :> Identifier.Path.Type.t)
           map.path_type;
     }
 
   and module_ parent map id =
-    let name = Ident.Name.unsafe_module id in
+    let name = Ident.Name.module_ id in
     let typed_name =
-      if List.mem name map.shadowed.s_modules then
-        ModuleName.internal_of_string name
+      if List.mem_assoc name map.shadowed.s_modules then
+        List.assoc name map.shadowed.s_modules
       else Ident.Name.typed_module id
     in
     let identifier = Identifier.Mk.module_ (parent, typed_name) in
     { map with module_ = Component.ModuleMap.add id identifier map.module_ }
 
   and module_type parent map id =
-    let name = Ident.Name.unsafe_module_type id in
+    let name = Ident.Name.module_type id in
     let typed_name =
-      if List.mem name map.shadowed.s_module_types then
-        ModuleTypeName.internal_of_string name
+      if List.mem_assoc name map.shadowed.s_module_types then
+        List.assoc name map.shadowed.s_module_types
       else Ident.Name.typed_module_type id
     in
     let identifier = Identifier.Mk.module_type (parent, typed_name) in
@@ -342,47 +317,47 @@ module ExtractIDs = struct
     }
 
   and class_ parent map id =
-    let name = Ident.Name.unsafe_class id in
+    let name = Ident.Name.type_ id in
     let typed_name =
-      if List.mem name map.shadowed.s_classes then
-        ClassName.internal_of_string name
-      else Ident.Name.typed_class id
+      if List.mem_assoc name map.shadowed.s_classes then
+        List.assoc name map.shadowed.s_classes
+      else Ident.Name.typed_type id
     in
     let identifier = Identifier.Mk.class_ (parent, typed_name) in
     {
       map with
       class_ = (id, identifier) :: map.class_;
       path_class_type =
-        Component.PathClassTypeMap.add
-          (id :> Ident.path_class_type)
+        Component.TypeMap.add
+          (id :> Ident.type_)
           (identifier :> Identifier.Path.ClassType.t)
           map.path_class_type;
       path_type =
-        Component.PathTypeMap.add
-          (id :> Ident.path_type)
+        Component.TypeMap.add
+          (id :> Ident.type_)
           (identifier :> Identifier.Path.Type.t)
           map.path_type;
     }
 
-  and class_type parent map (id : Ident.class_type) =
-    let name = Ident.Name.unsafe_class_type id in
+  and class_type parent map (id : Ident.type_) =
+    let name = Ident.Name.type_ id in
     let typed_name =
-      if List.mem name map.shadowed.s_class_types then
-        ClassTypeName.internal_of_string name
-      else Ident.Name.typed_class_type id
+      if List.mem_assoc name map.shadowed.s_class_types then
+        List.assoc name map.shadowed.s_class_types
+      else Ident.Name.typed_type id
     in
     let identifier = Identifier.Mk.class_type (parent, typed_name) in
     {
       map with
-      class_type = ((id :> Ident.class_type), identifier) :: map.class_type;
+      class_type = ((id :> Ident.type_), identifier) :: map.class_type;
       path_class_type =
-        Component.PathClassTypeMap.add
-          (id :> Ident.path_class_type)
+        Component.TypeMap.add
+          (id :> Ident.type_)
           (identifier :> Identifier.Path.ClassType.t)
           map.path_class_type;
       path_type =
-        Component.PathTypeMap.add
-          (id :> Ident.path_type)
+        Component.TypeMap.add
+          (id :> Ident.type_)
           (identifier :> Identifier.Path.Type.t)
           map.path_type;
     }
@@ -423,7 +398,6 @@ end
 
 let rec signature_items id map items =
   let open Component.Signature in
-  let map = ExtractIDs.signature_items id map items in
   let parent = id in
   let rec inner : item list -> Odoc_model.Lang.Signature.item list -> _ =
    fun items acc ->
@@ -480,13 +454,30 @@ and signature :
     Lang.Signature.t =
  fun id map sg ->
   let open Component.Signature in
-  (* let map = { map with shadowed = empty_shadow } in *)
+  let map = ExtractIDs.signature_items id map sg.items in
+  let removed = List.map (removed_item map id) sg.removed in
   {
     items = signature_items id map sg.items;
     compiled = sg.compiled;
+    removed;
     doc = docs (id :> Identifier.LabelParent.t) sg.doc;
   }
 
+and removed_item :
+    maps ->
+    Identifier.Id.signature ->
+    Component.Signature.removed_item ->
+    Lang.Signature.removed_item =
+ fun map parent item ->
+  match item with
+  | RModule (id, m) -> RModule (id, Path.module_ map m)
+  | RType (id, texpr, eqn) ->
+      RType
+        ( id,
+          type_expr map (parent :> Identifier.LabelParent.t) texpr,
+          type_decl_equation map (parent :> Identifier.FieldParent.t) eqn )
+  | RModuleType (id, m) -> RModuleType (id, module_type_expr map parent m)
+
 and class_ map parent id c =
   let open Component.Class in
   let identifier = List.assoc id map.class_ in
@@ -497,7 +488,7 @@ and class_ map parent id c =
   in
   {
     id = identifier;
-    locs = c.locs;
+    source_loc = c.source_loc;
     doc = docs (parent :> Identifier.LabelParent.t) c.doc;
     virtual_ = c.virtual_;
     params = c.params;
@@ -535,7 +526,7 @@ and class_type map parent id c =
   in
   {
     Odoc_model.Lang.ClassType.id = identifier;
-    locs = c.locs;
+    source_loc = c.source_loc;
     doc = docs (parent :> Identifier.LabelParent.t) c.doc;
     virtual_ = c.virtual_;
     params = c.params;
@@ -614,13 +605,14 @@ and simple_expansion :
   | Signature sg -> Signature (signature id map sg)
   | Functor (Named arg, sg) ->
       let identifier = Identifier.Mk.result id in
-      let name = Ident.Name.typed_functor_parameter arg.id in
+      let name = Ident.Name.typed_module arg.id in
       let param_identifier = Identifier.Mk.parameter (id, name) in
       let map =
         {
           map with
           functor_parameter =
             (arg.id, param_identifier) :: map.functor_parameter;
+          module_ = Component.ModuleMap.add arg.id param_identifier map.module_;
         }
       in
       let arg = functor_parameter map arg in
@@ -630,13 +622,21 @@ and simple_expansion :
 
 and combine_shadowed s1 s2 =
   let open Odoc_model.Lang.Include in
+  (* If something was already shadowed in the include, it mustn't be
+     added to the combined map. *)
+  let combine s1 s2 =
+    List.fold_left
+      (fun acc (name, typed_name) ->
+        if List.mem_assoc name acc then acc else (name, typed_name) :: acc)
+      s2 s1
+  in
   {
-    s_modules = s1.s_modules @ s2.s_modules;
-    s_module_types = s1.s_module_types @ s2.s_module_types;
-    s_values = s1.s_values @ s2.s_values;
-    s_types = s1.s_types @ s2.s_types;
-    s_classes = s1.s_classes @ s2.s_classes;
-    s_class_types = s1.s_class_types @ s2.s_class_types;
+    s_modules = combine s1.s_modules s2.s_modules;
+    s_module_types = combine s1.s_module_types s2.s_module_types;
+    s_values = combine s1.s_values s2.s_values;
+    s_types = combine s1.s_types s2.s_types;
+    s_classes = combine s1.s_classes s2.s_classes;
+    s_class_types = combine s1.s_class_types s2.s_class_types;
   }
 
 and include_decl :
@@ -645,6 +645,8 @@ and include_decl :
     Component.Include.decl ->
     Odoc_model.Lang.Include.decl =
  fun map identifier d ->
+  let map = { map with shadowed = empty_shadow } in
+  (* Don't start shadowing within any signatures *)
   match d with
   | Alias p -> Alias (Path.module_ map p)
   | ModuleType mty -> ModuleType (u_module_type_expr map identifier mty)
@@ -677,14 +679,14 @@ and value_ map parent id v =
   let open Component.Value in
   let name = Ident.Name.value id in
   let typed_name =
-    if List.mem name map.shadowed.s_values then
-      ValueName.internal_of_string name
+    if List.mem_assoc name map.shadowed.s_values then
+      List.assoc name map.shadowed.s_values
     else Ident.Name.typed_value id
   in
   let identifier = Identifier.Mk.value (parent, typed_name) in
   {
     id = identifier;
-    locs = v.locs;
+    source_loc = v.source_loc;
     doc = docs (parent :> Identifier.LabelParent.t) v.doc;
     type_ = type_expr map (parent :> Identifier.LabelParent.t) v.type_;
     value = v.value;
@@ -694,7 +696,7 @@ and typ_ext map parent t =
   let open Component.Extension in
   {
     parent;
-    type_path = Path.type_ map t.type_path;
+    type_path = (Path.type_ map t.type_path :> Paths.Path.Type.t);
     doc = docs (parent :> Identifier.LabelParent.t) t.doc;
     type_params = t.type_params;
     private_ = t.private_;
@@ -708,7 +710,7 @@ and extension_constructor map parent c =
   in
   {
     id = identifier;
-    locs = c.locs;
+    source_loc = c.source_loc;
     doc = docs (parent :> Identifier.LabelParent.t) c.doc;
     args =
       type_decl_constructor_argument map
@@ -727,7 +729,7 @@ and module_ map parent id m =
     let map = { map with shadowed = empty_shadow } in
     {
       Odoc_model.Lang.Module.id;
-      locs = m.locs;
+      source_loc = m.source_loc;
       doc = docs (parent :> Identifier.LabelParent.t) m.doc;
       type_ = module_decl map identifier m.type_;
       canonical = m.canonical;
@@ -793,18 +795,10 @@ and u_module_type_expr map identifier = function
       With
         ( List.map (mty_substitution map identifier) subs,
           u_module_type_expr map identifier expr )
-  | TypeOf { t_desc = ModPath p; t_expansion } ->
-      TypeOf
-        {
-          t_desc = ModPath (Path.module_ map p);
-          t_expansion = Opt.map (simple_expansion map identifier) t_expansion;
-        }
-  | TypeOf { t_desc = StructInclude p; t_expansion } ->
-      TypeOf
-        {
-          t_desc = StructInclude (Path.module_ map p);
-          t_expansion = Opt.map (simple_expansion map identifier) t_expansion;
-        }
+  | TypeOf (ModPath p, original_path) ->
+      TypeOf (ModPath (Path.module_ map p), Path.module_ map original_path)
+  | TypeOf (StructInclude p, original_path) ->
+      TypeOf (StructInclude (Path.module_ map p), Path.module_ map original_path)
 
 and module_type_expr map identifier = function
   | Component.ModuleType.Path { p_path; p_expansion } ->
@@ -827,12 +821,13 @@ and module_type_expr map identifier = function
           w_expr = u_module_type_expr map identifier w_expr;
         }
   | Functor (Named arg, expr) ->
-      let name = Ident.Name.typed_functor_parameter arg.id in
+      let name = Ident.Name.typed_module arg.id in
       let identifier' = Identifier.Mk.parameter (identifier, name) in
       let map =
         {
           map with
           functor_parameter = (arg.id, identifier') :: map.functor_parameter;
+          module_ = Component.ModuleMap.add arg.id identifier' map.module_;
         }
       in
       Functor
@@ -840,16 +835,18 @@ and module_type_expr map identifier = function
           module_type_expr map (Identifier.Mk.result identifier) expr )
   | Functor (Unit, expr) ->
       Functor (Unit, module_type_expr map (Identifier.Mk.result identifier) expr)
-  | TypeOf { t_desc = ModPath p; t_expansion } ->
+  | TypeOf { t_desc = ModPath p; t_original_path; t_expansion } ->
       TypeOf
         {
           t_desc = ModPath (Path.module_ map p);
+          t_original_path = Path.module_ map t_original_path;
           t_expansion = Opt.map (simple_expansion map identifier) t_expansion;
         }
-  | TypeOf { t_desc = StructInclude p; t_expansion } ->
+  | TypeOf { t_desc = StructInclude p; t_original_path; t_expansion } ->
       TypeOf
         {
           t_desc = StructInclude (Path.module_ map p);
+          t_original_path = Path.module_ map t_original_path;
           t_expansion = Opt.map (simple_expansion map identifier) t_expansion;
         }
 
@@ -866,7 +863,7 @@ and module_type :
   let map = { map with shadowed = empty_shadow } in
   {
     Odoc_model.Lang.ModuleType.id = identifier;
-    locs = mty.locs;
+    source_loc = mty.source_loc;
     doc = docs (parent :> Identifier.LabelParent.t) mty.doc;
     canonical = mty.canonical;
     expr = Opt.map (module_type_expr map sig_id) mty.expr;
@@ -934,7 +931,7 @@ and type_decl map parent id (t : Component.TypeDecl.t) :
   let identifier = Component.TypeMap.find id map.type_ in
   {
     id = identifier;
-    locs = t.locs;
+    source_loc = t.source_loc;
     equation =
       type_decl_equation map (parent :> Identifier.FieldParent.t) t.equation;
     doc = docs (parent :> Identifier.LabelParent.t) t.doc;
@@ -995,7 +992,9 @@ and type_expr map (parent : Identifier.LabelParent.t) (t : Component.TypeExpr.t)
         Arrow (lbl, type_expr map parent t1, type_expr map parent t2)
     | Tuple ts -> Tuple (List.map (type_expr map parent) ts)
     | Constr (path, ts) ->
-        Constr (Path.type_ map path, List.map (type_expr map parent) ts)
+        Constr
+          ( (Path.type_ map path :> Paths.Path.Type.t),
+            List.map (type_expr map parent) ts )
     | Polymorphic_variant v ->
         Polymorphic_variant (type_expr_polyvar map parent v)
     | Object o -> Object (type_expr_object map parent o)
@@ -1007,7 +1006,9 @@ and type_expr map (parent : Identifier.LabelParent.t) (t : Component.TypeExpr.t)
     let bt = Printexc.get_backtrace () in
     Format.fprintf Format.err_formatter
       "Exception %s handling type_expr: %a\nbacktrace:\n%s\n%!"
-      (Printexc.to_string e) Component.Fmt.type_expr t bt;
+      (Printexc.to_string e)
+      Component.Fmt.(type_expr default)
+      t bt;
     raise e
 
 and type_expr_polyvar map parent v =
@@ -1058,7 +1059,7 @@ and exception_ map parent id (e : Component.Exception.t) :
   in
   {
     id = identifier;
-    locs = e.locs;
+    source_loc = e.source_loc;
     doc = docs (parent :> Identifier.LabelParent.t) e.doc;
     args =
       type_decl_constructor_argument map
@@ -1082,7 +1083,7 @@ and block_element parent
             raise Not_found
         in
         `Heading (attrs, label, text)
-    | `Tag t -> `Tag t
+    | (`Tag _ | `Media _) as orig -> orig
     | #Odoc_model.Comment.nestable_block_element as n -> n
   in
   { d with Odoc_model.Location_.value }
@@ -1091,7 +1092,12 @@ and docs :
     Identifier.LabelParent.t ->
     Component.CComment.docs ->
     Odoc_model.Comment.docs =
- fun parent ds -> List.rev_map (fun d -> block_element parent d) ds |> List.rev
+ fun parent ds ->
+  {
+    elements =
+      List.rev_map (fun d -> block_element parent d) ds.elements |> List.rev;
+    warnings_tag = ds.warnings_tag;
+  }
 
 and docs_or_stop parent (d : Component.CComment.docs_or_stop) =
   match d with `Docs d -> `Docs (docs parent d) | `Stop -> `Stop
