@@ -7,31 +7,36 @@ module Opt = struct
   let map f = function Some x -> Some (f x) | None -> None
 end
 
-let locations env id locs =
+(* omg. Our current warning system is spread on different system. Hence this
+   atrocity. *)
+let maybe_suppress env warnings_tag =
+  if Env.should_suppress_warnings env warnings_tag then fun f ->
+    Lookup_failures.catch_failures ~filename:"" (fun () ->
+        Error.catch_warnings f |> fun x ->
+        Error.unpack_warnings x |> fst |> Error.unpack_warnings |> fst)
+    |> Error.unpack_warnings |> fst
+  else fun f -> f () |> Error.raise_warnings
+
+let source_loc env id loc =
   let id = (id :> Id.NonSrc.t) in
-  match locs with
-  | Some _ as locs -> locs
-  | None -> Shape_tools.lookup_def env id
+  match loc with Some _ as loc -> loc | None -> Shape_tools.lookup_def env id
 
 (** Equivalent to {!Comment.synopsis}. *)
 let synopsis_from_comment (docs : Component.CComment.docs) =
-  match docs with
+  match docs.elements with
   | ({ value = #Comment.nestable_block_element; _ } as e) :: _ ->
       (* Only the first element is considered. *)
       Comment.synopsis [ e ]
   | _ -> None
 
 let synopsis_of_module env (m : Component.Module.t) =
-  let open Utils.ResultMonad in
+  let open Odoc_utils.ResultMonad in
   match synopsis_from_comment m.doc with
   | Some _ as s -> s
   | None -> (
       let rec handle_expansion : Tools.expansion -> _ = function
         | Functor (_, expr) -> (
-            match
-              Tools.expansion_of_module_type_expr ~mark_substituted:true env
-                expr
-            with
+            match Tools.expansion_of_module_type_expr env expr with
             | Ok e -> handle_expansion e
             | Error _ as e -> e)
         | Signature sg -> Ok sg
@@ -79,8 +84,9 @@ let expansion_needed self target =
   let hidden_alias = Paths.Path.Resolved.is_hidden self
   and self_canonical =
     let i = Paths.Path.Resolved.identifier self in
-    i = (target :> Paths.Identifier.t)
+    i = Some (target :> Paths.Identifier.t)
   in
+
   self_canonical || hidden_alias
 
 exception Loop
@@ -92,11 +98,13 @@ let rec is_forward : Paths.Path.Module.t -> bool = function
   | `Identifier _ -> false
   | `Dot (p, _) -> is_forward p
   | `Apply (p1, p2) -> is_forward p1 || is_forward p2
+  | `Substituted s -> is_forward s
 
 let rec should_reresolve : Paths.Path.Resolved.t -> bool =
  fun p ->
   let open Paths.Path.Resolved in
   match p with
+  | `CoreType _ -> false
   | `Identifier _ -> false
   | `Subst (x, y) -> should_reresolve (x :> t) || should_reresolve (y :> t)
   | `Hidden p -> should_reresolve (p :> t)
@@ -106,8 +114,6 @@ let rec should_reresolve : Paths.Path.Resolved.t -> bool =
       should_reresolve (x :> t) || should_resolve (y :> Paths.Path.t)
   | `CanonicalType (x, y) ->
       should_reresolve (x :> t) || should_resolve (y :> Paths.Path.t)
-  | `CanonicalDataType (x, y) ->
-      should_reresolve (x :> t) || should_resolve (y :> Paths.Path.t)
   | `Apply (x, y) ->
       should_reresolve (x :> t) || should_reresolve (y :> Paths.Path.Resolved.t)
   | `SubstT (x, y) -> should_reresolve (x :> t) || should_reresolve (y :> t)
@@ -122,19 +128,16 @@ let rec should_reresolve : Paths.Path.Resolved.t -> bool =
   | `ModuleType (p, _)
   | `Module (p, _) ->
       should_reresolve (p :> t)
-  | `Constructor (p, _) -> should_reresolve (p :> t)
   | `OpaqueModule m -> should_reresolve (m :> t)
   | `OpaqueModuleType m -> should_reresolve (m :> t)
+  | `Substituted m -> should_reresolve (m :> t)
+  | `SubstitutedMT m -> should_reresolve (m :> t)
+  | `SubstitutedT m -> should_reresolve (m :> t)
+  | `SubstitutedCT m -> should_reresolve (m :> t)
 
 and should_resolve : Paths.Path.t -> bool =
  fun p -> match p with `Resolved p -> should_reresolve p | _ -> true
 
-(* and should_resolve_constructor : Paths.Path.Constructor.t -> bool = *)
-(*  fun p -> *)
-(*   match p with *)
-(*   | `Resolved p -> should_reresolve (p :> Paths.Path.Resolved.t) *)
-(*   | _ -> true *)
-
 let type_path : Env.t -> Paths.Path.Type.t -> Paths.Path.Type.t =
  fun env p ->
   if not (should_resolve (p :> Paths.Path.t)) then p
@@ -153,42 +156,23 @@ let type_path : Env.t -> Paths.Path.Type.t -> Paths.Path.Type.t =
             Errors.report ~what:(`Type_path cp) ~tools_error:e `Lookup;
             p)
 
-(* let value_path : Env.t -> Paths.Path.Value.t -> Paths.Path.Value.t = *)
-(*  fun env p -> *)
-(*   if not (should_resolve (p :> Paths.Path.t)) then p *)
-(*   else *)
-(*     let cp = Component.Of_Lang.(value_path (empty ()) p) in *)
-(*     match cp with *)
-(*     | `Resolved p -> *)
-(*         let result = Tools.reresolve_value env p in *)
-(*         `Resolved Lang_of.(Path.resolved_value (empty ()) result) *)
-(*     | _ -> ( *)
-(*         match Tools.resolve_value_path env cp with *)
-(*         | Ok p' -> *)
-(*             let result = Tools.reresolve_value env p' in *)
-(*             `Resolved Lang_of.(Path.resolved_value (empty ()) result) *)
-(*         | Error e -> *)
-(*             Errors.report ~what:(`Value_path cp) ~tools_error:e `Lookup; *)
-(*             p) *)
-
-(* let constructor_path : *)
-(*     Env.t -> Paths.Path.Constructor.t -> Paths.Path.Constructor.t = *)
-(*  fun env p -> *)
-(*   if not (should_resolve_constructor p) then p *)
-(*   else *)
-(*     let cp = Component.Of_Lang.(constructor_path (empty ()) p) in *)
-(*     match cp with *)
-(*     | `Resolved p -> *)
-(*         let result = Tools.reresolve_constructor env p in *)
-(*         `Resolved Lang_of.(Path.resolved_constructor (empty ()) result) *)
-(*     | _ -> ( *)
-(*         match Tools.resolve_constructor_path env cp with *)
-(*         | Ok p' -> *)
-(*             let result = Tools.reresolve_constructor env p' in *)
-(*             `Resolved Lang_of.(Path.resolved_constructor (empty ()) result) *)
-(*         | Error e -> *)
-(*             Errors.report ~what:(`Constructor_path cp) ~tools_error:e `Lookup; *)
-(*             p) *)
+let value_path : Env.t -> Paths.Path.Value.t -> Paths.Path.Value.t =
+ fun env p ->
+  if not (should_resolve (p :> Paths.Path.t)) then p
+  else
+    let cp = Component.Of_Lang.(value_path (empty ()) p) in
+    match cp with
+    | `Resolved p ->
+        let result = Tools.reresolve_value env p in
+        `Resolved Lang_of.(Path.resolved_value (empty ()) result)
+    | _ -> (
+        match Tools.resolve_value_path env cp with
+        | Ok p' ->
+            let result = Tools.reresolve_value env p' in
+            `Resolved Lang_of.(Path.resolved_value (empty ()) result)
+        | Error e ->
+            Errors.report ~what:(`Value_path cp) ~tools_error:e `Lookup;
+            p)
 
 let class_type_path : Env.t -> Paths.Path.ClassType.t -> Paths.Path.ClassType.t
     =
@@ -248,56 +232,66 @@ and module_path : Env.t -> Paths.Path.Module.t -> Paths.Path.Module.t =
             p)
 
 let rec comment_inline_element :
-    loc:_ -> Env.t -> Comment.inline_element -> Comment.inline_element =
- fun ~loc:_ env x ->
+    loc:_ ->
+    Env.t ->
+    string option ->
+    Comment.inline_element ->
+    Comment.inline_element =
+ fun ~loc:_ env warnings_tag x ->
   match x with
   | `Styled (s, ls) ->
-      `Styled (s, List.map (with_location (comment_inline_element env)) ls)
+      `Styled
+        ( s,
+          List.map (with_location (comment_inline_element env warnings_tag)) ls
+        )
   | `Reference (r, content) as orig -> (
-      match Ref_tools.resolve_reference env r |> Error.raise_warnings with
-      | Ok x ->
+      match
+        maybe_suppress env warnings_tag (fun () ->
+            Ref_tools.resolve_reference env r)
+      with
+      | Ok (ref_, c) ->
           let content =
             (* In case of labels, use the heading text as reference text if
                it's not specified. *)
-            match (content, x) with
-            | [], `Identifier ({ iv = #Id.Label.t_pv; _ } as i) -> (
-                match Env.lookup_by_id Env.s_label i env with
-                | Some (`Label (_, lbl)) ->
-                    Odoc_model.Comment.link_content_of_inline_elements
-                      lbl.Component.Label.text
-                | None -> [])
+            match (content, c) with
+            | [], Some content ->
+                Comment.link_content_of_inline_elements content
             | content, _ -> content
           in
-          `Reference (`Resolved x, content)
+          `Reference (`Resolved ref_, content)
       | Error e ->
-          Errors.report ~what:(`Reference r) ~tools_error:(`Reference e)
-            `Resolve;
+          if not (Env.should_suppress_warnings env warnings_tag) then
+            Errors.report ~what:(`Reference r) ~tools_error:(`Reference e)
+              `Resolve;
           orig)
   | y -> y
 
-and paragraph env elts =
-  List.map (with_location (comment_inline_element env)) elts
+and paragraph env warnings_tag elts =
+  List.map (with_location (comment_inline_element env warnings_tag)) elts
 
-and resolve_external_synopsis env synopsis =
+and resolve_external_synopsis env warnings_tag synopsis =
   let env = Env.inherit_resolver env in
-  paragraph env synopsis
+  paragraph env warnings_tag synopsis
 
-and comment_nestable_block_element env parent ~loc:_
+and comment_nestable_block_element env warnings_tag parent ~loc:_
     (x : Comment.nestable_block_element) =
   match x with
-  | `Paragraph elts -> `Paragraph (paragraph env elts)
+  | `Paragraph elts -> `Paragraph (paragraph env warnings_tag elts)
   | (`Code_block _ | `Math_block _ | `Verbatim _) as x -> x
   | `List (x, ys) ->
       `List
         ( x,
-          List.rev_map (comment_nestable_block_element_list env parent) ys
+          List.rev_map
+            (comment_nestable_block_element_list env warnings_tag parent)
+            ys
           |> List.rev )
   | `Table { data; align } ->
       let data =
         let map f x = List.rev_map f x |> List.rev in
         map
           (map (fun (cell, cell_type) ->
-               (comment_nestable_block_element_list env parent cell, cell_type)))
+               ( comment_nestable_block_element_list env warnings_tag parent cell,
+                 cell_type )))
           data
       in
       `Table { Comment.data; align }
@@ -306,75 +300,110 @@ and comment_nestable_block_element env parent ~loc:_
         List.rev_map
           (fun (r : Comment.module_reference) ->
             match
-              Ref_tools.resolve_module_reference env r.module_reference
-              |> Error.raise_warnings
+              maybe_suppress env warnings_tag (fun () ->
+                  Ref_tools.resolve_module_reference env r.module_reference)
             with
             | Ok (r, _, m) ->
                 let module_synopsis =
                   Opt.map
-                    (resolve_external_synopsis env)
+                    (resolve_external_synopsis env warnings_tag)
                     (synopsis_of_module env m)
                 in
                 { Comment.module_reference = `Resolved r; module_synopsis }
             | Error e ->
-                Errors.report
-                  ~what:(`Reference (r.module_reference :> Paths.Reference.t))
-                  ~tools_error:(`Reference e) `Resolve;
+                if not (Env.should_suppress_warnings env warnings_tag) then
+                  Errors.report
+                    ~what:(`Reference (r.module_reference :> Paths.Reference.t))
+                    ~tools_error:(`Reference e) `Resolve;
                 r)
           refs
         |> List.rev
       in
       `Modules refs
+  | `Media (`Reference r, m, content) as orig -> (
+      match
+        maybe_suppress env warnings_tag (fun () ->
+            Ref_tools.resolve_asset_reference env r)
+      with
+      | Ok x -> `Media (`Reference (`Resolved x), m, content)
+      | Error e ->
+          if not (Env.should_suppress_warnings env warnings_tag) then
+            Errors.report
+              ~what:(`Reference (r :> Paths.Reference.t))
+              ~tools_error:(`Reference e) `Resolve;
+          orig)
+  | `Media _ as orig -> orig
 
-and comment_nestable_block_element_list env parent
+and comment_nestable_block_element_list env warnings_tag parent
     (xs : Comment.nestable_block_element Comment.with_location list) =
-  List.rev_map (with_location (comment_nestable_block_element env parent)) xs
+  List.rev_map
+    (with_location (comment_nestable_block_element env warnings_tag parent))
+    xs
   |> List.rev
 
-and comment_tag env parent ~loc:_ (x : Comment.tag) =
+and comment_tag env warnings_tag parent ~loc:_ (x : Comment.tag) =
   match x with
   | `Deprecated content ->
-      `Deprecated (comment_nestable_block_element_list env parent content)
+      `Deprecated
+        (comment_nestable_block_element_list env warnings_tag parent content)
   | `Param (name, content) ->
-      `Param (name, comment_nestable_block_element_list env parent content)
+      `Param
+        ( name,
+          comment_nestable_block_element_list env warnings_tag parent content )
   | `Raise ((`Reference (r, reference_content) as orig), content) -> (
-      match Ref_tools.resolve_reference env r |> Error.raise_warnings with
-      | Ok x ->
+      match
+        maybe_suppress env warnings_tag (fun () ->
+            Ref_tools.resolve_reference env r)
+      with
+      | Ok (x, _) ->
           `Raise
             ( `Reference (`Resolved x, reference_content),
-              comment_nestable_block_element_list env parent content )
+              comment_nestable_block_element_list env warnings_tag parent
+                content )
       | Error e ->
-          Errors.report ~what:(`Reference r) ~tools_error:(`Reference e)
-            `Resolve;
-          `Raise (orig, comment_nestable_block_element_list env parent content))
+          if not (Env.should_suppress_warnings env warnings_tag) then
+            Errors.report ~what:(`Reference r) ~tools_error:(`Reference e)
+              `Resolve;
+          `Raise
+            ( orig,
+              comment_nestable_block_element_list env warnings_tag parent
+                content ))
   | `Raise ((`Code_span _ as orig), content) ->
-      `Raise (orig, comment_nestable_block_element_list env parent content)
+      `Raise
+        ( orig,
+          comment_nestable_block_element_list env warnings_tag parent content )
   | `Return content ->
-      `Return (comment_nestable_block_element_list env parent content)
+      `Return
+        (comment_nestable_block_element_list env warnings_tag parent content)
   | `See (kind, target, content) ->
-      `See (kind, target, comment_nestable_block_element_list env parent content)
+      `See
+        ( kind,
+          target,
+          comment_nestable_block_element_list env warnings_tag parent content )
   | `Before (version, content) ->
-      `Before (version, comment_nestable_block_element_list env parent content)
+      `Before
+        ( version,
+          comment_nestable_block_element_list env warnings_tag parent content )
   | `Author _ | `Since _ | `Alert _ | `Version _ ->
       x (* only contain primitives *)
 
-and comment_block_element env parent ~loc (x : Comment.block_element) =
+and comment_block_element env warnings_tag parent ~loc
+    (x : Comment.block_element) =
   match x with
   | #Comment.nestable_block_element as x ->
-      (comment_nestable_block_element env parent ~loc x
+      (comment_nestable_block_element env warnings_tag parent ~loc x
         :> Comment.block_element)
   | `Heading (attrs, label, elems) ->
-      let cie = comment_inline_element env in
+      let cie = comment_inline_element env warnings_tag in
       let elems =
         List.rev_map (fun ele -> with_location cie ele) elems |> List.rev
       in
       let h = (attrs, label, elems) in
       check_ambiguous_label ~loc env h;
       `Heading h
-  | `Tag t -> `Tag (comment_tag env parent ~loc t)
+  | `Tag t -> `Tag (comment_tag env warnings_tag parent ~loc t)
 
-and with_location :
-    type a.
+and with_location : type a.
     (loc:_ -> a -> a) -> a Location_.with_location -> a Location_.with_location
     =
  fun fn { value; location = loc } ->
@@ -382,10 +411,16 @@ and with_location :
   { value; location = loc }
 
 and comment_docs env parent d =
-  List.rev_map
-    (with_location (comment_block_element env (parent :> Id.LabelParent.t)))
-    d
-  |> List.rev
+  {
+    Comment.elements =
+      List.rev_map
+        (with_location
+           (comment_block_element env d.Comment.warnings_tag
+              (parent :> Id.LabelParent.t)))
+        d.Comment.elements
+      |> List.rev;
+    warnings_tag = d.warnings_tag;
+  }
 
 and comment env parent = function
   | `Stop -> `Stop
@@ -395,22 +430,70 @@ and open_ env parent = function
   | { Odoc_model__Lang.Open.doc; _ } as open_ ->
       { open_ with doc = comment_docs env parent doc }
 
+let warn_on_hidden_representation (id : Id.Type.t)
+    (r : Lang.TypeDecl.Representation.t) =
+  let open Paths.Identifier in
+  let rec internal_typ_exp typ_expr =
+    let open Lang.TypeExpr in
+    let open Paths.Path in
+    match typ_expr with
+    | Constr (p, ts) ->
+        is_hidden (p :> Paths.Path.t)
+        || List.exists (fun t -> internal_typ_exp t) ts
+    | Poly (_, t) | Alias (t, _) -> internal_typ_exp t
+    | Arrow (_, t, t2) -> internal_typ_exp t || internal_typ_exp t2
+    | Tuple ts | Class (_, ts) -> List.exists (fun t -> internal_typ_exp t) ts
+    | _ -> false
+  in
+
+  let internal_cstr_arg t =
+    let open Lang.TypeDecl.Constructor in
+    let open Lang.TypeDecl.Field in
+    match t.args with
+    | Tuple type_exprs ->
+        List.exists (fun type_expr -> internal_typ_exp type_expr) type_exprs
+    | Record fields ->
+        List.exists (fun field -> internal_typ_exp field.type_) fields
+  in
+
+  let internal_field t =
+    let open Lang.TypeDecl.Field in
+    internal_typ_exp t.type_
+  in
+
+  let fmt_cfg = Component.Fmt.{ default with short_paths = true } in
+  match r with
+  | Variant constructors ->
+      if List.exists internal_cstr_arg constructors then
+        Lookup_failures.report_warning "@[<2>Hidden constructors in type '%a'@]"
+          Component.Fmt.(model_identifier fmt_cfg)
+          (id :> Id.any)
+  | Record fields ->
+      if List.exists internal_field fields then
+        Lookup_failures.report_warning "@[<2>Hidden fields in type '%a'@]"
+          Component.Fmt.(model_identifier fmt_cfg)
+          (id :> Id.any)
+  | Extensible -> ()
+
 let rec unit env t =
   let open Compilation_unit in
   let content =
-    match t.content with
-    | Module sg ->
-        let sg = signature env (t.id :> Id.Signature.t) sg in
-        Module sg
-    | Pack _ as p -> p
+    if t.hidden then t.content
+    else
+      match t.content with
+      | Module sg ->
+          let sg = signature env (t.id :> Id.Signature.t) sg in
+          Module sg
+      | Pack _ as p -> p
   in
-  { t with content; linked = true }
+  let source_loc = source_loc env t.id t.source_loc in
+  { t with content; linked = true; source_loc }
 
 and value_ env parent t =
   let open Value in
   {
     t with
-    locs = locations env t.id t.locs;
+    source_loc = source_loc env t.id t.source_loc;
     doc = comment_docs env parent t.doc;
     type_ = type_expression env parent [] t.type_;
   }
@@ -419,9 +502,9 @@ and exception_ env parent e =
   let open Exception in
   let res = Opt.map (type_expression env parent []) e.res in
   let args = type_decl_constructor_argument env parent e.args in
-  let locs = locations env e.id e.locs in
+  let source_loc = source_loc env e.id e.source_loc in
   let doc = comment_docs env parent e.doc in
-  { e with locs; res; args; doc }
+  { e with source_loc; res; args; doc }
 
 and extension env parent t =
   let open Extension in
@@ -429,7 +512,7 @@ and extension env parent t =
     let open Constructor in
     {
       c with
-      locs = locations env c.id c.locs;
+      source_loc = source_loc env c.id c.source_loc;
       args = type_decl_constructor_argument env parent c.args;
       res = Opt.map (type_expression env parent []) c.res;
       doc = comment_docs env parent c.doc;
@@ -452,7 +535,7 @@ and class_type env parent c =
   let doc = comment_docs env parent c.doc in
   {
     c with
-    locs = locations env c.id c.locs;
+    source_loc = source_loc env c.id c.source_loc;
     expr = class_type_expr env parent c.expr;
     doc;
   }
@@ -504,9 +587,9 @@ and class_ env parent c =
         Arrow (lbl, type_expression env parent [] expr, map_decl decl)
   in
   let doc = comment_docs env parent c.doc in
-  let locs = locations env c.id c.locs in
+  let source_loc = source_loc env c.id c.source_loc in
   let type_ = map_decl c.type_ in
-  { c with locs; type_; doc }
+  { c with source_loc; type_; doc }
 
 and module_substitution env parent m =
   let open ModuleSubstitution in
@@ -569,7 +652,7 @@ and simple_expansion :
 and module_ : Env.t -> Module.t -> Module.t =
  fun env m ->
   let open Module in
-  let open Utils.ResultMonad in
+  let open Odoc_utils.ResultMonad in
   let sg_id = (m.id :> Id.Signature.t) in
   if m.hidden then m
   else
@@ -591,9 +674,9 @@ and module_ : Env.t -> Module.t -> Module.t =
           else type_
       | Alias _ | ModuleType _ -> type_
     in
-    let locs = locations env m.id m.locs in
+    let source_loc = source_loc env m.id m.source_loc in
     let doc = comment_docs env sg_id m.doc in
-    { m with locs; doc; type_ }
+    { m with source_loc; doc; type_ }
 
 and module_decl : Env.t -> Id.Signature.t -> Module.decl -> Module.decl =
  fun env id decl ->
@@ -626,8 +709,8 @@ and module_type : Env.t -> ModuleType.t -> ModuleType.t =
        | _ -> false
      in*)
   let doc = comment_docs env sg_id m.doc in
-  let locs = (locations env m.id) m.locs in
-  { m with locs; expr = expr'; doc }
+  let source_loc = (source_loc env m.id) m.source_loc in
+  { m with source_loc; expr = expr'; doc }
 
 and module_type_substitution :
     Env.t -> ModuleTypeSubstitution.t -> ModuleTypeSubstitution.t =
@@ -691,7 +774,7 @@ and handle_fragments env id sg subs =
             | _ -> frag
           in
           let sg' =
-            Tools.fragmap ~mark_substituted:true env
+            Tools.fragmap env
               Component.Of_Lang.(with_module_type_substitution (empty ()) lsub)
               sg
           in
@@ -709,7 +792,7 @@ and handle_fragments env id sg subs =
             | _ -> frag
           in
           let sg' =
-            Tools.fragmap ~mark_substituted:true env
+            Tools.fragmap env
               Component.Of_Lang.(with_module_type_substitution (empty ()) lsub)
               sg
           in
@@ -727,7 +810,7 @@ and handle_fragments env id sg subs =
             | _ -> frag
           in
           let sg' =
-            Tools.fragmap ~mark_substituted:true env
+            Tools.fragmap env
               Component.Of_Lang.(with_module_type_substitution (empty ()) lsub)
               sg
           in
@@ -745,7 +828,7 @@ and handle_fragments env id sg subs =
             | _ -> frag
           in
           let sg' =
-            Tools.fragmap ~mark_substituted:true env
+            Tools.fragmap env
               Component.Of_Lang.(with_module_type_substitution (empty ()) lsub)
               sg
           in
@@ -763,7 +846,7 @@ and handle_fragments env id sg subs =
             | _ -> frag
           in
           let sg' =
-            Tools.fragmap ~mark_substituted:true env
+            Tools.fragmap env
               Component.Of_Lang.(with_module_type_substitution (empty ()) lsub)
               sg
           in
@@ -781,7 +864,7 @@ and handle_fragments env id sg subs =
             | _ -> frag
           in
           let sg' =
-            Tools.fragmap ~mark_substituted:true env
+            Tools.fragmap env
               Component.Of_Lang.(with_module_type_substitution (empty ()) lsub)
               sg
           in
@@ -799,24 +882,22 @@ and u_module_type_expr :
   | Path p -> Path (module_type_path env p)
   | With (subs, expr) as unresolved -> (
       let cexpr = Component.Of_Lang.(u_module_type_expr (empty ()) expr) in
-      match
-        Tools.signature_of_u_module_type_expr ~mark_substituted:true env cexpr
-      with
+      match Tools.signature_of_u_module_type_expr env cexpr with
       | Ok sg ->
           With (handle_fragments env id sg subs, u_module_type_expr env id expr)
       | Error e ->
           Errors.report ~what:(`Module_type_U cexpr) ~tools_error:e `Resolve;
           unresolved)
-  | TypeOf { t_desc = StructInclude p; t_expansion } ->
-      TypeOf { t_desc = StructInclude (module_path env p); t_expansion }
-  | TypeOf { t_desc = ModPath p; t_expansion } ->
-      TypeOf { t_desc = ModPath (module_path env p); t_expansion }
+  | TypeOf (StructInclude p, original_path) ->
+      TypeOf (StructInclude (module_path env p), original_path)
+  | TypeOf (ModPath p, original_path) ->
+      TypeOf (ModPath (module_path env p), original_path)
 
 and module_type_expr :
     Env.t -> Id.Signature.t -> ModuleType.expr -> ModuleType.expr =
  fun env id expr ->
   let open ModuleType in
-  let open Utils.ResultMonad in
+  let open Odoc_utils.ResultMonad in
   let do_expn cur (e : Paths.Path.ModuleType.t option) =
     match (cur, e) with
     | Some e, _ ->
@@ -827,7 +908,7 @@ and module_type_expr :
             Component.Of_Lang.(resolved_module_type_path (empty ()) p_path)
           in
           match
-            Tools.expansion_of_module_type_expr ~mark_substituted:false env
+            Tools.expansion_of_module_type_expr env
               (Path { p_path = `Resolved cp; p_expansion = None })
             >>= Expand_tools.handle_expansion env (id :> Id.Signature.t)
           with
@@ -845,9 +926,7 @@ and module_type_expr :
       Path { p_path; p_expansion = do_expn p_expansion (Some p_path) }
   | With { w_substitutions; w_expansion; w_expr } as unresolved -> (
       let cexpr = Component.Of_Lang.(u_module_type_expr (empty ()) w_expr) in
-      match
-        Tools.signature_of_u_module_type_expr ~mark_substituted:true env cexpr
-      with
+      match Tools.signature_of_u_module_type_expr env cexpr with
       | Ok sg ->
           With
             {
@@ -863,17 +942,19 @@ and module_type_expr :
       let env = Env.add_functor_parameter arg env in
       let res' = module_type_expr env (Paths.Identifier.Mk.result id) res in
       Functor (arg', res')
-  | TypeOf { t_desc = StructInclude p; t_expansion } ->
+  | TypeOf { t_desc = StructInclude p; t_expansion; t_original_path } ->
       TypeOf
         {
           t_desc = StructInclude (module_path env p);
           t_expansion = do_expn t_expansion None;
+          t_original_path;
         }
-  | TypeOf { t_desc = ModPath p; t_expansion } ->
+  | TypeOf { t_desc = ModPath p; t_expansion; t_original_path } ->
       TypeOf
         {
           t_desc = ModPath (module_path env p);
           t_expansion = do_expn t_expansion None;
+          t_original_path;
         }
 
 and type_decl_representation :
@@ -893,20 +974,25 @@ and type_decl : Env.t -> Id.Signature.t -> TypeDecl.t -> TypeDecl.t =
   let open TypeDecl in
   let equation = type_decl_equation env parent t.equation in
   let doc = comment_docs env parent t.doc in
-  let locs = locations env t.id t.locs in
+  let source_loc = source_loc env t.id t.source_loc in
   let hidden_path =
     match equation.Equation.manifest with
     | Some (Constr (`Resolved path, params))
       when Paths.Path.Resolved.(is_hidden (path :> t))
            || Paths.Path.Resolved.(identifier (path :> t))
-              = (t.id :> Paths.Identifier.t) ->
+              = Some (t.id :> Paths.Identifier.t) ->
         Some (path, params)
     | _ -> None
   in
   let representation =
-    Opt.map (type_decl_representation env parent) t.representation
+    Opt.map
+      (fun r ->
+        let r' = type_decl_representation env parent r in
+        warn_on_hidden_representation t.id r';
+        r')
+      t.representation
   in
-  let default = { t with locs; equation; doc; representation } in
+  let default = { t with source_loc; equation; doc; representation } in
   match hidden_path with
   | Some (p, params) -> (
       let p' = Component.Of_Lang.(resolved_type_path (empty ()) p) in
@@ -922,7 +1008,9 @@ and type_decl : Env.t -> Id.Signature.t -> TypeDecl.t -> TypeDecl.t =
             with _ -> default.equation
           in
           { default with equation = type_decl_equation env parent equation }
-      | Ok (`FClass _ | `FClassType _ | `FType_removed _) | Error _ -> default)
+      | Ok (`FClass _ | `FClassType _ | `FType_removed _ | `CoreType _)
+      | Error _ ->
+          default)
   | None -> default
 
 and type_decl_equation env parent t =
@@ -1022,7 +1110,7 @@ and type_expression : Env.t -> Id.Signature.t -> _ -> _ =
       if not (Paths.Path.is_hidden (path :> Paths.Path.t)) then Constr (path, ts)
       else
         let cp = Component.Of_Lang.(type_path (empty ()) path') in
-        match Tools.resolve_type env ~add_canonical:true cp with
+        match Tools.resolve_type env cp with
         | Ok (cp', `FType (_, t)) ->
             let cp' = Tools.reresolve_type env cp' in
             let p = Lang_of.(Path.resolved_type (empty ()) cp') in
@@ -1056,7 +1144,7 @@ and type_expression : Env.t -> Id.Signature.t -> _ -> _ =
                       Constr (`Resolved p, ts))
               | _ -> Constr (`Resolved p, ts)
             else Constr (`Resolved p, ts)
-        | Ok (cp', (`FClass _ | `FClassType _)) ->
+        | Ok (cp', (`FClass _ | `FClassType _ | `CoreType _)) ->
             let p = Lang_of.(Path.resolved_type (empty ()) cp') in
             Constr (`Resolved p, ts)
         | Ok (_cp, `FType_removed (_, x, _eq)) ->
@@ -1083,24 +1171,26 @@ and type_expression : Env.t -> Id.Signature.t -> _ -> _ =
 
 let link ~filename x y =
   Lookup_failures.catch_failures ~filename (fun () ->
-      if y.Lang.Compilation_unit.linked || y.hidden then y else unit x y)
+      if y.Lang.Compilation_unit.linked then y else unit x y)
 
 let page env page =
   let () =
     List.iter
       (fun child ->
-        let check_resolves ~what f name =
-          match f name env with
-          | Some _ -> ()
-          | None -> Errors.report ~what `Lookup
-        in
         match child with
-        | Page.Asset_child _ | Page.Source_tree_child _ -> ()
-        | Page.Page_child page ->
-            check_resolves ~what:(`Child_page page) Env.lookup_page page
-        | Page.Module_child mod_ ->
-            check_resolves ~what:(`Child_module mod_) Env.lookup_root_module
-              mod_)
+        | Page.Page_child page -> (
+            match Env.lookup_page_by_name page env with
+            | Ok _ -> ()
+            | Error `Not_found -> Errors.report ~what:(`Child_page page) `Lookup
+            )
+        | Page.Module_child mod_ -> (
+            match
+              Env.lookup_root_module
+                (Odoc_model.Names.ModuleName.make_std mod_)
+                env
+            with
+            | Some _ -> ()
+            | None -> Errors.report ~what:(`Child_module mod_) `Lookup))
       page.Lang.Page.children
   in
   {
@@ -1109,6 +1199,52 @@ let page env page =
     linked = true;
   }
 
+let source_info env infos =
+  let open Source_info in
+  let jump_to v f_impl f_doc =
+    let documentation =
+      match v.documentation with Some p -> Some (f_doc p) | None -> None
+    in
+    let implementation =
+      match v.implementation with
+      | Some (Unresolved p) -> (
+          match f_impl p with
+          | Some x -> Some (Resolved x)
+          | None -> v.implementation)
+      | x -> x
+    in
+    { documentation; implementation }
+  in
+  List.map
+    (fun (i, pos) ->
+      let info =
+        match i with
+        | Value v ->
+            Value
+              (jump_to v (Shape_tools.lookup_value_path env) (value_path env))
+        | Module v ->
+            Module
+              (jump_to v (Shape_tools.lookup_module_path env) (module_path env))
+        | ModuleType v ->
+            ModuleType
+              (jump_to v
+                 (Shape_tools.lookup_module_type_path env)
+                 (module_type_path env))
+        | Type v ->
+            Type (jump_to v (Shape_tools.lookup_type_path env) (type_path env))
+        | i -> i
+      in
+      (info, pos))
+    infos
+
+let impl env i =
+  let open Implementation in
+  { i with source_info = source_info env i.source_info; linked = true }
+
 let resolve_page ~filename env p =
   Lookup_failures.catch_failures ~filename (fun () ->
       if p.Lang.Page.linked then p else page env p)
+
+let resolve_impl ~filename env i =
+  Lookup_failures.catch_failures ~filename (fun () ->
+      if i.Lang.Implementation.linked then i else impl env i)
