@@ -1,7 +1,8 @@
 open Odoc_model.Paths
 open Odoc_model.Names
 open Reference
-open Utils.ResultMonad
+open Odoc_utils
+open ResultMonad
 
 type module_lookup_result =
   Resolved.Module.t * Cpath.Resolved.module_ * Component.Module.t
@@ -20,15 +21,20 @@ type class_type_lookup_result = Resolved.ClassType.t * Component.ClassType.t
 
 type page_lookup_result = Resolved.Page.t * Odoc_model.Lang.Page.t
 
+type asset_lookup_result = Resolved.Asset.t
+
 type type_lookup_result =
   [ `T of datatype_lookup_result
   | `C of class_lookup_result
   | `CT of class_type_lookup_result ]
 
+type any_path_lookup_result =
+  [ `P of page_lookup_result | `S of signature_lookup_result ]
+
 type label_parent_lookup_result =
-  [ `S of signature_lookup_result
-  | type_lookup_result
-  | `P of page_lookup_result ]
+  [ type_lookup_result
+  | `P of page_lookup_result
+  | `S of signature_lookup_result ]
 
 type fragment_type_parent_lookup_result =
   [ `S of signature_lookup_result | `T of datatype_lookup_result ]
@@ -84,6 +90,7 @@ let ref_kind_of_find = function
   | `FClass _ -> "class"
   | `FClassType _ -> "class-type"
   | `FConstructor _ | `In_type (_, _, `FConstructor _) -> "constructor"
+  | `In_type (_, _, `FPoly _) -> "polymorphic constructor"
   | `FExn _ -> "exception"
   | `FExt _ -> "extension"
   | `FExtDecl _ -> "extension-decl"
@@ -138,17 +145,15 @@ let find_ambiguous ?(kind = `Any) find sg name =
       Ok x
   | [] -> Error (`Find_by_name (kind, name))
 
-let find find sg name =
+let find find sg conv name =
   match find sg name with
   | Some x -> Ok x
-  | None -> Error (`Find_by_name (`Any, name))
+  | None -> Error (`Find_by_name (`Any, (conv name : string)))
 
 let module_lookup_to_signature_lookup env (ref, cp, m) =
   let rec handle_expansion : Tools.expansion -> _ = function
     | Functor (_, expr) -> (
-        match
-          Tools.expansion_of_module_type_expr ~mark_substituted:true env expr
-        with
+        match Tools.expansion_of_module_type_expr env expr with
         | Ok e -> handle_expansion e
         | Error _ as e -> e)
     | Signature sg -> Ok ((ref :> Resolved.Signature.t), `Module cp, sg)
@@ -187,7 +192,7 @@ module M = struct
       else (base_path', base_ref')
     in
     let p, r =
-      match Tools.get_module_path_modifiers env ~add_canonical:true m with
+      match Tools.get_module_path_modifiers env m with
       | None -> (base_path, base_ref)
       | Some (`Aliased cp) ->
           let cp = Tools.reresolve_module env cp in
@@ -203,7 +208,8 @@ module M = struct
       =
     let parent_cp = Tools.reresolve_parent env parent_cp in
     let sg = Tools.prefix_signature (parent_cp, sg) in
-    find Find.module_in_sig sg name >>= fun (`FModule (name, m)) ->
+    find Find.module_in_sig sg ModuleName.to_string name
+    >>= fun (`FModule (name, m)) ->
     Ok (of_component env m (`Module (parent_cp, name)) (`Module (parent, name)))
 
   let of_element env (`Module (id, m)) : t =
@@ -214,7 +220,46 @@ module M = struct
   let in_env env name =
     match env_lookup_by_name Env.s_module name env with
     | Ok e -> Ok (of_element env e)
-    | Error _ -> Error (`Parent (`Parent_module (`Lookup_failure_root name)))
+    | Error _ ->
+        Error
+          (`Parent
+             (`Parent_module (`Lookup_failure_root (ModuleName.make_std name))))
+end
+
+module Path = struct
+  (* let first_seg (`Root (s, _) | `Slash (_, s)) = s *)
+
+  let mk_lookup_error (tag, path) = Error (`Path_error (`Not_found, tag, path))
+
+  let handle_lookup_error p = function
+    | Ok _ as ok -> ok
+    | Error `Not_found -> mk_lookup_error p
+
+  let page_in_env env p : page_lookup_result ref_result =
+    Env.lookup_page_by_path p env |> handle_lookup_error p >>= fun p ->
+    Ok (`Identifier p.name, p)
+
+  let asset_in_env env p : asset_lookup_result ref_result =
+    Env.lookup_asset_by_path p env |> handle_lookup_error p >>= fun p ->
+    Ok (`Identifier p.name)
+
+  let module_in_env env p : module_lookup_result ref_result =
+    Env.lookup_unit_by_path p env |> handle_lookup_error p >>= fun m ->
+    Ok (M.of_element env m)
+
+  let any_in_env env p : any_path_lookup_result ref_result =
+    (* TODO: Resolve modules *)
+    let page_result = page_in_env env p in
+    let module_result = module_in_env env p in
+    match (page_result, module_result) with
+    | Ok page, Error _ -> Ok (`P page)
+    | Error _, Ok m ->
+        module_lookup_to_signature_lookup env m >>= fun s -> Ok (`S s)
+    | Ok page, Ok _ ->
+        let name = List.last (snd p) in
+        ambiguous_generic_ref_warning name [ "module"; "page" ];
+        Ok (`P page)
+    | Error _, Error _ -> mk_lookup_error p
 end
 
 module MT = struct
@@ -223,7 +268,7 @@ module MT = struct
   type t = module_type_lookup_result
 
   let of_component env mt base_path base_ref : t =
-    match Tools.get_module_type_path_modifiers env ~add_canonical:true mt with
+    match Tools.get_module_type_path_modifiers env mt with
     | None -> (base_ref, base_path, mt)
     | Some (`AliasModuleType cp) ->
         let cp = Tools.reresolve_module_type env cp in
@@ -233,7 +278,8 @@ module MT = struct
   let in_signature env ((parent', parent_cp, sg) : signature_lookup_result) name
       =
     let sg = Tools.prefix_signature (parent_cp, sg) in
-    find Find.module_type_in_sig sg name >>= fun (`FModuleType (name, mt)) ->
+    find Find.module_type_in_sig sg ModuleTypeName.to_string name
+    >>= fun (`FModuleType (name, mt)) ->
     Ok
       (of_component env mt
          (`ModuleType (parent_cp, name))
@@ -290,7 +336,7 @@ module DT = struct
   let in_signature _env ((parent', parent_cp, sg) : signature_lookup_result)
       name =
     let sg = Tools.prefix_signature (parent_cp, sg) in
-    find Find.datatype_in_sig sg name >>= function
+    find Find.datatype_in_sig sg TypeName.to_string name >>= function
     | `FType (name, t) -> Ok (`T (`Type (parent', name), t))
 end
 
@@ -311,7 +357,7 @@ module T = struct
   let in_signature _env ((parent', parent_cp, sg) : signature_lookup_result)
       name =
     let sg = Tools.prefix_signature (parent_cp, sg) in
-    find Find.type_in_sig sg name >>= function
+    find Find.type_in_sig sg TypeName.to_string name >>= function
     | `FType (name, t) -> Ok (`T (`Type (parent', name), t))
     | `FClass (name, c) -> Ok (`C (`Class (parent', name), c))
     | `FClassType (name, ct) -> Ok (`CT (`ClassType (parent', name), ct))
@@ -328,9 +374,9 @@ module V = struct
 
   let of_component _env ~parent_ref name = Ok (`Value (parent_ref, name))
 
-  let in_signature _env ((parent', _, sg) : signature_lookup_result) name =
-    find_ambiguous ~kind:`S Find.value_in_sig sg (ValueName.to_string name)
-    >>= fun _ -> Ok (`Value (parent', name))
+  let in_signature _env ((parent, _, sg) : signature_lookup_result) name =
+    find Find.value_in_sig sg ValueName.to_string name >>= function
+    | `FValue (name, _) -> Ok (`Value (parent, name))
 end
 
 module L = struct
@@ -338,9 +384,9 @@ module L = struct
 
   type t = Resolved.Label.t
 
-  let in_env env name : t ref_result =
-    env_lookup_by_name Env.s_label name env >>= fun (`Label (id, _)) ->
-    Ok (`Identifier id)
+  let in_env env name : (t * _) ref_result =
+    env_lookup_by_name Env.s_label name env >>= fun (`Label (id, lbl)) ->
+    Ok (`Identifier id, lbl.text)
 
   let in_page _env (`P (_, p)) name =
     let rec find = function
@@ -350,26 +396,30 @@ module L = struct
               ( _,
                 ({ Odoc_model.Paths.Identifier.iv = `Label (_, name'); _ } as
                  label),
-                _ )
+                content )
             when name = LabelName.to_string name' ->
-              Ok (`Identifier label)
+              Ok (`Identifier label, content)
           | _ -> find tl)
       | [] -> Error (`Find_by_name (`Page, name))
     in
-    find p.Odoc_model.Lang.Page.content
+    find p.Odoc_model.Lang.Page.content.elements
 
   let of_component _env ~parent_ref label =
     Ok
-      (`Label
-        ( (parent_ref :> Resolved.LabelParent.t),
-          Ident.Name.typed_label label.Component.Label.label ))
+      ( `Label
+          ( (parent_ref :> Resolved.LabelParent.t),
+            Ident.Name.typed_label label.Component.Label.label ),
+        label.text )
 
   let in_label_parent env (parent : label_parent_lookup_result) name =
     match parent with
-    | `S (p, _, sg) ->
-        find_ambiguous ~kind:`Label Find.label_in_sig sg
-          (LabelName.to_string name)
-        >>= fun _ -> Ok (`Label ((p :> Resolved.LabelParent.t), name))
+    | `S (p, _, sg) -> (
+        find_ambiguous ~kind:`Label
+          (fun sg l -> Find.label_in_sig sg (LabelName.make_std l))
+          sg (LabelName.to_string name)
+        >>= function
+        | `FLabel lbl ->
+            Ok (`Label ((p :> Resolved.LabelParent.t), name), lbl.text))
     | (`T _ | `C _ | `CT _) as r -> wrong_kind_error [ `S; `Page ] r
     | `P _ as page -> in_page env page (LabelName.to_string name)
 end
@@ -389,7 +439,7 @@ module EC = struct
   let in_signature _env ((parent', parent_cp, sg) : signature_lookup_result)
       name =
     let sg = Tools.prefix_signature (parent_cp, sg) in
-    find Find.extension_in_sig sg (ExtensionName.to_string name) >>= fun _ ->
+    find Find.extension_in_sig sg ExtensionName.to_string name >>= fun _ ->
     Ok (`Extension (parent', name))
 end
 
@@ -407,14 +457,15 @@ module ED = struct
         let id_parent = match id.iv with `Extension (p, _) -> p in
         Ok
           (`Identifier
-            (Identifier.Mk.extension_decl
-               ( id_parent,
-                 (ExtensionName.make_std c.name, ExtensionName.make_std name) )))
+             (Identifier.Mk.extension_decl
+                ( id_parent,
+                  (ExtensionName.make_std c.name, ExtensionName.make_std name)
+                )))
 
   let in_signature _env ((parent', parent_cp, sg) : signature_lookup_result)
       name =
     let sg = Tools.prefix_signature (parent_cp, sg) in
-    find Find.extension_in_sig sg (ExtensionName.to_string name)
+    find Find.extension_in_sig sg ExtensionName.to_string name
     >>= fun (`FExt (ext, _) : Find.extension) ->
     (* Type extensions always have at least 1 constructor.
        The reference to the type extension shares the same name as the first constructor. *)
@@ -438,7 +489,7 @@ module EX = struct
   let in_signature _env ((parent', parent_cp, sg) : signature_lookup_result)
       name =
     let sg = Tools.prefix_signature (parent_cp, sg) in
-    find Find.exception_in_sig sg (ExceptionName.to_string name) >>= fun _ ->
+    find Find.exception_in_sig sg ExceptionName.to_string name >>= fun _ ->
     Ok (`Exception (parent', name))
 end
 
@@ -480,18 +531,32 @@ module CS = struct
         let sg = Tools.prefix_signature (parent_cp, sg) in
         find_ambiguous Find.any_in_type_in_sig sg name_s >>= function
         | `In_type (_, _, `FField _) -> got_a_field name_s
+        | `In_type (typ_name, _, `FPoly cs) ->
+            Ok
+              (`PolyConstructor
+                 (`Type (parent', typ_name), ConstructorName.make_std cs.name))
         | `In_type (typ_name, _, `FConstructor _) ->
             Ok (`Constructor (`Type (parent', typ_name), name)))
     | `T (parent', t) -> (
-        find Find.any_in_type t name_s >>= function
+        find Find.any_in_type t (fun x -> x) name_s >>= function
         | `FField _ -> got_a_field name_s
+        | `FPoly cs ->
+            Ok
+              (`PolyConstructor
+                 ( (parent' : Resolved.DataType.t),
+                   ConstructorName.make_std cs.name ))
         | `FConstructor _ ->
             Ok (`Constructor ((parent' : Resolved.DataType.t), name)))
 
   let of_component _env parent name =
     Ok
       (`Constructor
-        ((parent : Resolved.DataType.t), ConstructorName.make_std name))
+         ((parent : Resolved.DataType.t), ConstructorName.make_std name))
+
+  let poly_of_component _env parent name =
+    Ok
+      (`PolyConstructor
+         ((parent : Resolved.DataType.t), ConstructorName.make_std name))
 end
 
 module F = struct
@@ -514,20 +579,22 @@ module F = struct
         let sg = Tools.prefix_signature (parent_cp, sg) in
         find_ambiguous Find.any_in_type_in_sig sg name_s >>= function
         | `In_type (_, _, `FConstructor _) -> got_a_constructor name_s
+        | `In_type (_, _, `FPoly _) -> got_a_constructor name_s
         | `In_type (typ_name, _, `FField _) ->
             Ok
               (`Field
-                ((`Type (parent', typ_name) :> Resolved.FieldParent.t), name)))
+                 ((`Type (parent', typ_name) :> Resolved.FieldParent.t), name)))
     | `T (parent', t) -> (
-        find Find.any_in_type t name_s >>= function
+        find Find.any_in_type t (fun x -> x) name_s >>= function
         | `FConstructor _ -> got_a_constructor name_s
+        | `FPoly _ -> got_a_constructor name_s
         | `FField _ -> Ok (`Field ((parent' :> Resolved.FieldParent.t), name)))
 
   let of_component _env parent name =
     Ok
       (`Field
-        ( (parent : Resolved.DataType.t :> Resolved.FieldParent.t),
-          FieldName.make_std name ))
+         ( (parent : Resolved.DataType.t :> Resolved.FieldParent.t),
+           FieldName.make_std name ))
 end
 
 module MM = struct
@@ -539,7 +606,7 @@ module MM = struct
   let in_env _env name : t ref_result = Error (`Lookup_by_name (`Any, name))
 
   let in_class_signature _env (parent', cs) name =
-    find Find.method_in_class_signature cs (MethodName.to_string name)
+    find Find.method_in_class_signature cs MethodName.to_string name
     >>= fun _ -> Ok (`Method (parent', name))
 
   let of_component _env parent' name = Ok (`Method (parent', name))
@@ -555,7 +622,7 @@ module MV = struct
 
   let in_class_signature _env (parent', cs) name =
     find Find.instance_variable_in_class_signature cs
-      (InstanceVariableName.to_string name)
+      InstanceVariableName.to_string name
     >>= fun _ -> Ok (`InstanceVariable (parent', name))
 
   let of_component _env parent' name = Ok (`InstanceVariable (parent', name))
@@ -565,13 +632,22 @@ module Page = struct
   type t = page_lookup_result
 
   let in_env env name : t ref_result =
-    match Env.lookup_page name env with
-    | Some p -> Ok (`Identifier p.Odoc_model.Lang.Page.name, p)
-    | None -> Error (`Lookup_by_name (`Page, name))
+    match Env.lookup_page_by_name name env with
+    | Ok p -> Ok (`Identifier p.Odoc_model.Lang.Page.name, p)
+    | Error `Not_found -> Error (`Lookup_by_name (`Page, name))
 
   let of_element _env (`Page (id, page)) : t = (`Identifier id, page)
 end
 
+module Asset = struct
+  type t = asset_lookup_result
+
+  let in_env env name : t ref_result =
+    match Env.lookup_asset_by_name name env with
+    | Ok p -> Ok (`Identifier p.Odoc_model.Lang.Asset.name)
+    | Error `Not_found -> Error (`Lookup_by_name (`Page (* TODO *), name))
+end
+
 module LP = struct
   (** Label parent *)
 
@@ -617,7 +693,7 @@ module LP = struct
         Ok (`CT ct)
 end
 
-let rec resolve_label_parent_reference env r =
+let rec resolve_label_parent_reference env (r : LabelParent.t) =
   let label_parent_res_of_type_res : type_lookup_result -> _ =
    fun r -> Ok (r :> label_parent_lookup_result)
   in
@@ -629,19 +705,16 @@ let rec resolve_label_parent_reference env r =
   | `Root (name, `TType) -> T.in_env env name >>= label_parent_res_of_type_res
   | `Type (parent, name) ->
       resolve_signature_reference env parent >>= fun p ->
-      T.in_signature env p (TypeName.to_string name)
-      >>= label_parent_res_of_type_res
+      T.in_signature env p name >>= label_parent_res_of_type_res
   | `Root (name, `TClass) -> CL.in_env env name >>= fun r -> Ok (`C r)
   | `Class (parent, name) ->
       resolve_signature_reference env parent >>= fun p ->
-      T.in_signature env p (ClassName.to_string name)
-      >>= class_lookup_result_of_type
-      >>= fun r -> Ok (`C r)
+      T.in_signature env p name >>= class_lookup_result_of_type >>= fun r ->
+      Ok (`C r)
   | `Root (name, `TClassType) -> CT.in_env env name >>= fun r -> Ok (`CT r)
   | `ClassType (parent, name) ->
       resolve_signature_reference env parent >>= fun p ->
-      T.in_signature env p (ClassTypeName.to_string name)
-      >>= class_type_lookup_result_of_type
+      T.in_signature env p name >>= class_type_lookup_result_of_type
       >>= fun r -> Ok (`CT r)
   | `Dot (parent, name) ->
       resolve_label_parent_reference env parent
@@ -652,6 +725,12 @@ let rec resolve_label_parent_reference env r =
   | `Root (name, `TChildModule) ->
       resolve_signature_reference env (`Root (name, `TModule)) >>= fun s ->
       Ok (`S s)
+  | `Page_path p -> Path.page_in_env env p >>= fun r -> Ok (`P r)
+  | `Module_path p ->
+      Path.module_in_env env p >>= module_lookup_to_signature_lookup env
+      >>= fun r -> Ok (`S r)
+  | `Any_path p ->
+      Path.any_in_env env p >>= fun r -> Ok (r :> label_parent_lookup_result)
 
 and resolve_fragment_type_parent_reference (env : Env.t)
     (r : FragmentTypeParent.t) : (fragment_type_parent_lookup_result, _) result
@@ -668,11 +747,14 @@ and resolve_fragment_type_parent_reference (env : Env.t)
       DT.in_env env name >>= fragment_type_parent_res_of_type_res
   | `Type (parent, name) ->
       resolve_signature_reference env parent >>= fun p ->
-      DT.in_signature env p (TypeName.to_string name)
+      DT.in_signature env p name
   | `Dot (parent, name) ->
       resolve_label_parent_reference env parent
       >>= signature_lookup_result_of_label_parent
-      >>= fun p -> DT.in_signature env p name
+      >>= fun p -> DT.in_signature env p (TypeName.make_std name)
+  | `Module_path p ->
+      Path.module_in_env env p >>= module_lookup_to_signature_lookup env
+      >>= fun r -> Ok (`S r)
 
 and resolve_signature_reference :
     Env.t -> Signature.t -> signature_lookup_result ref_result =
@@ -686,13 +768,12 @@ and resolve_signature_reference :
         M.in_env env name >>= module_lookup_to_signature_lookup env
     | `Module (parent, name) ->
         resolve_signature_reference env parent >>= fun p ->
-        M.in_signature env p (ModuleName.to_string name)
-        >>= module_lookup_to_signature_lookup env
+        M.in_signature env p name >>= module_lookup_to_signature_lookup env
     | `Root (name, `TModuleType) ->
         MT.in_env env name >>= module_type_lookup_to_signature_lookup env
     | `ModuleType (parent, name) ->
         resolve_signature_reference env parent >>= fun p ->
-        MT.in_signature env p (ModuleTypeName.to_string name)
+        MT.in_signature env p name
         >>= module_type_lookup_to_signature_lookup env
     | `Root (name, `TUnknown) -> (
         env_lookup_by_name Env.s_signature name env >>= function
@@ -717,6 +798,8 @@ and resolve_signature_reference :
               (MT.of_component env mt
                  (`ModuleType (parent_cp, name))
                  (`ModuleType (parent, name))))
+    | `Module_path p ->
+        Path.module_in_env env p >>= module_lookup_to_signature_lookup env
   in
   resolve env'
 
@@ -727,11 +810,12 @@ and resolve_module_reference env (r : Module.t) : M.t ref_result =
   | `Dot (parent, name) ->
       resolve_label_parent_reference env parent
       >>= signature_lookup_result_of_label_parent
-      >>= fun p -> M.in_signature env p name
+      >>= fun p -> M.in_signature env p (ModuleName.make_std name)
   | `Module (parent, name) ->
       resolve_signature_reference env parent >>= fun p ->
-      M.in_signature env p (ModuleName.to_string name)
+      M.in_signature env p name
   | `Root (name, _) -> M.in_env env name
+  | `Module_path p -> Path.module_in_env env p
 
 let resolve_class_signature_reference env (r : ClassSignature.t) =
   (* Casting from ClassSignature to LabelParent.
@@ -742,17 +826,26 @@ let resolve_class_signature_reference env (r : ClassSignature.t) =
 
 (***)
 
-let resolved1 r = Ok (r :> Resolved.t)
+let resolved1 r = Ok ((r :> Resolved.t), None)
+
+let resolved_with_text (r, txt) = Ok ((r :> Reference.Resolved.t), Some txt)
 
 let resolved3 (r, _, _) = resolved1 r
 
 and resolved2 (r, _) = resolved1 r
 
+let resolve_asset_reference env (r : Reference.Asset.t) : Asset.t ref_result =
+  match r with `Resolved r -> Ok r | `Asset_path p -> Path.asset_in_env env p
+
 let resolved_type_lookup = function
   | `T (r, _) -> resolved1 r
   | `C (r, _) -> resolved1 r
   | `CT (r, _) -> resolved1 r
 
+let resolved_page_path_lookup = function
+  | `S (r, _, _) -> resolved1 r
+  | `P (r, _) -> resolved1 r
+
 let resolve_reference_dot_sg env ~parent_path ~parent_ref ~parent_sg name =
   let parent_path = Tools.reresolve_parent env parent_path in
   let parent_sg = Tools.prefix_signature (parent_path, parent_sg) in
@@ -772,23 +865,25 @@ let resolve_reference_dot_sg env ~parent_path ~parent_ref ~parent_sg name =
   | `FClassType (name, ct) ->
       CT.of_component env ct ~parent_ref name >>= resolved2
   | `FValue (name, _) -> V.of_component env ~parent_ref name >>= resolved1
-  | `FLabel label -> L.of_component env ~parent_ref label >>= resolved1
+  | `FLabel label -> L.of_component env ~parent_ref label >>= resolved_with_text
   | `FExn (name, _) -> EX.of_component env ~parent_ref name >>= resolved1
   | `FExt _ -> EC.of_component env ~parent_ref name >>= resolved1
   | `In_type (typ_name, _, r) -> (
       let parent = `Type (parent_ref, typ_name) in
       match r with
       | `FConstructor _ -> CS.of_component env parent name >>= resolved1
+      | `FPoly p -> CS.poly_of_component env parent p.name >>= resolved1
       | `FField _ -> F.of_component env parent name >>= resolved1)
   | `FModule_subst _ | `FType_subst _ | `FModuleType_subst _ ->
       Error (`Find_by_name (`Any, name))
 
 let resolve_reference_dot_page env page name =
-  L.in_page env page name >>= resolved1
+  L.in_page env page name >>= resolved_with_text
 
 let resolve_reference_dot_type env ~parent_ref t name =
-  find Find.any_in_type t name >>= function
+  find Find.any_in_type t (fun x -> x) name >>= function
   | `FConstructor _ -> CS.of_component env parent_ref name >>= resolved1
+  | `FPoly p -> CS.poly_of_component env parent_ref p.name >>= resolved1
   | `FField _ -> F.of_component env parent_ref name >>= resolved1
 
 let resolve_reference_dot_class env p name =
@@ -807,18 +902,29 @@ let resolve_reference_dot env parent name =
   | `P _ as page -> resolve_reference_dot_page env page name
 
 (** Warnings may be generated with [Error.implicit_warning] *)
-let resolve_reference =
+let resolve_reference :
+    Env.t ->
+    Reference.t ->
+    ( Reference.Resolved.t * Odoc_model.Comment.paragraph option,
+      Errors.Tools_error.reference_lookup_error )
+    result =
   let resolved = resolved3 in
   fun env r ->
     match r with
     | `Root (name, `TUnknown) -> (
-        let identifier id = Ok (`Identifier (id :> Identifier.t)) in
+        let identifier ?text id = Ok (`Identifier (id :> Identifier.t), text) in
         env_lookup_by_name Env.s_any name env >>= function
         | `Module (_, _) as e -> resolved (M.of_element env e)
         | `ModuleType (_, _) as e -> resolved (MT.of_element env e)
         | `Value (id, _) -> identifier id
         | `Type (id, _) -> identifier id
-        | `Label (id, _) -> identifier id
+        | `Label (id, _) ->
+            let text =
+              match Env.lookup_by_id Env.s_label id env with
+              | Some (`Label (_, lbl)) -> Some lbl.Component.Label.text
+              | None -> None
+            in
+            identifier ?text id
         | `Class (id, _) -> identifier id
         | `ClassType (id, _) -> identifier id
         | `Constructor (id, _) -> identifier id
@@ -827,38 +933,38 @@ let resolve_reference =
         | `ExtensionDecl (id, _) -> identifier id
         | `Field (id, _) -> identifier id
         | `Page (id, _) -> identifier id)
-    | `Resolved r -> Ok r
+    | `Resolved r -> Ok (r, None)
     | `Root (name, (`TModule | `TChildModule)) -> M.in_env env name >>= resolved
     | `Module (parent, name) ->
         resolve_signature_reference env parent >>= fun p ->
-        M.in_signature env p (ModuleName.to_string name) >>= resolved
+        M.in_signature env p name >>= resolved
     | `Root (name, `TModuleType) -> MT.in_env env name >>= resolved
     | `ModuleType (parent, name) ->
         resolve_signature_reference env parent >>= fun p ->
-        MT.in_signature env p (ModuleTypeName.to_string name) >>= resolved
+        MT.in_signature env p name >>= resolved
     | `Root (name, `TType) -> T.in_env env name >>= resolved_type_lookup
     | `Type (parent, name) ->
         resolve_signature_reference env parent >>= fun p ->
-        T.in_signature env p (TypeName.to_string name) >>= resolved_type_lookup
+        T.in_signature env p name >>= resolved_type_lookup
     | `Root (name, `TClass) -> CL.in_env env name >>= resolved2
     | `Class (parent, name) ->
         resolve_signature_reference env parent >>= fun p ->
-        T.in_signature env p (ClassName.to_string name)
-        >>= class_lookup_result_of_type >>= resolved2
+        T.in_signature env p name >>= class_lookup_result_of_type >>= resolved2
     | `Root (name, `TClassType) -> CT.in_env env name >>= resolved2
     | `ClassType (parent, name) ->
         resolve_signature_reference env parent >>= fun p ->
-        T.in_signature env p (ClassTypeName.to_string name)
-        >>= class_type_lookup_result_of_type >>= resolved2
+        T.in_signature env p name >>= class_type_lookup_result_of_type
+        >>= resolved2
     | `Root (name, `TValue) -> V.in_env env name >>= resolved1
     | `Value (parent, name) ->
         resolve_signature_reference env parent >>= fun p ->
         V.in_signature env p name >>= resolved1
-    | `Root (name, `TLabel) -> L.in_env env name >>= resolved1
+    | `Root (name, `TLabel) -> L.in_env env name >>= resolved_with_text
     | `Label (parent, name) ->
         resolve_label_parent_reference env parent >>= fun p ->
-        L.in_label_parent env p name >>= resolved1
+        L.in_label_parent env p name >>= resolved_with_text
     | `Root (name, (`TPage | `TChildPage)) -> Page.in_env env name >>= resolved2
+    | `Root (name, `TAsset) -> Asset.in_env env name >>= resolved1
     | `Dot (parent, name) -> resolve_reference_dot env parent name
     | `Root (name, `TConstructor) -> CS.in_env env name >>= resolved1
     | `Constructor (parent, name) ->
@@ -888,9 +994,19 @@ let resolve_reference =
     | `InstanceVariable (parent, name) ->
         resolve_class_signature_reference env parent >>= fun p ->
         MV.in_class_signature env p name >>= resolved1
+    | `Page_path p -> Path.page_in_env env p >>= resolved2
+    | `Asset_path a -> Path.asset_in_env env a >>= resolved1
+    | `Module_path p ->
+        Path.module_in_env env p
+        >>= module_lookup_to_signature_lookup env
+        >>= resolved
+    | `Any_path p -> Path.any_in_env env p >>= resolved_page_path_lookup
 
 let resolve_module_reference env m =
   Odoc_model.Error.catch_warnings (fun () -> resolve_module_reference env m)
 
+let resolve_asset_reference env m =
+  Odoc_model.Error.catch_warnings (fun () -> resolve_asset_reference env m)
+
 let resolve_reference env m =
   Odoc_model.Error.catch_warnings (fun () -> resolve_reference env m)
