@@ -2,8 +2,6 @@ open Component
 
 exception Invalidated
 
-exception MTOInvalidated
-
 type ('a, 'b) or_replaced = Not_replaced of 'a | Replaced of 'b
 
 type 'a type_or_replaced = ('a, TypeExpr.t * TypeDecl.Equation.t) or_replaced
@@ -21,14 +19,13 @@ type nonrec t = t
 
 let identity =
   {
-    module_ = PathModuleMap.empty;
+    module_ = ModuleMap.empty;
     module_type = ModuleTypeMap.empty;
     module_type_replacement = ModuleTypeMap.empty;
-    type_ = PathTypeMap.empty;
-    class_type = PathClassTypeMap.empty;
-    type_replacement = PathTypeMap.empty;
+    type_ = TypeMap.empty;
+    class_type = TypeMap.empty;
+    type_replacement = TypeMap.empty;
     path_invalidating_modules = [];
-    module_type_of_invalidating_modules = [];
     unresolve_opaque_paths = false;
   }
 
@@ -37,15 +34,8 @@ let unresolve_opaque_paths s = { s with unresolve_opaque_paths = true }
 let path_invalidate_module id t =
   { t with path_invalidating_modules = id :: t.path_invalidating_modules }
 
-let mto_invalidate_module id t =
-  {
-    t with
-    module_type_of_invalidating_modules =
-      id :: t.module_type_of_invalidating_modules;
-  }
-
 let add_module id p rp t =
-  { t with module_ = PathModuleMap.add id (`Prefixed (p, rp)) t.module_ }
+  { t with module_ = ModuleMap.add id (`Prefixed (p, rp)) t.module_ }
 
 let add_module_type id p rp t =
   {
@@ -55,50 +45,40 @@ let add_module_type id p rp t =
 
 let add_type : Ident.type_ -> Cpath.type_ -> Cpath.Resolved.type_ -> t -> t =
  fun id p rp t ->
-  {
-    t with
-    type_ = PathTypeMap.add (id :> Ident.path_type) (`Prefixed (p, rp)) t.type_;
-  }
+  { t with type_ = TypeMap.add (id :> Ident.type_) (`Prefixed (p, rp)) t.type_ }
 
 let add_class :
-    Ident.class_ -> Cpath.class_type -> Cpath.Resolved.class_type -> t -> t =
+    Ident.type_ -> Cpath.class_type -> Cpath.Resolved.class_type -> t -> t =
  fun id p rp t ->
   {
     t with
     type_ =
-      PathTypeMap.add
-        (id :> Ident.path_type)
+      TypeMap.add
+        (id :> Ident.type_)
         (`Prefixed ((p :> Cpath.type_), (rp :> Cpath.Resolved.type_)))
         t.type_;
     class_type =
-      PathClassTypeMap.add
-        (id :> Ident.path_class_type)
-        (`Prefixed (p, rp))
-        t.class_type;
+      TypeMap.add (id :> Ident.type_) (`Prefixed (p, rp)) t.class_type;
   }
 
 let add_class_type :
-    Ident.class_type -> Cpath.class_type -> Cpath.Resolved.class_type -> t -> t
-    =
+    Ident.type_ -> Cpath.class_type -> Cpath.Resolved.class_type -> t -> t =
  fun id p rp t ->
   {
     t with
     type_ =
-      PathTypeMap.add
-        (id :> Ident.path_type)
+      TypeMap.add
+        (id :> Ident.type_)
         (`Prefixed ((p :> Cpath.type_), (rp :> Cpath.Resolved.type_)))
         t.type_;
     class_type =
-      PathClassTypeMap.add
-        (id :> Ident.path_class_type)
-        (`Prefixed (p, rp))
-        t.class_type;
+      TypeMap.add (id :> Ident.type_) (`Prefixed (p, rp)) t.class_type;
   }
 
 let add_type_replacement id texp equation t =
   {
     t with
-    type_replacement = PathTypeMap.add id (texp, equation) t.type_replacement;
+    type_replacement = TypeMap.add id (texp, equation) t.type_replacement;
   }
 
 let add_module_type_replacement path mty t =
@@ -108,38 +88,31 @@ let add_module_type_replacement path mty t =
       ModuleTypeMap.add path mty t.module_type_replacement;
   }
 
-let add_module_substitution : Ident.path_module -> t -> t =
+let add_module_substitution : Ident.module_ -> t -> t =
  fun id t ->
   {
     t with
-    module_type_of_invalidating_modules =
-      id :: t.module_type_of_invalidating_modules;
     path_invalidating_modules = id :: t.path_invalidating_modules;
-    module_ = PathModuleMap.add id `Substituted t.module_;
+    module_ = ModuleMap.add id `Substituted t.module_;
   }
 
-let rename_module : Ident.path_module -> Ident.path_module -> t -> t =
- fun id id' t ->
-  { t with module_ = PathModuleMap.add id (`Renamed id') t.module_ }
+let rename_module : Ident.module_ -> Ident.module_ -> t -> t =
+ fun id id' t -> { t with module_ = ModuleMap.add id (`Renamed id') t.module_ }
 
 let rename_module_type : Ident.module_type -> Ident.module_type -> t -> t =
  fun id id' t ->
   { t with module_type = ModuleTypeMap.add id (`Renamed id') t.module_type }
 
-let rename_type : Ident.path_type -> Ident.path_type -> t -> t =
- fun id id' t -> { t with type_ = PathTypeMap.add id (`Renamed id') t.type_ }
+let rename_type : Ident.type_ -> Ident.type_ -> t -> t =
+ fun id id' t -> { t with type_ = TypeMap.add id (`Renamed id') t.type_ }
 
-let rename_class_type : Ident.path_class_type -> Ident.path_class_type -> t -> t
-    =
+let rename_class_type : Ident.type_ -> Ident.type_ -> t -> t =
  fun id id' t ->
   {
     t with
-    class_type = PathClassTypeMap.add id (`Renamed id') t.class_type;
+    class_type = TypeMap.add id (`Renamed id') t.class_type;
     type_ =
-      PathTypeMap.add
-        (id :> Ident.path_type)
-        (`Renamed (id' :> Ident.path_type))
-        t.type_;
+      TypeMap.add (id :> Ident.type_) (`Renamed (id' :> Ident.type_)) t.type_;
   }
 
 let rec substitute_vars vars t =
@@ -191,7 +164,7 @@ let rec resolved_module_path :
   | `Local id -> (
       if List.mem id s.path_invalidating_modules then raise Invalidated;
       match
-        try Some (PathModuleMap.find (id :> Ident.path_module) s.module_)
+        try Some (ModuleMap.find (id :> Ident.module_) s.module_)
         with _ -> None
       with
       | Some (`Renamed x) -> `Local x
@@ -252,7 +225,7 @@ and module_path : t -> Cpath.module_ -> Cpath.module_ =
   | `Apply (p1, p2) -> `Apply (module_path s p1, module_path s p2)
   | `Local (id, b) -> (
       match
-        try Some (PathModuleMap.find (id :> Ident.path_module) s.module_)
+        try Some (ModuleMap.find (id :> Ident.module_) s.module_)
         with _ -> None
       with
       | Some (`Prefixed (p, _rp)) -> p
@@ -331,7 +304,7 @@ and module_type_path :
         in
         Not_replaced r
   | `Identifier _ -> Not_replaced p
-  | `Dot (p, n) -> Not_replaced (`Dot (module_path s p, n))
+  | `DotMT (p, n) -> Not_replaced (`DotMT (module_path s p, n))
   | `ModuleType (p', str) ->
       Not_replaced (`ModuleType (resolved_parent_path s p', str))
 
@@ -341,13 +314,12 @@ and resolved_type_path :
     (Cpath.Resolved.type_, TypeExpr.t * TypeDecl.Equation.t) or_replaced =
  fun s p ->
   match p with
+  | `CoreType _ as c -> Not_replaced c
   | `Local id -> (
-      if PathTypeMap.mem id s.type_replacement then
-        Replaced (PathTypeMap.find id s.type_replacement)
+      if TypeMap.mem id s.type_replacement then
+        Replaced (TypeMap.find id s.type_replacement)
       else
-        match
-          try Some (PathTypeMap.find id s.type_) with Not_found -> None
-        with
+        match try Some (TypeMap.find id s.type_) with Not_found -> None with
         | Some (`Prefixed (_p, rp)) -> Not_replaced rp
         | Some (`Renamed x) -> Not_replaced (`Local x)
         | None -> Not_replaced (`Local id))
@@ -372,17 +344,15 @@ and type_path : t -> Cpath.type_ -> Cpath.type_ type_or_replaced =
         type_path s path')
   | `Substituted p -> type_path s p |> map_replaced (fun r -> `Substituted r)
   | `Local (id, b) -> (
-      if PathTypeMap.mem id s.type_replacement then
-        Replaced (PathTypeMap.find id s.type_replacement)
+      if TypeMap.mem id s.type_replacement then
+        Replaced (TypeMap.find id s.type_replacement)
       else
-        match
-          try Some (PathTypeMap.find id s.type_) with Not_found -> None
-        with
+        match try Some (TypeMap.find id s.type_) with Not_found -> None with
         | Some (`Prefixed (p, _rp)) -> Not_replaced p
         | Some (`Renamed x) -> Not_replaced (`Local (x, b))
         | None -> Not_replaced (`Local (id, b)))
   | `Identifier _ -> Not_replaced p
-  | `Dot (p, n) -> Not_replaced (`Dot (module_path s p, n))
+  | `DotT (p, n) -> Not_replaced (`DotT (module_path s p, n))
   | `Type (p, n) -> Not_replaced (`Type (resolved_parent_path s p, n))
   | `Class (p, n) -> Not_replaced (`Class (resolved_parent_path s p, n))
   | `ClassType (p, n) -> Not_replaced (`ClassType (resolved_parent_path s p, n))
@@ -392,9 +362,7 @@ and resolved_class_type_path :
  fun s p ->
   match p with
   | `Local id -> (
-      match
-        try Some (PathClassTypeMap.find id s.class_type) with _ -> None
-      with
+      match try Some (TypeMap.find id s.class_type) with _ -> None with
       | Some (`Prefixed (_p, rp)) -> rp
       | Some (`Renamed x) -> `Local x
       | None -> `Local id)
@@ -412,15 +380,13 @@ and class_type_path : t -> Cpath.class_type -> Cpath.class_type =
         let path' = Cpath.unresolve_resolved_class_type_path r in
         class_type_path s path')
   | `Local (id, b) -> (
-      match
-        try Some (PathClassTypeMap.find id s.class_type) with _ -> None
-      with
+      match try Some (TypeMap.find id s.class_type) with _ -> None with
       | Some (`Prefixed (p, _rp)) -> p
       | Some (`Renamed x) -> `Local (x, b)
       | None -> `Local (id, b))
   | `Identifier _ -> p
   | `Substituted p -> `Substituted (class_type_path s p)
-  | `Dot (p, n) -> `Dot (module_path s p, n)
+  | `DotT (p, n) -> `DotT (module_path s p, n)
   | `Class (p, n) -> `Class (resolved_parent_path s p, n)
   | `ClassType (p, n) -> `ClassType (resolved_parent_path s p, n)
 
@@ -613,7 +579,7 @@ and module_type s t =
   let expr =
     match t.expr with Some m -> Some (module_type_expr s m) | None -> None
   in
-  { expr; locs = t.locs; doc = t.doc; canonical = t.canonical }
+  { expr; source_loc = t.source_loc; doc = t.doc; canonical = t.canonical }
 
 and module_type_substitution s t =
   let open Component.ModuleTypeSubstitution in
@@ -627,51 +593,11 @@ and functor_parameter s t =
   | Unit -> Unit
 
 and module_type_type_of_desc s t =
-  let open Component.ModuleType in
-  match t with
-  | ModPath p ->
-      if mto_module_path_invalidated s p then raise MTOInvalidated
-      else ModPath (module_path s p)
-  | StructInclude p ->
-      if mto_module_path_invalidated s p then raise MTOInvalidated
-      else StructInclude (module_path s p)
-
-and module_type_type_of_desc_noexn s t =
   let open Component.ModuleType in
   match t with
   | ModPath p -> ModPath (module_path s p)
   | StructInclude p -> StructInclude (module_path s p)
 
-and mto_module_path_invalidated : t -> Cpath.module_ -> bool =
- fun s p ->
-  match p with
-  | `Resolved p' -> mto_resolved_module_path_invalidated s p'
-  | `Substituted p' | `Dot (p', _) -> mto_module_path_invalidated s p'
-  | `Module (`Module p', _) -> mto_resolved_module_path_invalidated s p'
-  | `Module (_, _) -> false
-  | `Apply (p1, p2) ->
-      mto_module_path_invalidated s p1 || mto_module_path_invalidated s p2
-  | `Local (id, _) -> List.mem id s.module_type_of_invalidating_modules
-  | `Identifier _ -> false
-  | `Forward _ -> false
-  | `Root _ -> false
-
-and mto_resolved_module_path_invalidated s p =
-  match p with
-  | `Local id -> List.mem id s.module_type_of_invalidating_modules
-  | `Gpath _ -> false
-  | `Apply (p1, p2) ->
-      mto_resolved_module_path_invalidated s p1
-      || mto_resolved_module_path_invalidated s p2
-  | `Module (`Module p, _) | `Substituted p ->
-      mto_resolved_module_path_invalidated s p
-  | `Module (_, _) -> false
-  | `Alias (p1, _p2, _) -> mto_resolved_module_path_invalidated s p1
-  | `Subst (_p1, p2) -> mto_resolved_module_path_invalidated s p2
-  | `Hidden p -> mto_resolved_module_path_invalidated s p
-  | `Canonical (p1, _p2) -> mto_resolved_module_path_invalidated s p1
-  | `OpaqueModule p -> mto_resolved_module_path_invalidated s p
-
 and u_module_type_expr s t =
   let open Component.ModuleType.U in
   match t with
@@ -682,7 +608,7 @@ and u_module_type_expr s t =
           match eqn with
           | Path p -> Path p.p_path
           | Signature s -> Signature s
-          | TypeOf t -> TypeOf t
+          | TypeOf t -> TypeOf (t.t_desc, t.t_original_path)
           | With w -> With (w.w_substitutions, w.w_expr)
           | Functor _ ->
               (* non functor cannot be substituted away to a functor *)
@@ -691,24 +617,8 @@ and u_module_type_expr s t =
   | With (subs, e) ->
       With
         (List.map (with_module_type_substitution s) subs, u_module_type_expr s e)
-  | TypeOf { t_desc; t_expansion = Some (Signature e) } -> (
-      try
-        TypeOf
-          {
-            t_desc = module_type_type_of_desc s t_desc;
-            t_expansion = Some (Signature (apply_sig_map_sg s e));
-          }
-      with MTOInvalidated -> u_module_type_expr s (Signature e))
-  | TypeOf { t_expansion = Some (Functor _); _ } -> assert false
-  | TypeOf { t_desc; t_expansion = None } ->
-      TypeOf
-        { t_desc = module_type_type_of_desc_noexn s t_desc; t_expansion = None }
-
-and module_type_of_simple_expansion :
-    Component.ModuleType.simple_expansion -> Component.ModuleType.expr =
-  function
-  | Signature sg -> Signature sg
-  | Functor (arg, e) -> Functor (arg, module_type_of_simple_expansion e)
+  | TypeOf (t_desc, t_original_path) ->
+      TypeOf (module_type_type_of_desc s t_desc, t_original_path)
 
 and module_type_expr s t =
   let open Component.ModuleType in
@@ -729,18 +639,13 @@ and module_type_expr s t =
           w_expansion = option_ simple_expansion s w_expansion;
           w_expr = u_module_type_expr s w_expr;
         }
-  | TypeOf { t_desc; t_expansion = Some e } -> (
-      try
-        TypeOf
-          {
-            t_desc = module_type_type_of_desc s t_desc;
-            t_expansion = Some (simple_expansion s e);
-          }
-      with MTOInvalidated ->
-        module_type_expr s (module_type_of_simple_expansion e))
-  | TypeOf { t_desc; t_expansion = None } ->
+  | TypeOf t ->
       TypeOf
-        { t_desc = module_type_type_of_desc_noexn s t_desc; t_expansion = None }
+        {
+          t with
+          t_desc = module_type_type_of_desc s t.t_desc;
+          t_expansion = option_ simple_expansion s t.t_expansion;
+        }
 
 and with_module_type_substitution s sub =
   let open Component.ModuleType in
@@ -894,7 +799,7 @@ and rename_bound_idents s sg =
   let open Component.Signature in
   let new_module_id id =
     try
-      match PathModuleMap.find (id :> Ident.path_module) s.module_ with
+      match ModuleMap.find (id :> Ident.module_) s.module_ with
       | `Renamed (`LModule _ as x) -> x
       | `Prefixed (_, _) ->
           (* This is unusual but can happen when we have TypeOf expressions. It means
@@ -913,44 +818,37 @@ and rename_bound_idents s sg =
   in
   let new_type_id id =
     try
-      match PathTypeMap.find (id :> Ident.path_type) s.type_ with
+      match TypeMap.find (id :> Ident.type_) s.type_ with
       | `Renamed (`LType _ as x) -> x
       | `Prefixed (_, _) -> Ident.Rename.type_ id
-      | _ -> failwith "Error"
     with Not_found -> Ident.Rename.type_ id
   in
   let new_class_id id =
     try
-      match
-        PathClassTypeMap.find (id :> Ident.path_class_type) s.class_type
-      with
-      | `Renamed (`LClass _ as x) -> x
-      | `Prefixed (_, _) -> Ident.Rename.class_ id
-      | _ -> failwith "Error"
-    with Not_found -> Ident.Rename.class_ id
+      match TypeMap.find (id :> Ident.type_) s.class_type with
+      | `Renamed (`LType _ as x) -> x
+      | `Prefixed (_, _) -> Ident.Rename.type_ id
+    with Not_found -> Ident.Rename.type_ id
   in
   let new_class_type_id id =
     try
-      match
-        PathClassTypeMap.find (id :> Ident.path_class_type) s.class_type
-      with
-      | `Renamed (`LClassType _ as x) -> x
-      | `Prefixed (_, _) -> Ident.Rename.class_type id
-      | _ -> failwith "Error!"
-    with Not_found -> Ident.Rename.class_type id
+      match TypeMap.find (id :> Ident.type_) s.class_type with
+      | `Renamed (`LType _ as x) -> x
+      | `Prefixed (_, _) -> Ident.Rename.type_ id
+    with Not_found -> Ident.Rename.type_ id
   in
   function
   | [] -> (s, List.rev sg)
   | Module (id, r, m) :: rest ->
       let id' = new_module_id id in
       rename_bound_idents
-        (rename_module (id :> Ident.path_module) (id' :> Ident.path_module) s)
+        (rename_module (id :> Ident.module_) (id' :> Ident.module_) s)
         (Module (id', r, m) :: sg)
         rest
   | ModuleSubstitution (id, m) :: rest ->
       let id' = new_module_id id in
       rename_bound_idents
-        (rename_module (id :> Ident.path_module) (id' :> Ident.path_module) s)
+        (rename_module (id :> Ident.module_) (id' :> Ident.module_) s)
         (ModuleSubstitution (id', m) :: sg)
         rest
   | ModuleType (id, mt) :: rest ->
@@ -968,13 +866,13 @@ and rename_bound_idents s sg =
   | Type (id, r, t) :: rest ->
       let id' = new_type_id id in
       rename_bound_idents
-        (rename_type (id :> Ident.path_type) (id' :> Ident.path_type) s)
+        (rename_type (id :> Ident.type_) (id' :> Ident.type_) s)
         (Type (id', r, t) :: sg)
         rest
   | TypeSubstitution (id, t) :: rest ->
       let id' = new_type_id id in
       rename_bound_idents
-        (rename_type (id :> Ident.path_type) (id' :> Ident.path_type) s)
+        (rename_type (id :> Ident.type_) (id' :> Ident.type_) s)
         (TypeSubstitution (id', t) :: sg)
         rest
   | Exception (id, e) :: rest ->
@@ -987,19 +885,13 @@ and rename_bound_idents s sg =
   | Class (id, r, c) :: rest ->
       let id' = new_class_id id in
       rename_bound_idents
-        (rename_class_type
-           (id :> Ident.path_class_type)
-           (id' :> Ident.path_class_type)
-           s)
+        (rename_class_type (id :> Ident.type_) (id' :> Ident.type_) s)
         (Class (id', r, c) :: sg)
         rest
   | ClassType (id, r, c) :: rest ->
       let id' = new_class_type_id id in
       rename_bound_idents
-        (rename_class_type
-           (id :> Ident.path_class_type)
-           (id' :> Ident.path_class_type)
-           s)
+        (rename_class_type (id :> Ident.type_) (id' :> Ident.type_) s)
         (ClassType (id', r, c) :: sg)
         rest
   | Include ({ expansion_; _ } as i) :: rest ->
@@ -1019,13 +911,10 @@ and removed_items s items =
   let open Component.Signature in
   List.map
     (function
-      | RModule (id, _) as x -> (
-          try
-            match PathModuleMap.find (id :> Ident.path_module) s.module_ with
-            | `Prefixed (_, x) -> RModule (id, x)
-            | _ -> x
-          with Not_found -> x)
-      | x -> x)
+      | RModule (id, p) -> RModule (id, module_path s p)
+      | RType (id, exp, eqn) ->
+          RType (id, type_expr s exp, type_decl_equation s eqn)
+      | RModuleType (id, mty) -> RModuleType (id, module_type_expr s mty))
     items
 
 and signature s sg =
@@ -1037,60 +926,44 @@ and apply_sig_map_sg s (sg : Component.Signature.t) =
   let items, removed, dont_recompile = apply_sig_map s sg.items sg.removed in
   { sg with items; removed; compiled = sg.compiled && dont_recompile }
 
-and apply_sig_map s items removed =
+and apply_sig_map_item s item =
   let open Component.Signature in
-  let rec inner items acc =
-    match items with
-    | [] -> List.rev acc
-    | Module (id, r, m) :: rest ->
-        inner rest
-          (Module
-             ( id,
-               r,
-               Component.Delayed.put (fun () ->
-                   module_ s (Component.Delayed.get m)) )
-          :: acc)
-    | ModuleSubstitution (id, m) :: rest ->
-        inner rest (ModuleSubstitution (id, module_substitution s m) :: acc)
-    | ModuleType (id, mt) :: rest ->
-        inner rest
-          (ModuleType
-             ( id,
-               Component.Delayed.put (fun () ->
-                   module_type s (Component.Delayed.get mt)) )
-          :: acc)
-    | ModuleTypeSubstitution (id, mt) :: rest ->
-        inner rest
-          (ModuleTypeSubstitution (id, module_type_substitution s mt) :: acc)
-    | Type (id, r, t) :: rest ->
-        inner rest
-          (Type
-             ( id,
-               r,
-               Component.Delayed.put (fun () ->
-                   type_ s (Component.Delayed.get t)) )
-          :: acc)
-    | TypeSubstitution (id, t) :: rest ->
-        inner rest (TypeSubstitution (id, type_ s t) :: acc)
-    | Exception (id, e) :: rest ->
-        inner rest (Exception (id, exception_ s e) :: acc)
-    | TypExt e :: rest -> inner rest (TypExt (extension s e) :: acc)
-    | Value (id, v) :: rest ->
-        inner rest
-          (Value
-             ( id,
-               Component.Delayed.put (fun () ->
-                   value s (Component.Delayed.get v)) )
-          :: acc)
-    | Class (id, r, c) :: rest -> inner rest (Class (id, r, class_ s c) :: acc)
-    | ClassType (id, r, c) :: rest ->
-        inner rest (ClassType (id, r, class_type s c) :: acc)
-    | Include i :: rest -> inner rest (Include (include_ s i) :: acc)
-    | Open o :: rest -> inner rest (Open (open_ s o) :: acc)
-    | Comment c :: rest -> inner rest (Comment c :: acc)
-  in
-  let dont_recompile =
-    List.length s.path_invalidating_modules = 0
-    && List.length s.module_type_of_invalidating_modules = 0
-  in
-  (inner items [], removed_items s removed, dont_recompile)
+  match item with
+  | Module (id, r, m) ->
+      Module
+        ( id,
+          r,
+          Component.Delayed.put (fun () -> module_ s (Component.Delayed.get m))
+        )
+  | ModuleSubstitution (id, m) ->
+      ModuleSubstitution (id, module_substitution s m)
+  | ModuleType (id, mt) ->
+      ModuleType
+        ( id,
+          Component.Delayed.put (fun () ->
+              module_type s (Component.Delayed.get mt)) )
+  | ModuleTypeSubstitution (id, mt) ->
+      ModuleTypeSubstitution (id, module_type_substitution s mt)
+  | Type (id, r, t) ->
+      Type
+        ( id,
+          r,
+          Component.Delayed.put (fun () -> type_ s (Component.Delayed.get t)) )
+  | TypeSubstitution (id, t) -> TypeSubstitution (id, type_ s t)
+  | Exception (id, e) -> Exception (id, exception_ s e)
+  | TypExt e -> TypExt (extension s e)
+  | Value (id, v) ->
+      Value
+        (id, Component.Delayed.put (fun () -> value s (Component.Delayed.get v)))
+  | Class (id, r, c) -> Class (id, r, class_ s c)
+  | ClassType (id, r, c) -> ClassType (id, r, class_type s c)
+  | Include i -> Include (include_ s i)
+  | Open o -> Open (open_ s o)
+  | Comment c -> Comment c
+
+and apply_sig_map_items s items =
+  List.rev_map (apply_sig_map_item s) items |> List.rev
+
+and apply_sig_map s items removed =
+  let dont_recompile = List.length s.path_invalidating_modules = 0 in
+  (apply_sig_map_items s items, removed_items s removed, dont_recompile)
