@@ -57,6 +57,10 @@
     Code can appear {b inside [other] markup}. Its display shouldn't be
     affected.
 
+    There is no differences between [a b] and [a
+    b].
+
+    Consecutive whitespaces not after a newline are conserved as they are: [a   b].
 
     {1 Links and references}
 
