@@ -1,4 +1,5 @@
 open Result
+open Odoc_model.Names
 
 (* Example usage of these:
 
@@ -40,8 +41,8 @@ let cmt_of_string s =
     Typemod.type_implementation Unit_info.(make ~source_file:"" Impl "") env p
 #endif
 
-let parent = Odoc_model.Paths.Identifier.Mk.page (None, Odoc_model.Names.PageName.make_std "None")
-let id = Odoc_model.Paths.Identifier.Mk.root (Some parent, Odoc_model.Names.ModuleName.make_std "Root")
+let parent = Odoc_model.Paths.Identifier.Mk.page (None, PageName.make_std "None")
+let id = Odoc_model.Paths.Identifier.Mk.root (Some parent, ModuleName.make_std "Root")
 
 let root_of_compilation_unit ~package ~hidden ~module_name ~digest =
   ignore(package);
@@ -63,13 +64,13 @@ let root =
 
 let root_identifier = `Identifier id
 
-let root_module name = Odoc_model.Paths.Identifier.Mk.module_ (id, Odoc_model.Names.ModuleName.make_std name)
+let root_module name = Odoc_model.Paths.Identifier.Mk.module_ (id, ModuleName.make_std name)
 
 let root_pp fmt (_ : Odoc_model.Root.t) = Format.fprintf fmt "Common.root"
 
 let model_of_string str = 
     let cmti = cmti_of_string str in
-    Odoc_loader__Cmti.read_interface (Some parent) "Root" cmti
+    Odoc_loader__Cmti.read_interface (Some parent) "Root" ~warnings_tag:None cmti
 
 let model_of_string_impl str =
 #if OCAML_VERSION < (4,13,0)
@@ -77,43 +78,13 @@ let model_of_string_impl str =
 #else
     let cmt = (cmt_of_string str).structure in
 #endif
-    Odoc_loader__Cmt.read_implementation (Some parent) "Root" cmt
+    Odoc_loader__Cmt.read_implementation (Some parent) "Root" ~warnings_tag:None cmt
 
 let signature_of_mli_string str =
     Odoc_xref2.Ident.reset ();
     let _, sg, _ = model_of_string str in
     sg
 
-let string_of_file f =
-    let ic = open_in f in
-    let buffer = Buffer.create 100 in
-    let rec loop () =
-        try
-            Buffer.add_channel buffer ic 1024;
-            loop ()
-        with End_of_file ->
-            ()
-    in loop ();
-    close_in ic;
-    Buffer.contents buffer
-
-let file_of_string ~filename str =
-    let oc = open_out filename in
-    Printf.fprintf oc "%s%!" str;
-    close_out oc
-
-let list_files path =
-    Sys.readdir path |> Array.to_list
-
-let load_cmti filename =
-  let make_root = root_of_compilation_unit ~package:"nopackage" ~hidden:false in
-  Odoc_loader.read_cmti ~make_root ~filename
-
-let load_cmt filename =
-    let make_root = root_of_compilation_unit ~package:"nopackage" ~hidden:false in
-    Odoc_loader.read_cmt ~make_root ~filename
-  
-
 module Ident = Ident
 
 module LangUtils = struct
@@ -407,7 +378,7 @@ module LangUtils = struct
                         let id = m.Odoc_model.Lang.Module.id in
                         match id.iv with
                         | `Module (_, mname') ->
-                            if Odoc_model.Names.ModuleName.to_string mname' = mname
+                            if ModuleName.to_string mname' = mname
                             then m
                             else inner rest
                         | _ -> inner rest
@@ -456,7 +427,7 @@ module LangUtils = struct
 
         let identifier ppf i =
             Format.fprintf ppf "%a"
-                Odoc_xref2.Component.Fmt.model_identifier
+                Odoc_xref2.Component.Fmt.(model_identifier default)
                 (i :> Odoc_model.Paths.Identifier.t)
 
         let rec signature ppf sg =
@@ -567,20 +538,23 @@ module LangUtils = struct
             | `Alias (dest, src) -> Format.fprintf ppf "(%a -> %a)" path (src :> Odoc_model.Paths.Path.t) resolved_path (cast dest)
             | `AliasModuleType (path, realpath) -> Format.fprintf ppf "(%a -> %a)" resolved_path (cast path) resolved_path (cast realpath)
             | `Subst (modty, m) -> Format.fprintf ppf "(%a subst-> %a)" resolved_path (cast modty) resolved_path (cast m)
-            | `Module (p, m) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (Odoc_model.Names.ModuleName.to_string m)
-            | `ModuleType (p, mt) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (Odoc_model.Names.ModuleTypeName.to_string mt)
-            | `Type (p, t) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (Odoc_model.Names.TypeName.to_string t)
-            | `Value (p, t) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (Odoc_model.Names.ValueName.to_string t)
-            | `Constructor (p, t) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (Odoc_model.Names.ConstructorName.to_string t)
+            | `Module (p, m) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (ModuleName.to_string m)
+            | `ModuleType (p, mt) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (ModuleTypeName.to_string mt)
+            | `Type (p, t) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (TypeName.to_string t)
+            | `Value (p, t) -> Format.fprintf ppf "%a.%s" resolved_path (cast p) (ValueName.to_string t)
             | `OpaqueModule m -> Format.fprintf ppf "opaquemodule(%a)" resolved_path (cast m)
             | `OpaqueModuleType m -> Format.fprintf ppf "opaquemoduletype(%a)" resolved_path (cast m)
             | `SubstT (_, _)
             | `CanonicalModuleType (_, _)
             | `CanonicalType (_, _)
-            | `CanonicalDataType (_, _)
             | `Class (_, _)
             | `ClassType (_, _)
             | `Hidden _
+            | `SubstitutedT _
+            | `SubstitutedMT _
+            | `CoreType _
+            | `Substituted _
+            | `SubstitutedCT _
             | `Canonical _ -> Format.fprintf ppf "unimplemented resolved_path"
 
         and path : Format.formatter -> Odoc_model.Paths.Path.t -> unit =
@@ -588,10 +562,14 @@ module LangUtils = struct
             match p with
             | `Resolved rp -> Format.fprintf ppf "resolved[%a]" resolved_path (rp :> Odoc_model.Paths.Path.Resolved.t)
             | `Identifier (i,b) -> Format.fprintf ppf "identifier(%a,%b)" identifier i b
-            | `Root s -> Format.fprintf ppf "%s" s
+            | `Root s -> Format.fprintf ppf "%a" ModuleName.fmt s
             | `Forward s -> Format.fprintf ppf "%s" s
-            | `Dot (parent,s) -> Format.fprintf ppf "%a.%s" path (parent :> Odoc_model.Paths.Path.t) s
+            | `Dot (parent,s) -> Format.fprintf ppf "%a.%a" path (parent :> Odoc_model.Paths.Path.t) ModuleName.fmt s
+            | `DotMT (parent,s) -> Format.fprintf ppf "%a.%a" path (parent :> Odoc_model.Paths.Path.t) ModuleTypeName.fmt s
+            | `DotT (parent,s) -> Format.fprintf ppf "%a.%a" path (parent :> Odoc_model.Paths.Path.t) TypeName.fmt s
+            | `DotV (parent,s) -> Format.fprintf ppf "%a.%a" path (parent :> Odoc_model.Paths.Path.t) ValueName.fmt s
             | `Apply (func,arg) -> Format.fprintf ppf "%a(%a)" path (func :> Odoc_model.Paths.Path.t) path (arg :> Odoc_model.Paths.Path.t)
+            | `SubstitutedT _|`SubstitutedMT _|`Substituted _|`SubstitutedCT _  -> Format.fprintf ppf "Unimplemented path"
 
         and model_fragment ppf (f : Odoc_model.Paths.Fragment.t) =
             match f with
@@ -603,8 +581,8 @@ module LangUtils = struct
             match f with
             | `Root (`Module p) -> Format.fprintf ppf "root_module(%a)" resolved_path (p :> Odoc_model.Paths.Path.Resolved.t) 
             | `Root (`ModuleType p) -> Format.fprintf ppf "root_module_type(%a)" resolved_path (p :> Odoc_model.Paths.Path.Resolved.t) 
-            | `Module (sg, m) -> Format.fprintf ppf "%a.%s" model_resolved_fragment (sg :> Odoc_model.Paths.Fragment.Resolved.t) (Odoc_model.Names.ModuleName.to_string m)
-            | `Type (sg, m) -> Format.fprintf ppf "%a.%s" model_resolved_fragment (sg :> Odoc_model.Paths.Fragment.Resolved.t) (Odoc_model.Names.TypeName.to_string m)
+            | `Module (sg, m) -> Format.fprintf ppf "%a.%s" model_resolved_fragment (sg :> Odoc_model.Paths.Fragment.Resolved.t) (ModuleName.to_string m)
+            | `Type (sg, m) -> Format.fprintf ppf "%a.%s" model_resolved_fragment (sg :> Odoc_model.Paths.Fragment.Resolved.t) (TypeName.to_string m)
             | _ -> Format.fprintf ppf "UNIMPLEMENTED model_resolved_fragment"
 
     end
@@ -624,13 +602,12 @@ let my_compilation_unit id (s : Odoc_model.Lang.Signature.t) =
     ; expansion = None
     ; linked = false
     ; canonical = None
-    ; source_info = None
-    ; shape_info = None
+    ; source_loc = None
 }
 
 let mkresolver () =
   Odoc_odoc.Resolver.create
-    ~important_digests:false
+ ~roots:None     ~important_digests:false
     ~directories:(List.map Odoc_odoc.Fs.Directory.of_string
 #if OCAML_VERSION >= (5,2,0)
   (let paths = Load_path.get_paths () in
@@ -643,7 +620,7 @@ let mkresolver () =
     ) ~open_modules:[]
 
 let warnings_options =
-  { Odoc_model.Error.warn_error = false; print_warnings = true }
+  { Odoc_model.Error.warn_error = false; print_warnings = true; warnings_tag = None }
 
 let handle_warnings ww =
   match Odoc_model.Error.handle_warnings ~warnings_options ww with
